#!/bin/sh
# -*- tcl -*- \
exec wish8.0 "$0" "$@"

#	Copyright (c) 1996,1997 Australian National University
#
#--------
#
# Copyright (c) 1995,1996 ANU and CSIRO on behalf of the participants in the 
# CRC for Advanced Computational Systems ('ACSys').
#
# ACSys makes this software and all associated data and documentation
# ('Software') available free of charge for non-commercial purposes only. You
# may make copies of the Software but you must include all of this notice on
# any copy.
#
# The Software was developed for research purposes and ACSys does not warrant
# that it is error free or fit for any purpose.  ACSys disclaims any
# liability for all claims, expenses, losses, damages and costs any user may
# incur as a result of using, copying or modifying the Software.

array set scripts [list \
plume-1.tcl {# Check that we're running the correct version

if {[info tclversion] < 7.6} {
    puts stderr "This application requires Tcl 7.6 or better.  This is only Tcl [info tclversion]"
    exit 1
}

if {![info exists tk_version]} {
    puts stderr "This application requires Tk"
    exit 1
}
if {$tk_version < 4.2} {
    puts stderr "This application requires Tk 4.2 or better.  This is only Tk $tk_version"
    exit 1
}

# This version uses features of Tk that are incompatible with Tk 4.2 beta releases.
# Eg. grid command options.
if {[regexp {4\.2(a|b).*} $tk_patchLevel]} {
    puts stderr "This application will not work with a Tk 4.2 beta release"
    exit 1
}

# Process command line arguments - pretty dodgey!

if {[info exists gk_library]} {
    gk_initConf $argv
} else {
    if {[catch {array set argva $argv}]} {
	puts stderr "no value given for option [lindex $argv end]"
	exit 2
    }
}

if {[file readable prot.tcl]} {
    # Allow it to work in the source directory before installation
    set auto_path [linsert $auto_path 0 .]
}

if {[info exists argva(-debug)]} {
    if {[string match *.* $argva(-debug)]} {
	set debug_fd [open $argva(-debug) w]
	fconfigure $debug_fd -buffering line
	proc debug_puts {text} {global debug_fd; puts $debug_fd $text}
    } elseif {$argva(-debug)} {
	proc debug_puts {text} {puts stderr $text}
    }
} else {
    proc debug_puts {args} {}
}

# Set defaults 

package require utilities 1.0

array set surfit [list version [list Plume 0.6.2alpha] CFGwwwhome {} userdir [File_pwd] CFGusername nobody]

switch $tcl_platform(platform) {
    macintosh {
	set surfit(CFGusername) $env(USER)
	set surfit(userdir) [file split :Plume:]
    }
    windows {
	catch {set surfit(CFGusername) $env(USER)}
	
	set surfit(userdir) [file split C:\\plume]
	catch {set surfit(userdir) [file split [File_normalizeFile $env(TEMP)]]}
	catch {set surfit(userdir) [file split [File_normalizeFile $env(TMP)]]}
	catch {set surfit(userdir) [file split [File_normalizeFile $env(HOME)]]}
	lappend surfit(userdir) plume
	catch {set surfit(userdir) [file split [File_normalizeFile $env(PLUMEDIR)]]} 
    }
    unix -
    default {
	catch {set surfit(CFGusername) $env(USER)}
	if {[file isdirectory [eval file join [file split $env(HOME)] .plume]]} {
	    set surfit(userdir) "[file split $env(HOME)] .plume"
	}
	catch {set surfit(userdir) [file split $env(PLUMEDIR)]}
    }
}

# cmd line args override defaults, environment settings

if {[info exists argva(-userdir)]} {
    set surfit(userdir) [file split [File_normalizeFile $argva(-userdir)]]
}

} \
plume-2.tcl {
###
### Setup environment
###

# Basic utility modules

package require fsbox 1.0

# General modules

package require uri 1.0
package require protocols 1.0
package require cache 1.0
set htmllib htmllib.tcl
package require html_library 1.0
package require safe-tk 1.0

# Instantiate overridden packages
PR_init
HMlibrary_init
interp_init

# Application-specific modules

package require applets 1.0; Applet_init
package require handlers 1.0; handlers_init
package require hypertools 1.0; HT_init
package require mailto 1.0; mailto_init
package require internals 1.0; surfit_init

# Third-party packages

# Jan Nijtman's Img package provides better image support
# as well as JPEG, TIFF images
catch {package require Img}

# Load icon entities into images
# Find the directory containing the icons
if {[info exists env(SURFIT_ICONPATH)]} {
    set iconPath [file split [File_normalizeFile $env(SURFIT_ICONPATH)]]
} elseif {[file isdirectory [eval file join "$datadir icons"]]} {
    set iconPath "$datadir icons"
} elseif {[file isdirectory [eval file join "$tk_library icons"]]} {
    set iconPath "$tk_library icons"
}
if {[info exists iconPath] && $iconPath != {}} {
    foreach i [glob [eval file join "$iconPath *.xbm"]] {
	regsub -all _ [file tail [file rootname $i]] . img
	catch {image create bitmap ICON$img -file $i}
    }
}


} \
plume-3.tcl {
# Check whether the user directory exists

if {[file exists [eval file join $surfit(userdir)]]} {
    if {![file isdirectory [eval file join $surfit(userdir)]]} {
	tk_dialog .err Error "Plume program directory \"[eval file join $surfit(userdir)]\" already exists " {} 0 OK
	exit 0
    }
} else {
    File_mkdir $surfit(userdir)
}

# Now invoke the user's preferences file, if any.
# This allows the user to override any standard definitions.

catch {source [eval file join $surfit(userdir) Prefs.tcl]}

# Once preferences are known, initialise modules

Cache_init

if {[info exists argva(-home)]} {
    set wwwhome $argva(-home)
} elseif {[info exists env(WWW_HOME)]} {
    set wwwhome $env(WWW_HOME)
} elseif {[info exists surfit(CFGwwwhome)] && $surfit(CFGwwwhome) != {}} {
    set wwwhome $surfit(CFGwwwhome)
} else {
    set wwwhome about:plume
}

###
### Window management 
###

set surfit_win_num 0

###
### Set up initial windows
###

wm maxsize . [winfo screenwidth .] [winfo screenheight .]
wm title . Plume
wm iconname . Plume
set surfit(appname) [tk appname surfit] ;# for compatibility with exmh
# set surfit(appname) [tk appname webbrowser] ;# future setting

catch {console hide}

#
# Preferences for Text widgets
#

option add *Text.height 40 startup
option add *Text.width 80 startup
bind Text <End> {%W yview end}
bind Text <Home> {%W yview 0.0}
bind Text <Next> {%W yview scroll 1 page}
bind Text <Prior> {%W yview scroll -1 page}

set s ""

# Control window

set p [frame $s.ctl-panel -relief raised -borderwidth 2]
menubutton $p.file -text File -menu $p.file.m -underline 0
pack $p.file -side left -anchor nw
menu $p.file.m
$p.file.m add command -label "Quit" -command {destroy .} -underline 0

menubutton $p.edit -text "Edit" -underline 0 -menu $p.edit.m
menu $p.edit.m
$p.edit.m add command -label "Preferences..." -underline 0 -command {SurfIt_Preferences}

menubutton $p.cache -text "Cache" -underline 0 -menu $p.cache.m
menu $p.cache.m
$p.cache.m add command -label "Empty Memory Cache" -underline 0 -command {Cache_flush}
$p.cache.m add command -label "Flush Entire Cache" -underline 0 -command {Cache_expire_all}
menubutton $p.applets -text "Applets" -menu $p.applets.m
menu $p.applets.m

debug_puts [list browser window done]

# Debugging aid menu

if {[info exists argva(-debug)] && ([string match *.* $argva(-debug)] || $argva(-debug))} {
    menubutton $p.debug -text "Debug" -menu $p.debug.m
    menu $p.debug.m
    if {$tcl_platform(platform) != "unix"} {
	$p.debug.m add command -label "Show Console" -command "debug_console $p.debug.m show"
	$p.debug.m add separator
    }
    $p.debug.m add command -label "htmllib"	-command {debug_load_file htmllib.tcl}
    $p.debug.m add command -label "internals"	-command {debug_load_file internal.tcl}
    $p.debug.m add command -label "prot.tcl"	-command {debug_load_file prot.tcl}
    $p.debug.m add command -label "cache.tcl"	-command {debug_load_file cache.tcl}
    $p.debug.m add command -label "handlers.tcl"	-command {debug_load_file handlers.tcl}
    $p.debug.m add command -label "applets.tcl"	-command {debug_load_file applets.tcl}
    $p.debug.m add command -label "safe-tk.tcl"	-command {debug_load_file safe-tk.tcl}
    $p.debug.m add command -label "hypetool.tcl"	-command {debug_load_file hypetool.tcl}
    $p.debug.m add command -label "mailto.tcl"	-command {debug_load_file mailto.tcl}
    $p.debug.m add command -label "uri.tcl"	-command {debug_load_file uri.tcl}
    pack $p.debug -side left -anchor w

    proc debug_load_file {file} {
	auto_load_file $file
	PR_init; HMlibrary_init; interp_init; 
	Applet_init; handlers_init; HT_init; mailto_init; surfit_init
    }
    proc debug_console {m state} {
	switch $state {
	    show {
		catch {console show}
		$m entryconfigure "Show Console" -label "Hide Console" -command "debug_console $m hide"
	    }
	    hide {
		catch {console hide}
		$m entryconfigure "Hide Console" -label "Show Console" -command "debug_console $m show"
	    }
	}
    }
}
menubutton $p.about -text "About" -menu $p.about.m -underline 0
menu $p.about.m
$p.about.m add command -label Plume -command {surfit_create_window http://tcltk.anu.edu.au/} -underline 0
$p.about.m add command -label Usage -command tkAboutDialog -underline 0

pack $p.edit $p.cache $p.applets -side left -anchor w
pack $p.about -side right -anchor e

pack $p -side top -fill x

set p [frame $s.open-panel]

button $p.home -text "Home" -command {surfit_create_window $wwwhome}
if {$wwwhome == {}} "$p.home configure -state disabled"

button $p.openlab -text "Open New: " -command {
    surfit_create_window $open_vars(.)
}
entry $p.open -textvariable open_vars(.)  -relief sunken -border 2
bind $p.open <Key-Return> {
    surfit_create_window $open_vars(.)
}
bind $p.open <Control-u> {set open_vars(.) {}}
set open_vars(.) {}

pack $p.home $p.openlab -side left -anchor nw
pack $p.open -side left -anchor nw -fill both -expand yes
pack $p -side top -fill x -expand yes -anchor sw

if {[info exists gk_library]} {
    wm withdraw .
}

# Initial (main) hyperdoc window

debug_puts [list about to create main window]

update
set main [surfit_create_window]

debug_puts [list main window created]

# Want to position the main window just beneath the control panel
# Need to get the x, y and height of the frame surrounding the control panel
# For now, I'll fudge it for olwm
wm geometry $main =+[expr [winfo rootx .] - 5]+[expr [winfo rooty .] + 5 + [winfo height .]]

# resize the browser control panel to be at least as wide as the main window
update
set height [winfo height $p]
pack propagate $p 0
$p configure -width [max [winfo width .] [winfo width $main]] -height $height

# Load initial page (if any)

if {[info exists gk_library]} {
    if {![gk_amOriginator]} {
	set wwwhome {}
    }
    gk_bind updateEntrant {
	gk_toUserNum %U surfit_realLoadURL $surfit_currPage([SurfIt_hyperpage $main]) [SurfIt_hyperpage $main]
    }
}

# Optionally load home page
if {$wwwhome != {}} {
    incr surfit_url_history_idx([SurfIt_hyperpage $main])
    SurfIt_loadURL $wwwhome [SurfIt_hyperpage $main]
}} \
utils.tcl {# $Id: utils.tcl,v 1.7 1996/06/26 04:32:12 steve Exp $
#
#	Copyright (c) 1996,1997 Australian National University
#
#--------
#
#	PASTIME Project
#	Cooperative Research Centre for Advanced Computational Systems
#	COPYRIGHT NOTICE AND DISCLAIMER.
#
# Copyright (c) 1995,1996 ANU and CSIRO on behalf of the participants in 
# the  CRC for Advanced Computational Systems ('ACSys').
# 
# ACSys makes this software and all associated data and documentation
# ('Software') available free of charge for non-commercial purposes only. You
# may make copies of the Software but you must include all of this notice on
# any copy.
# 
# The Software was developed for research purposes and ACSys does not warrant
# that it is error free or fit for any purpose.  ACSys disclaims any
# liability for all claims, expenses, losses, damages and costs any user may
# incur as a result of using, copying or modifying the Software.
#
###
#
# Define some general, utility functions
#

package provide utilities [lindex {$Revision: 1.7 $} 1]

if {[info commands infox] == {}} {
    # TclX is missing.  Do what we can do replace
    # its functions with Tcl (slower) equivalents

    proc auto_load_file {f} {
	global auto_path
	set found 0
	foreach p $auto_path {
	    if {$p == "."} {
		set path $f
	    } else {
		set path $p/$f
	    }
	    if {[file readable $path]} {
		set found 1
		uplevel #0 source $path
		break
	    }
	}
	if {!$found} {
	    error "unable to source \"$f\", file not found"
	}
    }

    proc lassign {list args} {
	foreach l $list {
	    if {$args == {}} break;
	    upvar [lindex $args 0] var
	    set var $l
	    set args [lrange $args 1 end]
	}
    }

    proc min {a b} {
	if {$a < $b} {return $a} else {return $b}
    }

    proc max {a b} {
	if {$a > $b} {return $a} else {return $b}
    }
    
    proc read_file {f} {
	set fd [open $f]
	if {[catch {read -nonewline $fd} data]} {
	    close $fd
	    error $data
	}
	close $fd
	return $data
    }

    # A pseudo-random number generator
    # Based on code by Mark Eichin, Cygnus. 
    # http://www.cygnus.com/%7Eeichin/random-tcl.html

    global UTIL_rndseed
    set UTIL_rndseed [expr [clock clicks] % 65536]
 
    proc UTIL_rawrand {} {
	global UTIL_rndseed
	# per Knuth 3.6:
	# 65277 mod 8 = 5 (since 65536 is a power of 2)
	# c/m = .5-(1/6)\sqrt{3}
	# c = 0.21132*m = 13849, and should be odd.
	set UTIL_rndseed [expr (65277 * $UTIL_rndseed +13849)%65536]
	set UTIL_rndseed [expr ($UTIL_rndseed+65536)%65536]
	return $UTIL_rndseed
    }
    proc random {arg1 {seed {}}} {
	if {$arg1 == "seed"} {
	    global UTIL_rndseed
	    set UTIL_rndseed $seed
	    return {}
	}
	return [expr ([UTIL_rawrand] * $arg1) / 65536]
    }
}

### Resize a Text widget to fit it's contents.

# This is a huge kludge that uses the Text's scrollbars
# to achieve the fit.
#
# It has been tested on static text, ie. where the Text widget
# has had all of its text inserted before attempting to resize
# it using this facility.  If you use it to size a Text widget
# and then subsequently insert more text to the widget and it
# doesn't work properly then let me know (also let me know if
# it does work!) <URL:mailto:Steve.Ball@tcltk.anu.edu.au>
# This facility definitely _not_ shrink a Text widget when text
# is removed.

# USAGE:
#
# Say you have a Text widget .text that you want resized.
#
# Firstly, everytime you want to resize the Text widget call the
# reset procedure:
#	UTIL_resetFitText .text
#
# To have the widget resized to fit it's contents lengthwise,
# call:
#	UTIL_fitTextVertical .text <min>
# where <min> is the minimum height you want the widget to be.
#
# To have the widget resized to fit it's contents widthwise,
# call:
#	UTIL_fitTextHorizontal .text <min>
# where <min> is the minimum width  you want the widget to be.
#
# The latter two procedures will automatically save any scrolling
# commands already defined and invoke them whenever the widget
# calls its scroll commands.
# Also, these procedures take an optional command that is 
# executed once the correct size has been achieved.  This facility
# is useful when you want the text widget sized to its contents
# both horizontally and vertically.  The widget should be sized
# vertically first, and then horizontally.  The command to
# achieve this is:
#	UTIL_fitTextVertical .text <min>  #		{UTIL_fitTextHorizontal .text <min>}


proc UTIL_fitTextHorizontal {win min {fitcmd {}}} {
    debug_puts [list UTIL_fitTextHorizontal $win $min $fitcmd]
    upvar #0 UTIL$win var

    if {![info exists var(lastXLast)]} {
	array set var {widthCorrect 0 lastXLast -1 widthAdjusted 0}
	$win configure -xscrollcommand [list UTIL_fitText_xscroll $win $min $fitcmd  [$win cget -xscrollcommand]]
    } elseif {$var(lastXLast) != -1 && !$var(widthCorrect) && !$var(widthAdjusted)} {
	set var(widthAdjusted) 1
	$win configure -width [expr [$win cget -width] + 1]
    } 
}

proc UTIL_fitTextVertical {win min {fitcmd {}} {adjust 0}} {
    debug_puts [list UTIL_fitTextVertical $win $min $fitcmd $adjust]
    upvar #0 UTIL$win var

    # Make a first guess at the correct size
    $win configure -height [set start [max $min [lindex [split [$win index end] .] 0]]]
    update idletasks
    UTIL_fitTextVertical:resize $win $adjust
    if {[string compare {} $fitcmd]} {
	eval $fitcmd
    }
    # Use yscrollcommand to know when contents gets bigger
    # NB. no need to keep any existing script, since widget contents
    # will always be fully visible
    $win configure -yscrollcommand [list UTIL_fitTextVertical:scroll $win $fitcmd $adjust]
}

proc UTIL_fitTextVertical:resize {win adjust} {
    set height [set start [$win cget -height]]
    debug_puts [list UTIL_fitTextVertical:resize $win starting height $start]
    set cnt 10	;# prevent infinite loop
    while {![string compare {} [$win dlineinfo {end - 1 line}]] && [incr cnt -1]} {
	incr height $start
	$win configure -height $height
	update idletasks
    }
    if {!$cnt} {
	# Something funny going on...
	# Leave it for now, and come back via the yscrollcommand
	# later
	$win configure -height $start
    }
    # Adjust height for snug fit
    # Use Tk 8.0's font metrics if available, otherwise leave as is
    if {[catch {array set metrics [font metrics [$win cget -font]]}]} {
	return
    }
    foreach {x y w h b} [$win dlineinfo {end - 1 line}] break
    $win configure -height [expr int( ceil( double($y + $b) / double($metrics(-linespace)) ) )]
    # Just to make sure...
    if {$adjust && $start == [$win cget -height]} {
	# Haven't changed anything
	$win configure -height [incr start]
    }
}

proc UTIL_fitTextVertical:scroll {win fitcmd adjust first last} {
    global UTIL_vertical
    debug_puts [list UTIL_fitTextVertical:scroll $win fitcmd $adjust $first $last]
    switch -glob -- $first,$last {
	0,0 -
	0,1 {
	    # All done
	    catch {unset UTIL_vertical($win,last); unset UTIL_vertical($win,cnt)}
	}
	default {
	    # Catch infinite loops.  This often occurs when fitcmd also resizes
	    # the widget or its parent
	    if {[info exists UTIL_vertical($win,last)] && $UTIL_vertical($win,last) == $last} {
		if {[incr UTIL_vertical($win,cnt) -1] <= 0} {
		    return
		}
	    } else {
		array set UTIL_vertical [list $win,last $last $win,cnt 10]
	    }

	    # Resize the widget.  Disable yscrollcommand to avoid calling ourself
	    $win configure -yscrollcommand {}
	    UTIL_fitTextVertical:resize $win $adjust
	    $win configure -yscrollcommand [list UTIL_fitTextVertical:scroll $win $fitcmd $adjust]
	    catch {eval $fitcmd}
	}
    }
}

proc UTIL_resetFitText {win} {
    global UTIL_vertical
    debug_puts [list UTIL_resetFitText $win]
    upvar #0 UTIL$win var
    catch {unset var}
    set horizontal [$win cget -xscrollcommand]
    if {[lindex $horizontal 0] == "UTIL_fitText_xscroll"} {
	$win configure -xscrollcommand [lindex $horizontal end]
    }
    catch {unset UTIL_vertical($win,last); unset UTIL_vertical($win,cnt)}
}

# We need to dampen last values which are very small

proc UTIL_fitText_xscroll {win min fitcmd cmd first last} {
    debug_puts [list UTIL_fitText_xscroll $win $min $fitcmd $cmd $first $last]
    upvar #0 UTIL$win var

    if {$first == 0 && $last >= 0.99} {
	if {!$var(widthCorrect) && $fitcmd != {}} {
	    eval $fitcmd
	}
	array set var {widthCorrect 1 widthAdjusted 0}
    } else {
	if {$last == $var(lastXLast)} {
	    # Do nothing
	} elseif {$last <= 0.1} {
	    set var(widthCorrect) 0
	    $win configure -width [expr [$win cget -width] + 1]
	} else {
	    array set var {widthCorrect 0 widthAdjusted 0}
	    set winwidth [$win cget -width]
	    set reqwidth [winfo reqwidth $win]
	    set realwidth [winfo width $win]
	    # Work out the actual width (approximately)
	    if {$realwidth <= $reqwidth} {
		# Window's width has been honoured, so use it in calculating width
		set actualWidth $winwidth
	    } else {
		# Geometry manager is probably sizing the widget, so translate to chars
		catch {set actualWidth [expr int(ceil($winwidth * ($realwidth / $reqwidth)))]}
	    }
	    if {[catch {expr int(ceil($actualWidth / $last))} new]} {
		puts stderr "bad calculation: \"$new\""
	    }
	    catch {$win configure -width $new}
	    set var(adjustX) 10
	    if {![info exists var(checkXID)]} {
		set var(checkXID) [after 250 UTIL_fitText_checkWidth $win $min]
	    }
	}
    }
    set var(lastXLast) $last

    if {$cmd != {}} {eval $cmd $first $last}
}

proc UTIL_fitText_checkWidth {win min} {
    debug_puts [list UTIL_fitText_checkWidth $win $min]
    upvar #0 UTIL$win var

    if {$var(widthCorrect)} {
	catch {unset var(checkXID)}
    } elseif {[incr var(adjustX) -1] <= 0} {
	catch {unset var(checkXID)}
    } else {
	$win configure -width [max $min [expr [$win cget -width] + 1]]
	set var(checkXID) [after 250 UTIL_fitText_checkWidth $win $min]
    }
}

### Filesystem utilities

proc lsubst { listname index new } {
    upvar 1 $listname l
 
    set l [ lreplace $l $index $index $new ]
}

# When using a pathname to construct a URI we need to map
# special characters using x-www-url-encoded.

# File_stdname --
#
# This procedure maps from a native pathname to a x-www-url-encoded
# name
#
# Arguments:
# p - an absolute pathname (internal format) to map

proc File_stdname {p} {
    debug_puts [list File_stdname $p]

    set p [lrange $p 1 end]	;# Remove the root

    # Protect the path component separators used on various filesystems
    regsub -all / $p {%2F} p
    regsub -all : $p {%3A} p
    regsub -all {\\\\} $p {%5C} p

    # Map URI special characters
    set pl {}
    foreach c $p {
	regsub -all { } $c {%20} c
	regsub -all {\?} $c {%3F} c
	regsub -all {#} $c {%23} c
	lappend pl $c
    }

    # Construct a URI-compatible path.  URI's use '/' as the path
    # component separator

    debug_puts [list File_stdname $p -> /[join $pl /]]

    return /[join $pl /]
}

# File_nativename --
#
# This procedure takes a x-www-url-encoded pathname and returns
# a path in Tcl list representation.

proc File_nativename {p} {
    set pl [file split $p]

    # workaround windows "feature"  where first element may still have "/" in it 
    # e.g absolute & drive relative paths    
    #if { $tcl_platform(platform) == "windows" } {
    #    regsub {/\\\\} $pl {\\} pl
    #}

    # Decode special characters in each path component
    set path {}
    foreach c $pl {
	lappend path [URIcgiMap $c]
    }
    debug_puts [list File_nativename $p -> $path ]
    return $path
}

proc File_pwd {} {
    set dirname {}
    if {[catch {pwd} dirname]} {
	if {[catch {exec pwd} dirname]} {
	    catch {set dirname $env(PWD)}
	}
    }
    return [file split $dirname]
}

### From '96 Tcl Workshop tutorial examples
# steve: with some rewriting

# All of these procedures expect their pathname arguments to be in the
# Tcl internal representation, as produced by [file split]

if {$tcl_platform(platform) == "windows"} {
    if {[string tolower $tcl_platform(os)] == "windows nt"} {
	set OS(shell) cmd.exe
    } else {
	set OS(shell) command.com
    }
}

# File_makeAbsolute --
#
# This procedure converts a relative or absolute path into
# an absolute path name.
#
# Arguments:
# p -	The path name to convert

proc File_makeAbsolute {p} {
    eval file join [File_pwd] $file
}

# File_normalizeFile --
#
# This procedure takes a user entered file name and ensures
# that it will be interpreted properly.
#
# Arguments:
# file -	The file name to normalize

proc File_normalizeFile {file} {
    global tcl_platform

    switch $tcl_platform(platform) {
	macintosh {
	    if {[string first : $file] < 0} {
		return :$file
	    }
	}
	windows {
	    return [file join $file]
	}
    }

    return $file
}

# File_delete --
#
# This procedure hides platform-specific details
# of removing a file.
#
# Arguments:
# f - The path name of a single file to be removed

proc File_delete {f} {
    global tcl_platform OS

    debug_puts [list File_delete $f]

    set name [eval file join $f]
    if {[file isdir $name]} {
	error "path \"$f\" is a directory"
    }

    if {![catch {file delete $name}]} {
    	return
    }
    # Fall through to support for old Tcl
    switch $tcl_platform(platform) {
	unix {exec /bin/rm -f -- $name}
	windows {
	    regsub / $name {\\\\} name
	    exec $OS(shell) /c del \"$name\"
	}
	macintosh {rm $name}
	default {error "unknown platform \"$tcl_platform(platform)\""}
    }
}

# File_rmdir --
#
# This procedure hides platform-specific details
# of removing a directory.
#
# Arguments:
# p - The path name of a single directory to be removed

proc File_rmdir {p} {
    global tcl_platform OS

    debug_puts [list File_rmdir $p]

    set name [eval file join $p]
    if {![file isdir $name]} {
	error "path \"$p\" is not a directory"
    }

    if {![catch {file rmdir $name}]} {
    	return
    }
    # Fall through to support for old tcl

    switch $tcl_platform(platform) {
	unix {exec /bin/rmdir -- $name}
	windows {
	    regsub / $name {\\\\} name
	    exec $OS(shell) /c rd \"$name\"
	}
	macintosh {rmdir $name}
	default {error "unknown platform \"$tcl_platform(platform)\""}
    }
}

# File_mkdir --
#
# Create a directory (or folder).
#
# Arguments:
# p - the pathname of the directory to create

proc File_mkdir {p} {
    debug_puts [list File_mkdir $p]
    global tcl_platform OS
    set name [eval file join $p]
    if {![catch {file mkdir $name}]} {
    	return
    }
    # Fall through to support for old tcl

    switch $tcl_platform(platform) {
	unix {exec mkdir $name}
	windows {
	    regsub / $name "\\\\" name
	    debug_puts [list exec $OS(shell) /c mkdir \"$name\"]
	    exec $OS(shell) /c md \"$name\"
	}
	macintosh {mkdir $name}
    }
    debug_puts [list succeeded in creating $p]
}

#############################################
#
# Emulate fcopy for pre-Tcl8.0b1
#
#############################################

if {[info tclversion] < 8.0 || [regexp {8\.0a.*} [info patchlevel]]} {
    proc fcopy {in out args} {
	array set opts {-size {} -command {}}
	array set opts $args
	if {[string compare {} $opts(-command)]} {
	    fconfigure $in -blocking no
	    fconfigure $out -blocking no
	    fileevent $in readable [list UTIL_fcopy $in $out 0 $opts(-size) $opts(-command)]
	    return {}
	} else {
	    fconfigure $in -blocking yes
	    fconfigure $out -blocking yes
	    eval unsupported0 [list $in $out] $opts(-size)
	}
    }

    proc UTIL_fcopy {in out done size cmd} {
	if {[eof $in]} {
	    fileevent $in readable {}
	    eval $cmd $done
	} else {
	    set num [eval unsupported0 [list $in $out] $size]
	    fileevent $in readable [list UTIL_fcopy $in $out [expr $done + $num] $size $cmd]
	}
    }
}} \
fsbox.tcl #\ \$Id:\ fsbox.tcl,v\ 1.4\ 1996/08/06\ 11:54:24\ steve\ Exp\ \$\n#\n#\tCopyright\ (c)\ 1996,1997\ Australian\ National\ University\n#\n#--------\n#\n#\tPASTIME\ Project\n#\tCooperative\ Research\ Centre\ for\ Advanced\ Computational\ Systems\n#\tCOPYRIGHT\ NOTICE\ AND\ DISCLAIMER.\n#\n#\ Copyright\ (c)\ 1995,1996\ ANU\ and\ CSIRO\ on\ behalf\ of\ the\ participants\ in\ \n#\ the\ \ CRC\ for\ Advanced\ Computational\ Systems\ ('ACSys').\n#\ \n#\ ACSys\ makes\ this\ software\ and\ all\ associated\ data\ and\ documentation\n#\ ('Software')\ available\ free\ of\ charge\ for\ non-commercial\ purposes\ only.\ You\n#\ may\ make\ copies\ of\ the\ Software\ but\ you\ must\ include\ all\ of\ this\ notice\ on\n#\ any\ copy.\n#\ \n#\ The\ Software\ was\ developed\ for\ research\ purposes\ and\ ACSys\ does\ not\ warrant\n#\ that\ it\ is\ error\ free\ or\ fit\ for\ any\ purpose.\ \ ACSys\ disclaims\ any\n#\ liability\ for\ all\ claims,\ expenses,\ losses,\ damages\ and\ costs\ any\ user\ may\n#\ incur\ as\ a\ result\ of\ using,\ copying\ or\ modifying\ the\ Software.\n#\n#####\n#\n#\ File\ selection\ dialog,\ stolen\ from\ CMT\ 3.0\ distribution\n#\ Modified\ to\ allow\ specification\ of\ a\ file\ that\ does\ not\n#\ (yet)\ exist.\ \ Also\ cleaned\ up\ namespace\ usage\ -\ uses\n#\ prefix\ Fsbox\n\npackage\ provide\ fsbox\ \[lindex\ {\$Revision:\ 1.4\ \$}\ 1\]\n\nproc\ Fsbox_slistbox\ {path}\ {\n\ \ \ \ frame\ \$path\;\n\ \ \ \ set\ p\ \$path\;\n\n\ \ \ \ frame\ \$p.top\;\n\ \ \ \ frame\ \$p.bot\;\n\n\ \ \ \ listbox\ \$p.lbox\ -relief\ sunken\ -setgrid\ 1\ \ -yscrollcommand\ \"\$p.vert\ set\"\ -xscrollcommand\ \"\$p.horiz\ set\"\;\n\ \ \ \ scrollbar\ \$p.vert\ -orient\ vertical\ -relief\ sunken\ \ -command\ \"\$p.lbox\ yview\"\;\n\ \ \ \ scrollbar\ \$p.horiz\ -orient\ horizontal\ -relief\ sunken\ \ -command\ \"\$p.lbox\ xview\"\;\n\ \ \ \ frame\ \$p.nop\ -width\ 19\ -height\ 19\;\n\n\ \ \ \ pack\ \$p.vert\ \$p.lbox\ -in\ \$p.top\ -fill\ both\ -side\ right\;\n\ \ \ \ pack\ configure\ \$p.lbox\ -expand\ yes\;\n\ \ \ \ pack\ \$p.horiz\ \$p.nop\ -in\ \$p.bot\ -fill\ both\ -side\ left\;\n\ \ \ \ pack\ configure\ \$p.horiz\ -expand\ yes\;\n\n\ \ \ \ pack\ \$p.bot\ \$p.top\ -in\ \$p\ -fill\ both\ -side\ bottom\;\n\ \ \ \ pack\ configure\ \$p.top\ -expand\ yes\;\n\ \ \ \ return\ \$path\;\n}\n\n\nproc\ Fsbox\ {state}\ {\n\ \ \ \ upvar\ #0\ \$state\ var\n\n\ \ \ \ #\ defaults\n\ \ \ \ \n\ \ \ \ set\ fspath\ \$var(win)\;\n\ \ \ \ set\ var(currentDir)\ \[File_pwd\]\n\ \ \ \ if\ {\$var(currentDir)\ ==\ {}}\ {\n\terror\ \"cannot\ determine\ current\ directory\"\n\ \ \ \ }\n\n\ \ \ \ set\ var(filterList)\ \$var(flist)\;\n\n\ \ \ \ if\ {\"\$var(flist)\"\ ==\ {}}\ {\n\tset\ var(filterList)\ \[list\ \"All\ Files\\\\\\\\t\"\ *\]\;\n\ \ \ \ }\n\ \ \ \ \n\ \ \ \ set\ var(filter)\ \[lindex\ \$var(filterList)\ 0\]\;\n\ \ \ \ \n\ \ \ \ #\ toplevel\ frames\n\ \ \ \ \n\ \ \ \ frame\ \$fspath\;\n\ \ \ \ frame\ \$fspath.boxes\;\n\ \ \ \ frame\ \$fspath.options\;\n\ \ \ \ frame\ \$fspath.sep\ -borderwidth\ 1\ -height\ 2\ -relief\ sunken\;\n\ \ \ \ frame\ \$fspath.buttons\;\n\ \ \ \ \n\ \ \ #####\ Boxes\ frame\ #####\n\ \ \ \ \n\ \ \ \ set\ f\ \$fspath.boxes.files\;\n\ \ \ \ set\ d\ \$fspath.boxes.dirs\;\n\ \ \ \ \n\ \ \ \ frame\ \$f\;\n\ \ \ \ frame\ \$d\;\n\ \ \ \ \n\ \ \ \ label\ \$f.l\ -text\ Files:\ -anchor\ w\;\n\ \ \ \ entry\ \$f.e\ -relief\ sunken\ -width\ 1\;\n\ \ \ \ if\ {\[info\ exists\ var(initialSel)\]}\ {\n\t\$f.e\ insert\ 0\ \$var(initialSel)\n\ \ \ \ }\n\ \ \ \ Fsbox_slistbox\ \$f.list\;\n\ \ \ \ \n\ \ \ \ label\ \$d.l\ -text\ Directories:\ -anchor\ w\;\n\ \ \ \ label\ \$d.e\ -text\ \[eval\ file\ join\ \$var(currentDir)\]\ -width\ 1\ -anchor\ w\;\n\ \ \ \ Fsbox_slistbox\ \$d.list\;\n\ \ \ \ \n\ \ \ \ pack\ \$f.l\ \$f.e\ \$f.list\ -side\ top\ -pady\ 1\ -fill\ both\;\n\ \ \ \ pack\ configure\ \$f.list\ -expand\ yes\;\n\ \ \ \ \n\ \ \ \ pack\ \$d.l\ \$d.e\ \$d.list\ -side\ top\ -pady\ 1\ -fill\ both\;\n\ \ \ \ pack\ configure\ \$d.list\ -expand\ yes\;\n\ \ \ \ \n\ \ \ \ pack\ \$f\ \$d\ -in\ \$fspath.boxes\ -side\ left\ -expand\ yes\ -fill\ both\ -padx\ 10\;\n\ \ \ \ \n\ \ \ #####\ Options\ frame\ #####\n\ \ \ \ \n\ \ \ \ set\ o\ \$fspath.options\;\n\ \ \ \ \n\ \ \ \ menubutton\ \$o.filter\ -relief\ groove\ -pady\ 3\ -padx\ 3\ -width\ 1\ \\\n\t-menu\ \$o.filter.m\ -text\ \"\ \ \ \ \[Fsbox_desc\ \$state\]\"\ -anchor\ w\;\n\ \ \ \ menu\ \$o.filter.m\;\n\n#\ \ \ label\ \$o.hidden\ -pady\ 3\ -padx\ 3\ -width\ 1\;\n\ \ \ \n\ \ \ \ pack\ \$o.filter\ -expand\ yes\ -fill\ both\ -side\ left\ \ -in\ \$o\ -padx\ 10\;\n\ \ \n\ \ \ \ foreach\ fpair\ \$var(filterList)\ {\n\tset\ fdesc\ \[lindex\ \$fpair\ 0\]\;\n\tset\ fspec\ \[lindex\ \$fpair\ 1\]\;\n\t\$o.filter.m\ add\ radiobutton\ -label\ \"\$fdesc\ \ \ \$fspec\"\ \\\n\t\ \ \ \ -variable\ var(filter)\ -value\ \$fpair\ \\\n\t\ \ \ \ -command\ \[list\ Fsbox_changeFilter\ \$state\]\n\ \ \ }\n\ \ \ \ \n\ \ \ #####\ Buttons\ frame\ #####\n\ \ \ \ \n\ \ \ \ set\ b\ \$fspath.buttons\;\n\ \ \ \ \n\ \ \ \ frame\ \$b.ok\ -borderwidth\ 1\ -relief\ sunken\;\n\ \ \ \ button\ \$b.ok.b\ -text\ OK\ -pady\ 5\ -padx\ 8\ -command\ \[list\ FsboxOK\ \$state\]\n\ \ \ \ \n\ \ \ \ button\ \$b.cancel\ -text\ Cancel\ -pady\ 5\ -padx\ 8\ -command\ \[list\ FsboxExit\ \$state\]\n\ \ \ \ \n\ \ \ \ pack\ \$b.ok.b\ -in\ \$b.ok\ -pady\ 5\ -padx\ 5\;\n\ \ \ \ pack\ \$b.ok\ \$b.cancel\ -expand\ yes\ -in\ \$b\ -side\ left\;\n\ \ \ \ \n\ \ \ \ #\ Pack\ widgets\n\ \ \ \ \n\ \ \ \ pack\ \$b\ \$fspath.sep\ \$o\ \$fspath.boxes\ -side\ bottom\ -fill\ both\ \\\n\t-pady\ 5\ -padx\ 5\ -in\ \$fspath\;\n\ \ \ \ pack\ configure\ \$fspath.boxes\ -expand\ yes\;\n\ \ \ \ \n\ \ \ \ #\ Make\ neccesary\ bindings\n\ \ \ \ \n\ \ \ \ focus\ \$f.e\;\n\ \ \ \ \n\ \ \ \ bind\ \$f.e\ <Return>\ \[list\ Fsbox_keyEnter\ \$state\]\n\ \ \ \ bind\ \$f.e\ <Tab>\ \[list\ Fsbox_keyTab\ \$state\]\n\ \ \ \ \n\ \ \ \ bind\ \$f.list.lbox\ <1>\ \[list\ Fsbox_selFile\ \$state\ %W\ %x\ %y\]\n\n\ \ \ \ bind\ \$f.list.lbox\ <B1-Motion>\ \[list\ Fsbox_selFile\ \$state\ %W\ %x\ %y\]\n\t\n\ \ \ \ bind\ \$f.list.lbox\ <Double-1>\ \"FsboxOpenFile\ \$state\;\ break\"\n\n\ \ \ \ bind\ \$d.list.lbox\ <Double-1>\ \[list\ Fsbox_selDir\ \$state\ %W\ %x\ %y\]\n\ \ \ \ \n\ \ \ \ Fsbox_fillBoxes\ \$state\ \[Fsbox_spec\ \$state\]\;\n}\n\nproc\ Fsbox_changeFilter\ {state}\ {\n\ \ \ \ upvar\ #0\ \$state\ var\n\ \ \ \ \$var(win).options.filter\ configure\ -text\ \"\ \ \ \ \[Fsbox_desc\ \$state\]\"\;\n\ \ \ \ Fsbox_fillBoxes\ \$state\ \[Fsbox_spec\ \$state\]\;\ \n}\n\nproc\ Fsbox_selFile\ {state\ W\ x\ y}\ {\n\ \ \ \ upvar\ #0\ \$state\ var\n\ \ \ \ \$var(win).boxes.files.e\ delete\ 0\ end\;\n\ \ \ \ \$var(win).boxes.files.e\ insert\ 0\ \[\$W\ get\ \[\$W\ index\ @\$x,\$y\]\]\n}\n\nproc\ Fsbox_selDir\ {state\ W\ x\ y}\ {\n\ \ \ \ upvar\ #0\ \$state\ var\n\ \ \ \ Fsbox_changeDir\ \$state\ \[\$W\ get\ \[\$W\ index\ @\$x,\$y\]\]\;\n\ \ \ \ \$var(win).boxes.files.e\ delete\ 0\ end\;\n\ \ \ \ Fsbox_fillBoxes\ \$state\ \[Fsbox_spec\ \$state\]\;\n}\n\nproc\ FsboxOK\ {state}\ {\n\ \ \ \ upvar\ #0\ \$state\ var\n\n\ \ \ \ set\ fspath\ \$var(win)\n\ \ \ \ set\ var(selFile)\ \[\$fspath.boxes.files.e\ get\]\;\n\ \ \ \ cd\ \[eval\ file\ join\ \$var(currentDir)\]\n\ \ \ \ if\ {\$var(ignoreMissing)\ ||\ \[file\ isfile\ \$var(selFile)\]}\ {\n\tFsboxOpenFile\ \$state\n\ \ \ \ }\ else\ {\n\t\$fspath.boxes.files.e\ delete\ 0\ end\;\n\t\$fspath.boxes.files.e\ insert\ 0\ \"Filename\ Not\ Found\"\;\n\tupdate\;\n\tafter\ 500\n\t\$fspath.boxes.files.e\ delete\ 0\ end\;\n\t\$fspath.boxes.files.e\ insert\ 0\ \$var(selFile)\n\ \ \ \ }\n}\ \ \ \ \ \ \ \ \ \ \ \ \n\n#\ Takes\ a\ directory\ and\ a\ filter-spec,\ and\ returns\ a\ sorted\ list\ of\ the\n#\ files\ that\ match\ the\ filter-spec\ in\ that\ directory.\n\nproc\ Fsbox_GetFiles\ {dir\ fspec}\ {\n\ \ \ \ cd\ \[eval\ file\ join\ \$dir\]\;\n\ \ \ \ set\ fileList\ {}\;\n\ \ \ \ set\ fspec\ \[split\ \$fspec\ \\\;\\,\\:\\\ \]\;\n\ \ \ \ foreach\ filter\ \$fspec\ {\n\tset\ fileList\ \[concat\ \$fileList\ \[glob\ -nocomplain\ \$filter\]\]\;\n\ \ \ \ }\n\ \ \ \ set\ fileList\ \[lsort\ \$fileList\]\;\n\ \ \ \ return\ \$fileList\;\n}\n\n\nproc\ Fsbox_fillBoxes\ {state\ fspec}\ {\n\ \ \ \ upvar\ #0\ \$state\ var\n\n\ \ \ \ set\ dir\ \[eval\ file\ join\ \$var(currentDir)\]\n\ \ \ \ set\ fileList\ {}\;\n\ \ \ \ set\ dirList\ {}\;\n\ \ \ \ \n\ \ \ \ set\ files\ \[Fsbox_GetFiles\ \$dir\ \$fspec\]\;\n\ \ \ \ set\ dirs\ \[Fsbox_GetFiles\ \$dir\ *\]\;\n\n\ \ \ \ foreach\ f\ \$files\ {\n\tif\ {\[file\ isfile\ \$f\]\ &&\ \[file\ readable\ \$f\]}\ {\n\t\ \ \ \ lappend\ fileList\ \[file\ tail\ \$f\]\;\n\t}\n\ \ \ \ }\n\ \ \ \ \ \ \ \ \n\ \ \ \ foreach\ d\ \$dirs\ {\n\tif\ {\[file\ isdirectory\ \$d\]\ &&\ \[file\ readable\ \$d\]}\ {\n\t\ \ \ \ lappend\ dirList\ \[file\ tail\ \$d\]\;\n\t}\n\ \ \ \ }\n\ \ \ \ \n\ \ \ \ set\ dirList\ \[concat\ .\ ..\ \$dirList\]\;\n\ \ \ \ \n\ \ \ \ set\ flbox\ \$var(win).boxes.files.list.lbox\;\n\ \ \ \ set\ dlbox\ \$var(win).boxes.dirs.list.lbox\;\n\ \ \ \ \n\ \ \ \ \$flbox\ delete\ 0\ end\;\n\ \ \ \ \$dlbox\ delete\ 0\ end\;\n\ \ \ \ eval\ \$flbox\ insert\ 0\ \$fileList\;\n\ \ \ \ eval\ \$dlbox\ insert\ 0\ \$dirList\;\n}\n\nproc\ Fsbox_keyEnter\ {state}\ {\n\ \ \ \ upvar\ #0\ \$state\ var\n\ \ \ \ \n\ \ \ \ set\ entry\ \$var(win).boxes.files.e\;\n\ \ \ \ set\ file\ \[File_normalizeFile\ \[\$entry\ get\]\]\;\n\ \ \ \ \n\ \ \ \ #\ Check\ to\ see\ if\ we\ have\ a\ full\ pathname\ in\ \$file\n\ \ \ \ \n\ \ \ \ if\ {\[file\ dirname\ \$file\]\ !=\ \".\"}\ {\n\tset\ path\ \[file\ dirname\ \$file\]\;\n\tset\ fspec\ \[file\ tail\ \$file\]\;\n\ \ \ \ }\ else\ {\n\tset\ path\ \$var(currentDir)\n\tset\ fspec\ \$file\;\n\ \ \ \ }\n\ \ \ \ \n\ \ \ \ if\ {\"\$fspec\"\ ==\ \"\"}\ {\n\tset\ fspec\ .\;\n\ \ \ \ }\n\ \ \ \ \n\ \ \ \ set\ size\ \[llength\ \[Fsbox_GetFiles\ \$path\ \$fspec\]\]\;\n\n\ \ \ \ if\ {\[string\ first\ *\ \$fspec\]\ !=\ -1\ ||\n\t\[string\ first\ \\\;\ \$fspec\]\ !=\ -1\ ||\n\t\[string\ first\ ,\ \$fspec\]\ !=\ -1\ ||\n\t\[string\ first\ \\\ \ \$fspec\]\ !=\ -1}\ {\n\tset\ filtering\ 1\;\n\ \ \ \ }\ else\ {\n\tset\ filtering\ 0\;\n\ \ \ \ }\n\t\ \ \ \ \n\ \ \ \ \$entry\ delete\ 0\ end\;\n\n\ \ \ \ if\ {\$size\ ==\ 0\ &&\ !\$filtering}\ {\n\tif\ {\$var(ignoreMissing)}\ {\n\t\ \ \ \ set\ var(selFile)\ \$fspec\n\t\ \ \ \ FsboxOpenFile\ \$state\n\t\ \ \ \ return\;\n\t}\ else\ {\n\t\ \ \ \ \$entry\ insert\ 0\ \"Filename\ Not\ Found.\"\;\n\t\ \ \ \ update\;\n\t\ \ \ \ after\ 500\n\t\ \ \ \ \$entry\ delete\ 0\ end\;\n\t\ \ \ \ \$entry\ insert\ 0\ \$file\;\n\t\ \ \ \ Fsbox_fillBoxes\ \$state\ \[Fsbox_spec\ \$state\]\;\n\t\ \ \ \ return\;\n\t}\n\ \ \ \ }\ elseif\ {\$size\ >\ 1\ ||\ \$filtering}\ {\n\tFsbox_changeDir\ \$state\ \$path\;\n\tFsbox_newFilter\ \$state\ \$fspec\;\n\tFsbox_fillBoxes\ \$state\ \[Fsbox_spec\ \$state\]\;\n\treturn\;\n\ \ \ \ }\ else\ {\n\tif\ {\[file\ isdirectory\ \$fspec\]}\ {\n\t\ \ \ \ Fsbox_changeDir\ \$state\ \$fspec\;\n\t\ \ \ \ Fsbox_fillBoxes\ \$state\ \[Fsbox_spec\ \$state\]\;\n\t\ \ \ \ return\;\n\t}\ else\ {\n\t\ \ \ \ set\ var(selFile)\ \$fspec\n\t\ \ \ \ FsboxOpenFile\ \$state\n\t\ \ \ \ return\;\n\t}\n\ \ \ \ }\n}\n\nproc\ Fsbox_keyTab\ {state}\ {\n\ \ \ \ upvar\ #0\ \$state\ var\n\ \ \ \ \n\ \ \ \ set\ fentry\ \[\$var(win).boxes.files.e\ get\]\;\n\ \ \ \ \n\ \ \ \ #\ used\ in\ case\ we\ try\ to\ access\ ~user\ when\ user\ does\ not\ exist.\n\n\ \ \ \ set\ noUser\ \[catch\ {file\ dirname\ \$fentry}\ path\]\;\n\ \ \ \ if\ {\$noUser}\ {\n\treturn\;\n\ \ \ \ }\n\ \ \ \ \n\ \ \ \ if\ {\$path\ !=\ \".\"}\ {\n\tset\ path\ \[file\ dirname\ \$fentry\]\;\n\tset\ file\ \[file\ tail\ \$fentry\]\;\n\ \ \ \ }\ else\ {\n\tset\ path\ \$var(currentDir)\n\tset\ file\ \$fentry\;\n\ \ \ \ }\n\ \ \ \ \n\ \ \ \ #\ used\ in\ case\ they\ give\ a\ path\ that\ doesn't\ exist\n\n\ \ \ \ set\ noPath\ \[catch\ {cd\ \$path}\]\n\ \ \ \ if\ {\$noPath}\ {\n\treturn\;\n\ \ \ \ }\n\n\ \ \ \ set\ f\ \[glob\ -nocomplain\ \$file*\]\;\n\ \ \ \ set\ size\ \[llength\ \$f\]\;\n\ \ \ \ \n\ \ \ \ if\ {\$size\ ==\ 0}\ {\n\treturn\;\n\ \ \ \ }\ elseif\ {\$size\ ==\ 1}\ {\n\tset\ var(currentDir)\ \$path\n\t\$var(win).boxes.files.e\ delete\ 0\ end\;\n\tif\ {\[file\ dirname\ \$fentry\]\ ==\ \".\"}\ {\n\t\ \ \ \ \$var(win).boxes.files.e\ insert\ 0\ \$f\;\n\t}\ else\ {\n\t\ \ \ \ if\ {\[llength\ \$path\]\ ==\ 1}\ {\n\t\t\$var(win).boxes.files.e\ insert\ 0\ \[File_makeAbsolute\ \$f\]\;\n\t\ \ \ \ }\ else\ {\n\t\t\$var(win).boxes.files.e\ insert\ 0\ \[eval\ file\ join\ \[file\ split\ \$path\]\ \$f\]\;\n\t\ \ \ \ }\t\t\n\t}\n\tFsbox_fillBoxes\ \$state\ \$file*\;\n\treturn\;\n\ \ \ \ }\ else\ {\n\tset\ var(currentDir)\ \[file\ split\ \$path\]\n\tFsbox_fillBoxes\ \$state\ \$file*\;\n\treturn\;\n\ \ \ \ }\n}\n\nproc\ Fsbox_newFilter\ {state\ f}\ {\n\ \ \ \ upvar\ #0\ \$state\ var\n\ \ \ \ set\ var(filter)\ \[list\ \"User\ Defined\"\ \$f\]\;\n\ \ \ \ \$var(win).options.filter\ configure\ -text\ \"\ \ \ \ \[Fsbox_desc\ \$state\]\"\;\n}\n\nproc\ Fsbox_spec\ {state}\ {\n\ \ \ \ upvar\ #0\ \$state\ var\n\ \ \ \ return\ \[lindex\ \$var(filter)\ 1\]\;\n}\n\nproc\ Fsbox_desc\ {state}\ {\n\ \ \ \ upvar\ #0\ \$state\ var\n\ \ \ \ return\ \"\[lindex\ \$var(filter)\ 0\]\ \ \ \[lindex\ \$var(filter)\ 1\]\"\;\n}\n\nproc\ Fsbox_changeDir\ {state\ newDir}\ {\n\ \ \ \ upvar\ #0\ \$state\ var\n\n\ \ \ \ cd\ \$newDir\;\n\ \ \ \ set\ var(currentDir)\ \[File_pwd\]\;\n\ \ \ \ \$var(win).boxes.dirs.e\ conf\ -text\ \[eval\ file\ join\ \$var(currentDir)\]\n}\n\nproc\ FsboxExit\ {state}\ {\n\ \ \ \ upvar\ #0\ \$state\ var\n\ \ \ \ eval\ \$var(exit)\ \$state\n}\n\nproc\ FsboxOpenFile\ {state}\ {\n\ \ \ \ upvar\ #0\ \$state\ var\n\ \ \ \ eval\ \$var(open)\ \$state\n} \
uri.tcl {# $Id: uri.tcl,v 1.1 1996/08/06 11:54:38 steve Exp $
#
#	Copyright (c) 1996,1997 Australian National University
#
#--------
#
#	PASTIME Project
#	Cooperative Research Centre for Advanced Computational Systems
#	COPYRIGHT NOTICE AND DISCLAIMER.
#
# Copyright (c) 1995,1996 ANU and CSIRO on behalf of the participants in 
# the  CRC for Advanced Computational Systems ('ACSys').
# 
# ACSys makes this software and all associated data and documentation
# ('Software') available free of charge for non-commercial purposes only. You
# may make copies of the Software but you must include all of this notice on
# any copy.
# 
# The Software was developed for research purposes and ACSys does not warrant
# that it is error free or fit for any purpose.  ACSys disclaims any
# liability for all claims, expenses, losses, damages and costs any user may
# incur as a result of using, copying or modifying the Software.
#
###
# uri.tcl
#
# Routines for handling URIs
#
# All procedures are prefixed by "URI", "URL", "URN" or "URC"
# ("URL", "URN" and "URC" are currently unused)

package provide uri [lindex {$Revision: 1.1 $} 1]

# Parse a URI, assigning values to appropriate variables, and
# return the URL scheme

proc URI_parse {uri schemevar pathvar queryvar fragvar} {
    upvar $schemevar scheme $pathvar path $queryvar query $fragvar fragment

    set path {}
    set query {}
    set fragment {}

    # Extract out the components
    if {![regexp {^([^:]+):([^?#]*)(\?([^#]*))?(#(.*))?} $uri all scheme path qquery query ffrag fragment]} {
	error "URI \"$uri\" does not specify a scheme"
    }
    return $scheme
}

# Construct URI's, etc, from the pieces

# URI_make --
#
# Reconstructs an URL, without any fragment identifier.

proc URI_make {scheme path query} {
    if {$query != {}} {set query ?$query}
    return $scheme:$path$query
}

# Reconstitute the path
proc URI_makePath {path {query {}}} {
    if {[set what $path] == {}} {set what /}
    if {$query != {}} {
	append what "?" $query
    }
    return $what
}

# URI_merge --
#
# Takes a URI and another URI specifying a possibly relative 
# reference and returns an absolute URI for the referred URI.
# The final path is canonicalised using the supplied routine.
#
# It is assumed that "/" is used as the path component separator.

proc URI_merge {base rel pathhandler} {
    if {[catch {URI_parse $rel scheme path query fragment}]} {
	# rel is relative
	URI_parse $base scheme path query fragment
	if {[string match /* $rel]} {
	    # absolute path, preserve query and fragment
	    regexp {(/[^#?]*)} $rel all newpath
	    regexp {#(.*)} $rel all fragment
	    regexp {\?(.*)} $rel all query
	    # Handle special case for hostname specs
	    if {[string match //* $newpath]} {
		set path $newpath	;# will work because Tcl handles Unix-style paths
	    } else {
		# No hostname for scheme "file://"
		# Patch by Alain Seite, mailto:alain.seite@filnet.fr
		if {[string compare $scheme file]} {
		    regexp {^(//[^/]*)?(/.*)} $path all hostname basepath
		    set path $hostname$newpath
		} else {
		    set path /$newpath
		}
	    }
	} elseif {[string match #* $rel]} {
	    # Special case: relative destination anchor
	    set fragment $rel
	} elseif {[string match {\?*} $rel]} {
	    # Special case: relative search query
	    set query $rel
	} else {
	    regexp {([^#?]+)(\?([^#]*))?(#(.*))?} $rel all relpath optquery query optfrag fragment
	    if {[string match */ $path]} {
		set path $path$relpath
	    } else {
		set path [$pathhandler $scheme mergeURIrel [$pathhandler $scheme removetail $path] $relpath]
	    }
	}
    }

    if {$query != {}} {set query ?$query}
    if {$fragment != {}} {set fragment #$fragment}
    return $scheme:[$pathhandler $scheme canonicalise $path]$query$fragment
}

# URI_type --
#
# Returns whether a given URI is absolute or relative

proc URI_type {a} {
    if {[regexp {^[^:/?#]+:} $a]} {return abs} else {return rel}
}

# URI_dirname --
#
# Generic brand find-the-directory-part of a path
# Assumes "/" is the path component separator

proc URI_dirname {path} {
    if {$path == "/" || $path == ""} {return /}
    set length [string length $path]
    if {[string index $path [incr length -1]] == "/"} {
	set path [string range $path 0 [incr length -1]]
    }
    if {[set match [string last / $path]] == -1} {
	return ./
    } elseif {$match == 0} {
	return "/"
    } else {
	return [string range $path 0 $match]
    }
}

# URI_encode_comp --
#
# Map special characters :#/?\ and spaces in a single component of a pathname

proc URI_encode_comp {p} {
    regsub -all : $p {%3A} p
    regsub -all / $p {%2F} p
    regsub -all {\\\\} $p {%5C} p
    regsub -all { } $p {+} p
    regsub -all {\?} $p {%3F} p
    regsub -all {#} $p {%23} p
    return $p
}

# URIcgiMap --
#
# Just a copy of HMcgiMap from html_library, included here to remove
# dependency on html_library

proc URIcgiMap {data} {
	regsub -all {\+} $data " " data
	
	if {[regexp % $data]} {
		regsub -all {([][$\\])} $data {\\\1} data
		regsub -all {%([0-9a-fA-F][0-9a-fA-F])} $data  {[format %c 0x\1]} data
		return [subst $data]
	} else {
		return $data
	}
}
} \
prot.tcl {# $Id: prot.tcl,v 1.9 1996/08/06 11:54:31 steve Exp $
#
#	Copyright (c) 1996,1997 Australian National University
#
#--------
#
#	PASTIME Project
#	Cooperative Research Centre for Advanced Computational Systems
#	COPYRIGHT NOTICE AND DISCLAIMER.
#
# Copyright (c) 1995,1996 ANU and CSIRO on behalf of the participants in 
# the  CRC for Advanced Computational Systems ('ACSys').
# 
# ACSys makes this software and all associated data and documentation
# ('Software') available free of charge for non-commercial purposes only. You
# may make copies of the Software but you must include all of this notice on
# any copy.
# 
# The Software was developed for research purposes and ACSys does not warrant
# that it is error free or fit for any purpose.  ACSys disclaims any
# liability for all claims, expenses, losses, damages and costs any user may
# incur as a result of using, copying or modifying the Software.
#
###
# prot.tcl handles document loading.
#
# All procedures are prefixed by PR.  Global variables use either PR or PS.

package provide protocols [lindex {$Revision: 1.9 $} 1]

proc PR_init {} {}

# PRloadDocument handles loading the data and invoking the content-type handler
# If type is specified then the data is forced through the handler for that
# type.
#
# method specifies the type of transaction for obtaining the data.
# Valid values are "get" or "put".  These translate directly to HTTP methods.
# Other protocols may interpret the method as appropriate.
#
# win is a unique token for loading this document (typically the widget
# pathname that the document is to be loaded into).  From this token
# various state variables are generated.  By convention if win is not
# a pathname then array PS$win win will contain the window into which the
# document is being loaded.
#
# put_data is data to be sent to the URL as part of the request.
#
# PRloadDocument returns 1 if a document was successfully loaded, 0 otherwise.
# NB. Errors are propagated normally, so a return value of 0 indicates that
# no document was found to be loaded, rather than an error occurring in loading
# the document.

proc PRloadDocument {url win renderState {startproc PRnoop} {type {}} {method get} {put_data {}}} {
    upvar #0 $renderState ren

    # First, get the data
    if {[set result [PRloadData $url $win data dataState $method $put_data]] == {}} {return 0};
    upvar #0 $dataState var

    # Mutual pointers
    set var(renderState) $renderState
    set ren(dataState) $dataState

    # Other initialisation
    array set var  [list startproc $startproc forceType $type]

    if {$type == {}} {set var(type) $var(HDRcontent-type)} else {set var(type) $type}
    if {$var(type) != {}} {
	PRfind_handler $dataState $var(type)
	switch $result {
	    PRfile -
	    PRfd {
		if {[info exists var(read_handler)]} {
		    fileevent $var(fd) r [list PRstarthandler $dataState $result]
		}
		if {[info exists var(write_handler)]} {
		    fileevent $var(fd) w [list $var(write_handler) $dataState]
		}
	    }
	    PRpost {
		catch {eval $startproc [list $dataState $renderState]}
		eval $var(renderer) [list $dataState $renderState $data]
		PRhandlerFinished $dataState 0	;# Even though there was no handler
		catch {eval $var(closeproc) [list $dataState $renderState]}
	    }
	    PRdata {
		catch {eval $startproc [list $dataState $renderState]}
		PRprocess $dataState $renderState PRdata $data
		catch {eval $var(closeproc) [list $dataState $renderState]}
	    }
	}
    } elseif {$result == "PRfd"} {
	# Handling delayed
	if {[info exists var(read_handler)]} {
	    fileevent $var(fd) r [list PRdelayed_handler $dataState]
	}
	if {[info exists var(write_handler)]} {
	    fileevent $var(fd) w [list $var(write_handler) $dataState]
	}
    } else {
	error "unable to determine media type for \"$url\""
    }

    return 1
}

# Manage calling the read handler and processing the incoming data

proc PRstarthandler {dataState dataType args} {
    upvar #0 $dataState var
    debug_puts [list PRstarthandler $dataState $dataType $args]

    # fileevent and other processing must occur before startproc, 
    # since it may invoke update
    if {[info exists var(base_handler)]} {
	set var(base_handler) [list PRhandler $dataState $dataType]
    } else {
	fileevent $var(fd) r [list PRhandler $dataState $dataType]
    }
    # Start caching the document
    if {[PRtextData $var(HDRcontent-type)]} {
	Cache_add_document PRdata $dataState
    } else {
	Cache_add_document $dataType $dataState
    }

    catch {
	debug_puts [list PRstarthandler: invoking startproc $var(startproc)]
	eval $var(startproc) [list $dataState $var(renderState)]
    }
}

# PRprocess will call PRhandlerFinished for us

proc PRhandler {dataState dataType args} {
    upvar #0 $dataState var
    debug_puts [list PRhandler $dataState $dataType $args]

    if {[catch {eval $var(read_handler) $dataState $args} data]} {
	if {[PRhandlerError $dataState $data]} return;
	set data {}	;# error message no longer needed
    }
    Cache_append_data $dataState $data
    PRprocess $dataState $var(renderState) $dataType $data
    if {![info exists var(fd)] || [eof $var(fd)]} {
	catch {eval $var(closeproc) [list $dataState $var(renderState)]}
    }
}

# Manage calling the read handler & finding the media type handler

proc PRdelayed_handler {dataState args} {
    upvar #0 $dataState var
    debug_puts [list PRdelayed_handler $dataState $args]
  
    if {[catch {eval $var(read_handler) $dataState $args} data]} {
	if {[PRhandlerError $dataState $data]} return;
	set data {}	;# no longer need error message
    }
    if {$var(HDRcontent-type) != {}} {
	set var(type) $var(HDRcontent-type)
	PRfind_handler $dataState $var(type)
	if {[info exists var(fd)]} {
	    set var(putback_data) $data
	    if {[info exists var(base_handler)]} {
		set var(base_handler) [list PRstarthandler $dataState PRfd]
	    } else {
		fileevent $var(fd) r [list PRstarthandler $dataState PRfd]
	    }
	} else {
	    # All data has been read from the channel,
	    # now handle startproc, etc, directly
	    catch {eval $var(startproc) [list $dataState $var(renderState)]}
	    PRprocess $dataState $var(renderState) $var(dataType) $data
	}
    } elseif {$data != {}} {
	error "unable to determine media type for \"$var(url)\""
    }
}

# A handler for fcopy events.  fcopy is incompatible with fileevent, so
# manage the invocation of fileevent handler from here.

proc PRfcopy_handler {dataState args} {
    upvar #0 $dataState var
    debug_puts [list PRfcopy_handler $dataState $args]

    eval $var(base_handler) $args

    if {[info exists var(fd)]} {
	fcopy $var(fd) $var(cachefd) -size $var(chunkSize) -command [list PRfcopy_handler $dataState]
    } else {
	debug_puts [list haven't restarted fcopy]
    }
}

# Manage the handling and rendering process

proc PRprocess {dataState renderState dataType data} {
    upvar #0 $dataState var
    debug_puts [list PRprocess $dataState $renderState $dataType ...]

    if {[catch {eval $var(processor) [list $dataState $dataType $data]} result]} {
	if {[PRhandlerError $dataState $result]} return;
    } else {
	PRhandlerFinished $dataState $var(readsize)
    }

    eval $var(renderer) [list $dataState $renderState $result]
}

# PRfind_handler finds a content-type handler and renderer for the data
# and fills in the document's state variable with the details.

proc PRfind_handler {dataState type} {
    global PRtypes
    upvar #0 $dataState var

    if {[set handler [array get PRtypes $type]] == {}} {
	# Look for a more general handler
	if {[set handler [array get PRtypes  "[lindex [split $type /] 0]/*"]] == {}} {
	    # Look for the most general (catch-all) handler
	    set handler [array get PRtypes "*/*"]
	}
    }
    if {$handler == {}} {error "no handler found for content-type \"$type\""}

    # Find the bits and pieces
    lassign [lindex $handler 1]  var(processor) var(renderer) var(closeproc) var(typeHandlerQuality)
    if {$var(processor) == {}} {set var(processor) PRnoop}
    if {$var(renderer) == {}} {set var(renderer) PRnoop}
    if {$var(closeproc) == {}} {set var(closeproc) PRnoop}
    if {$var(typeHandlerQuality) == {}} {set var(typeHandlerQuality) 0}
}

proc PRnoop {args} {}

# A trivial data processor which just passes data unchanged
# Applications or content-handlers may find it useful

proc PRpassthrough {dataState dataType data} {
    return $data
}

# A slightly more complex data processor which passes the data
# through, as above, but also marks the data as post-processed

proc PRpassNpost {dataState dataType data} {
    # Not yet implemented
    return [PRpassthrough $dataState $dataType $data]
}

# Application callbacks for data loading, should be replaced by application

proc PRhandlerFinished {dataState size} {
    puts "==> handler for $dataState loaded $size bytes"
}

# If 1 is returned then further processing of the document
# is abandoned, otherwise processing continues.

proc PRhandlerError {dataState result} {
    error $result
}

# Simple test to determine whether data is plain text or binary
# given a document's content type.

proc PRtextData {type} {
    if {$type == "application/x-tcl" ||  $type == "application/x-cmt-script" ||  $type == "movie/cmt" ||  [string match text/* $type]}  {return 1}  else {return 0}
}

# Currently supported protocols include file:, http:, ftp:, about:
#
# The "PRloadData" command has arguments {url cookie data dataState method put_data}.
# "url" is the URL to be loaded, "cookie" is an application token (sometimes the
# pathname of a widget), "data" is the name of a variable to pass back the data,
# "dataState" is the name of a variable for the document's state variable name,
# "method" is either "get or "put" and indicates whether the URL is being read from or
# written to (NB. that HTTP's GET method allows data to be written along with the request)
# and "put_data" is data to be written to the URL.
#
# The return value is one of:
#	PRdata	the data itself has been returned
#	PRpost	post-processed data was available and has been returned
#		(probably from the cache)
#	PRfd	the return value is a file descriptor from which the data may be read.  
#		The socket is set to non-blocking mode.  Handlers are provided for managing 
#		reading data (see below).
#	PRfile	the return value is the filename of a file from which the data may be read
#
# The state variable also has elements set for the known MIME or HTTP attributes, if they
# are known at the time.  These values may be added at a later stage by the read handler.

if {![info exists PRdocumentID]} {set PRdocumentID 0}	;# protect from re-source
proc PRloadData {url cookie data dataState {method get} {put_data {}}} {
    global PRscheme PRdocumentID
    upvar $data data_return
    upvar $dataState state_return

    if {!($method == "get" || $method == "put")} {
	error "invalid method \"$method\""
    }

    set newID [incr PRdocumentID]
    set state_return PR$newID
    upvar #0 $state_return var

    # Clean out any old values in the state array
    catch {unset var}

    # Parse the URI
    URI_parse $url var(scheme) var(path) var(query) var(fragment)

    # add requestURI to state array (nb: fragment not incl.)
    array set var [list url [URI_make $var(scheme) $var(path) $var(query)] eof 0 HDRcontent-type {} HDRcontent-length 0 HDRcontent-encoding {} readsize 0 id $newID cookie $cookie method $method chunkSize 4096]

    # Check whether the document has been cached
    if {[set cache_ret [Cache_get_document $state_return $data]] != {}} {
	set var(dataType) $cache_ret
	if {$cache_ret == "PRdata" || $cache_ret == "PRpost"} {
	    return $cache_ret
	} else {set scheme file}	;# Must be in disk cache
    } else {set scheme $var(scheme)}

    # The document was not cached, so call the scheme handler to fetch the data
    if {[array names PRscheme $scheme] != {}} {
	set handler_return [[lindex $PRscheme($scheme) 0] $state_return $method $data $put_data]
    } else {
	error "unknown scheme \"$scheme\""
    }

    if {[set var(dataType) $handler_return] == {}} return {};
    return $handler_return
}

# Registration procedures
#
# The registration system allows dynamic extension and restriction of capabilities.

# Handlers are registered with a (numeric) "quality".  The handler with 
# the highest quality is used.
# Suggested values:
#	99	inline handlers
#	60-80	hypertool handlers
#	40	external (helper) handlers
#	20	default internal handlers
# Current bug is that only 2-digit values can be used.

### Scheme handlers

# Declare a URI scheme, such as http: or ftp:

# "handler" is a routine that takes care of opening a connection to the referred document.
# (whatever that means for that particular scheme).

# "URIparser" is a routine that performs scheme-specific processing of URLs.
# These routines take arguments: {scheme method args}
# Valid methods are:
#	canonicalise _path_	-- resolve an URL's path component to its canonical form
#	addcomponent _path_ _component_
#				-- append the component to the given path
#	removetail		-- remove the trailing path component
#	mergeURIrel		-- append a relative path given in URI syntax

proc PRregister_scheme {name handler URIparser {quality 0}} {
    global PRscheme PRschemeQ
    set PRschemeQ($name,$quality) [list $handler $URIparser]
    if {[lindex [split [lindex [lsort -decreasing [array names PRschemeQ $name,*]] 0] ,] 1] <= $quality} {
	set PRscheme($name) [list $handler $quality $URIparser]
    }
}

proc PRremove_scheme {name handler} {
    global PRscheme PRschemeQ
    if {[set hdlrs [array get PRschemeQ $name,*]] == {}} {error "scheme \"$name\" not found"}
    if {[set idx [lsearch -exact $hdlrs $handler]] == -1}  {error "handler \"$handler\" not found for scheme \"$name\""}
    unset PRschemeQ([lindex $hdlrs [expr $idx - 1]])
    if {[lindex $PRscheme($name) 0] == $handler} {
	set new [lindex [lsort -decreasing [array names PRschemeQ $name,*]] 0]
	lassign [split $new ,] name quality
	set PRscheme($name) [list [lindex $PRschemeQ($new) 0] $quality [lindex $PRschemeQ($new) 1]]
    }
}

proc PRregister_encoding {name suffixList handler {quality 0} {icon {}}} {
    global PRencodings PRencodingsQ
    set PRencodingsQ($name,$quality) $handler
    PRregister_suffixes $suffixList $name $icon
    if {[lindex [split [lindex [lsort -decreasing [array names PRencodingsQ $name,*]] 0] ,] 1] <= $quality} {
	set PRencodings($name) [list $handler $quality]
    }
}

proc PRregister_type {name suffixList handler renderer  {close_handler {}} {quality 0} {icon {}}} {
    global PRtypes PRtypesQ
    set PRtypesQ($name,$quality) [list $handler $renderer $close_handler]
    PRregister_suffixes $suffixList $name $icon
    if {[lindex [split [lindex [lsort -decreasing [array names PRtypesQ $name,*]] 0] ,] 1] <= $quality} {
	set PRtypes($name) [list $handler $renderer $close_handler $quality]
    }
}

proc PRregister_suffixes {suffixList type {icon {}}} {
    global PRsuffixes PRsuff_icon
    foreach suffix $suffixList {
	set suffix [string tolower $suffix]
	set PRsuffixes($suffix) $type
	if {$icon != {}} {
	    set PRsuff_icon($suffix) $icon
	}
    }
}

### Support routines

# PRpath_handler --
#
# Takes the path component of a URI and calls the scheme-specific method.
# See above for valid methods

proc PRpath_handler {scheme method args} {
    global PRscheme

    if {![info exists PRscheme($scheme)]} {error "scheme \"$scheme\" does not exist"}

    eval [lindex $PRscheme($scheme) 2] $scheme $method $args
}

### Private routines

# Encoding routines

# Try and work out the encoding from the given path
proc PRfind_encoding_from_path {path} {
    global PRencodings
    if {[set e [array get PRencodings [string tolower [file extension $path]]]] != {}} {
	return [lindex $e 1]
    } else {return {}}
}

# MIME type routines

# Try and work out the MIME-type from the given path
proc PRfind_type_from_path {path} {
    global PRsuffixes
    if {[set t [array get PRsuffixes [string tolower [file extension $path]]]] != {}} {
	return [lindex $t 1]
    } else {return {}}
}

# Failing simple extension tests, could attempt to read the magic number
# and derive file type.

proc PRfind_type_from_magic {path} {
    return {}
}

# Common URI parsers

# PRparser_generic --
#
# A simple URI parser routine which performs identity canonicalisation

proc PRparser_generic {scheme method args} {
    switch $method {
	removetail -
	canonicalise {
	    return [lindex $args 0]
	}
	mergeURIrel -
	addcomponent {
	    return [lindex $args 0][lindex $args 1]
	}
	default {
	    error "don't know how to \"$method\" a $scheme URL"
	}
    }
}

# PRparser_collapse --
#
# A URI parser which canonicalises paths by collapsing .. & .
# Suitable for http: and ftp: URLs.
# 
# Canonicalise a pathname:  Resolve ".", ".." and empty
# path components.  Eg. /a/b/../../c/./d//e => /c/d/e
# The result will be an absolute path.

proc PRparser_collapse {scheme method args} {
    switch $method {
	canonicalise {
	    debug_puts [list PRparse_collapse canonicalise args $args]
	    set path [lindex $args 0]
	    debug_puts [list path $path]
	    regexp {^(//[^/]*)?(.*)} $path all addr path
	    set trailingslash [expr {[string match */ $path] ? "/" : {}}]
	    set new {}
	    if {$path != {}} {
		foreach c [split $path /] {
		    switch $c {
			{} -
			. {
			    # Do nothing
			}
			.. {
			    # Remove previous path component
			    # test handles pathological case  /..
			    if { [llength $new] } { 
				set new [lreplace $new end end]
			    }
			}
			default {
			    lappend new $c
			}
		    }
		}
	    }
	    debug_puts [list returning $addr/[join $new /]$trailingslash]
	    return $addr/[join $new /]$trailingslash
	}
	addcomponent {
	    debug_puts [list PRparser_collapse $method returning [lindex $args 0]/[lindex $args 1]]
	    return [lindex $args 0]/[lindex $args 1]
	}
	mergeURIrel {
	    debug_puts [list PRparse_collapse mergeURI args $args]
	    switch -regexp [lindex $args 0] {
		^.*/$ {
		    # Strip trailing slash
		    regexp {(.*)/$} [lindex $args 0] x path
		}
		^.*/[^/]*\.[^/]*$ {
		    # Looks kinda like a file spec - return dirname
		    regexp {(.*)/[^/]*} [lindex $args 0] x path
		}
		default {
		    # Looks like a directory
		    set path [lindex $args 0]
		}
	    }
	    debug_puts [list PRparser_collapse $method returning $path/[lindex $args 1]]
	    return $path/[lindex $args 1]
	}
	removetail {
	    debug_puts [list PRparser_collapse removetail [lindex $args 0]]
	    regexp {^(//[^/]*)(/.*)?$} [lindex $args 0] x netaddr path
	    if {![string compare {} $path]} {set path /}
	    debug_puts [list returning $netaddr[URI_dirname $path]]
	    return $netaddr[URI_dirname $path]
	}
	default {
	    error "don't know how to \"$method\" a $scheme URL"
	}
    }
}

###
### Handlers for various URL schemes.
###
### These take {state method data put_data} as arguments and return information as per PRloadData.
### Prerequisite is that all necessary information has been set in the state variable.
###
### NB. That since Tcl is not 8-bit clean, binary data can only be dealt
### with by writing it to a file as it arrives from the network.
###

# The about: URL scheme provides pre-canned data.  The data is contained
# in the PRAbout global array, which should be maintained by the 
# application, and is expected to be HTML text.

PRregister_scheme about PRscheme_about PRparser_generic 99

proc PRscheme_about {state method data put_data} {
    global PRAbout
    upvar #0 $state var
    upvar 2 $data data_return

    if {$method != "get"} {error "about: scheme does not support method $method"}

    if {[info exists PRAbout($var(path))]} {
	set data_return $PRAbout($var(path))
	array set var [list HDRcontent-type text/html  HDRcontent-length [string length $data_return]  readsize [string length $data_return]  eof 1]
	return PRdata
    } else {
	error "don't know anything about \"$var(path)\""
    }
}

# The file: URL scheme reads from a file in the local filesystem.
# The path component is a native path.
# The path may be a script, in which case invoke it as a CGI script (not implemented).

PRregister_scheme file PRscheme_file PRparser_file 99

proc PRscheme_file {state method data put_data} {
    upvar #0 $state var
    upvar 2 $data data_return

    if {[info exists var(file)]} {
	set filename $var(file)
    } else {
	if {$var(path) == {}} {
	    debug_puts [list getting pwd]
	    set var(path) [eval file join [File_pwd]]
	} else {
	    debug_puts [list given path, normalizing]
	    # Accept native path
	    set var(path) [File_normalizeFile $var(path)]
	}
	# Convert to x-www-url-encoded representation
	#set var(path) [File_stdname $filename]
	set var(url) [URI_make file $var(path) {}]
	set filename [file split $var(path)]
    }
    debug_puts [list filename is $filename]

    # Check whether the path is a directory
    if {![catch {file isdirectory [eval file join $filename]} dir] && $dir} {
	# Build a long listing of all of the files in the directory
	# Windows crashes if /.. is accessed
	if {[llength $filename] > 1 && ![catch {file stat [eval file join $filename ..] stat}]} {
	    set dir "[PRmapPerms $stat(mode)] $stat(nlink) $stat(uid) $stat(gid) $stat(size) [clock format $stat(mtime) -format {%b %d %Y}] ..\n"
	}

	if {![catch {glob -- [eval file join $filename *]} globList]} {
	    foreach f [lsort $globList] {
		if {[catch {file stat $f stat}]} {
		    # Insert non-linked entry
		} else {
		    switch $stat(type) {
			directory	{append dir d}
			characterSpecial -
			blockSpecial	{append dir c}
			fifo		{append dir f}
			link		{append dir l}
			socket		{append dir s}
			file -
			default		{append dir -}
		    }
		    append dir "[PRmapPerms $stat(mode)] $stat(nlink) $stat(uid) $stat(gid) $stat(size) [clock format $stat(mtime) -format {%b %d %Y }] [file tail $f]\n"
		}
	    }
	}

	set data_return [PRdirlist2html $state {} $dir]
	array set var "eof 1 HDRcontent-type text/html  HDRcontent-length [string length $data_return]"
	return PRdata
    }
    # otherwise the path is to a regular file

    # Determine data encoding
    if {$var(HDRcontent-encoding) == {}} {
	if {[set var(HDRcontent-encoding) [PRfind_encoding_from_path $filename]] != {}} {
	    # Strip encoding suffix from path
	}
    }

    # Determine data type
    if {$var(HDRcontent-type) == {}} {
	if {[set var(HDRcontent-type) [PRfind_type_from_path $filename]] == {}} {
	    # Take a guess
	    set var(HDRcontent-type) "text/plain"
	}
    }

    switch $method {
	get {
	    debug_puts [list opening [eval file join $filename]]
	    set var(fd) [open [eval file join $filename] r]	;# if file doesn't exist an error will occur here
	    fconfigure $var(fd) -blocking no

	    set var(read_handler) PRfile_read

	    # Reading from a file is the same as reading from a cache file
	    set data_return $filename
	    set var(file) $filename

	    # Determining the size is OK since the file must exist
	    set var(HDRcontent-length) [file size [eval file join $filename]]
	}
	put {
	    set var(fd) [open [eval file join $filename] w]	;# If there's a problem an error will occur here

	    set var(write_handler) PRfile_write
	    set var(put_data) $put_data
	}
    }

    return PRfile
}

# PRparser_file --
#
# Perform operations on OS pathnames

proc PRparser_file {scheme method args} {
    switch $method {
	canonicalise {
	    set path [lindex $args 0]
	    return $path
	}
	addcomponent {
	    return [eval file join [file split [lindex $args 0]] [list [lindex $args 1]]]
	}
	removetail {
	    if {[llength [set dir [lindex $args 0]]] <= 1} {return [eval file join $dir]}
	    return [eval file join [lrange $dir 0 [expr [llength $dir] - 2]]]
	}
	mergeURIrel {
	    set path [expr {[string match */ [lindex $args 0]] ? [file split [lindex $args 0]] : [file split [file dirname [lindex $args 0]]]}]
	    # decode x-www-url-encoded URI relative path
	    
	    return [eval file join $path [File_nativename [lindex $args 1]]]
	}
	default {
	    error "don't know how to \"$method\" a $scheme URL"
	}
    }
}

proc PRfile_read {state} {
    upvar #0 $state var
    debug_puts [list PRfile_read $state]

    set var(readsize) 0
    if {[PRtextData $var(HDRcontent-type)]} {
	debug_puts [list reading text]
	set data [read $var(fd)]
	if {[eof $var(fd)]} {
	    debug_puts [list detected EOF]
	    set var(eof) 1
	    close $var(fd)
	    unset var(fd)
	}
    } else {
	debug_puts [list binary data]
	# Don't attempt to read binary data
	set data {}
	set var(eof) 1
	close $var(fd)
	unset var(fd)
    }
    set var(readsize) [string length $data]

    catch {set data $var(putback_data)$data}
    debug_puts [list returning [string length $data] bytes of data]
    return $data
}

proc PRfile_write {state} {
    upvar #0 $state var
    puts -nonewline $var(fd) $var(put_data)
    unset var(put_data)
    close $var(fd)
    unset var(fd)
}

array set PRpermMap {0 --- 1 --x 2 -w- 3 -wx 4 r-- 5 r-x 6 rw- 7 rwx}
proc PRmapPerms {perm} {
    global PRpermMap
    set perm [format %o $perm]
    for {set i [expr [string length $perm] - 3]; set j 0} {$j < 3} {incr i; incr j} {
	append ret $PRpermMap([string index $perm $i])
    }
    return $ret
}

# The wais: protocol handler uses a proxy server set up as a WWW-WAIS gateway

PRregister_scheme wais PRscheme_wais PRparser_generic 20

proc PRscheme_wais {state method data put_data} {
    global surfit

    if {![info exists surfit(CFGwaisproxy)] || ![info exists surfit(CFGwaisproxy-port)] ||  $surfit(CFGwaisproxy) == {} || $surfit(CFGwaisproxy-port) == {}} {
	error "WAIS proxy server not specified"
    }
    return [PRscheme_http $state get $data $put_data [list $surfit(CFGwaisproxy) $surfit(CFGwaisproxy-port)]]
}

# The http: URL scheme connects to a HTTP server, sends a request 
# and then uses a read handler to process with the response.

PRregister_scheme http PRscheme_http PRparser_collapse 99

proc PRscheme_http {state method data put_data {proxy {}}} {
    debug_puts [list PRscheme_http $state ...]
    global surfit
    upvar #0 $state var
    upvar 2 $data data_return

    # Recover window from state to get alternate state variable
    # This needs to be cleaned up (remove application dependency)
    if {![catch {set win [SurfIt_hyperpage $var(cookie)]}]} {
	upvar #0 PS$win altState
    }

    # Setup the connection. The file event handlers do the real work

    # Default port for HTTP
    set port 80

    # Parse path
    if {[regexp {//([^/:]*)(:([^/]+))?(.*)} $var(path) all host optport port path]} {
	if {$host == {}} {set host localhost}	;# Supply reasonable default
	if {$port == {}} {set port 80}		;# HTTP default service
	if {$path == {}} {set path /}
    } else {
	# Recover from malformed URL
	if {![string match */* $var(path)]} {
	    # Assume only the hostname:port has been given
	    regexp {([^:]*)(:(.*))?} $var(path) all host optport port
	    if {$host == {}} {set host localhost}
	    if {$port == {}} {set port 80}
	    set path /
	} else {
	    error "no hostname given in URL \"$var(url)\""
	}
    }

    # Check if this host should not go via proxy
    set noproxy 0
    if [info exists surfit(CFGnoproxy)] {
	foreach m $surfit(CFGnoproxy) {
	    set noproxy [string match $m $host]
	}
    }

    if {$proxy != {}} {
	set host [lindex $proxy 0]
	set port [lindex $proxy 1]
	set service "proxy HTTP server"
	set requestURI $var(url)
	upvar 3 $data data_return ;# Ugly!
    } elseif {!$noproxy && [info exists surfit(CFGproxy)] && $surfit(CFGproxy) != {}} {
	set host $surfit(CFGproxy)
	catch {set port $surfit(CFGproxy-port)}
	set service "proxy HTTP server"
	set requestURI $var(url)
    } else {
	array set var [list host $host port $port path $path]
	set service "host"
	set requestURI [URI_makePath $var(path) $var(query)]
    }

    if {$port == {}} {set port 80}	;# final check

    debug_puts [list connecting to $service $host on port $port]
    if {[catch {socket -async $host $port} var(fd)]} {
	unset var(fd)
	error "cannot connect to $service $host on port $port"
    }
    array set var [list write_handler PRhttp_connect HDRcontent-type {} HDRcontent-encoding {} HDRcontent-length {} length 0 requestURI $requestURI referer {} put_data $put_data]
    if {[info exists altState(referer)]} {
	set var(referer) $altState(referer)
	unset altState(referer);	# No longer needed
    }
    set data_return $var(fd)
    debug_puts [list returning with connection fd $var(fd)]
    return PRfd
}

proc PRhttp_connect {state} {
    debug_puts [list PRhttp_connect $state]
    global surfit PRtypes PRencodings
    upvar #0 $state var

    fconfigure $var(fd) -blocking no -translation crlf

    # Send the HTTP request.
    # The Accept: header is generated by registered content-type handlers
    # ditto for Accept-Encoding:
    # Current bug: Accept: entries are unordered - should be ordered by quality
    # User: ???

    debug_puts [list sending request: [string toupper $var(method)] $var(requestURI) HTTP/1.0]
    puts $var(fd) "[string toupper $var(method)] $var(requestURI) HTTP/1.0
User-Agent: $surfit(version)
Accept: [join [array names PRtypes] ,]"
    if {[array names PRencodings] != {}} {
	puts $var(fd) "Accept-Encoding: [join [array names PRencodings] ,]"
    }
    if {$var(referer) != {}} {
	puts $var(fd) "Referer: $var(referer)"
    }
    puts $var(fd) ""
    puts $var(fd) $var(put_data)
    unset var(put_data)	;# No longer needed
    flush $var(fd)
    fconfigure $var(fd) -translation auto	;# server may not use standard line terminators

    fileevent $var(fd) w {}
    set var(read_handler) PRhttp_read_header
    if {[info exists var(processor)]} {
	fileevent $var(fd) r [list PRstarthandler $state PRfd]
    } else {
	fileevent $var(fd) r [list PRdelayed_handler $state]
    }
    unset var(write_handler)
    debug_puts [list sent request, waiting on response]
}

# File event handler for HTTP responses.
# Pre-requisite is that all necessary information is in the state array variable.
# If var(putback_data) is set then prepend that data to the returned data,
# without causing it to be appended to the cache (since it has been already).

proc PRhttp_read_header {state} {
    debug_puts [list PRhttp_read_header $state]
    upvar #0 $state var

    set var(readsize) 0

    if {[catch {gets $var(fd) hdr} retCnt]} {
	# error occurred
	PRhandleEOF $state
	error "error occurred on connection to host $var(host), port $var(port)"
    }
    if {$retCnt < 0} {
	if {[eof $var(fd)]} {debug_puts [list got EOF]; return [PRhandleEOF $state]}  else {debug_puts [list returning, probably blocked]; return {}}
    } elseif {$retCnt == 0} {
	debug_puts [list end of headers]

	# Tcl is NOT 8-bit clean, and sometimes servers lie about the
	# MIME type of documents.  Hence we are forced to write the data into a file.
	# Copy the body of the document into the cache file.

	fconfigure $var(fd) -translation lf
	fconfigure $var(cachefd) -translation lf
	set var(read_handler) PRhttp_read_body
	set var(base_handler) [fileevent $var(fd) readable]
	fileevent $var(fd) readable {}
	trace variable var(fd) uw PRdbg_var_fd
	fcopy $var(fd) $var(cachefd) -size $var(chunkSize) -command [list PRfcopy_handler $state]

	# Handle automatic redirection.

	if {[info exists var(HTTPstatus)]} {
	    if {$var(HTTPstatus) == 301 || $var(HTTPstatus) == 302} {
		if {[info exists var(HDRlocation)]} {
		    debug_puts [list automatic redirection to $var(HDRlocation)]
		    PRhandleEOF $state
		    array set var {processor PRnoop renderer PRnoop closeproc PRnoop}
		    set fragment $var(fragment)
		    if {$fragment != {}} {set fragment #$fragment}
		    PRloadDocument $var(HDRlocation)$fragment $var(cookie) $var(renderState) $var(startproc) $var(forceType)
		    return {}
		} else {
		    error [list $var(HTTPstatus) $var(HTTPreason)]
		}
	    } elseif {$var(HTTPstatus) >= 400} {
		# Document content is assumed to be text/html
		PRfind_handler $state text/html
		error [list $var(HTTPstatus) $var(HTTPreason)]
	    }
	}

	# Check whether mandatory attributes have been set.
	# Current attributes are: Content-Type Content-Length
	if {$var(HDRcontent-type) == {}} {
	    # We'll have to try and guess what the Content-Type is.
	    if {[set var(HDRcontent-type) [PRfind_type_from_path $var(path)]] == {}} {
		# Take a guess
		set var(HDRcontent-type) text/plain
	    }
	}
	# Content-Length is fixed at EOF

	return {}
    }

    debug_puts [list got header $hdr]
    if {![regexp -nocase "HTTP/(\[^ \]*) (\[^ \]+) (.*)\r?" $hdr all var(HTTPversion) var(HTTPstatus) var(HTTPreason)]} {
	if {[regexp "(\[^:\]+):(\[^\r\]*)\r?" $hdr all key value]} {
	    set var(HDR[string tolower $key]) [string trim $value]
	}
    }
    return {}
}

proc PRdbg_var_fd {name1 name2 op} {
    debug_puts [list name1 $name2 name2 $name2 op $op caller: [info level -1]]
}

proc PRhttp_read_body {state bytes} {
    upvar #0 $state var
    debug_puts [list PRhttp_read_body $state $bytes]

    set var(readsize) $bytes
    incr var(length) $bytes

    set data {}

    if {$bytes} {
	if {[PRtextData $var(type)]} {
	    # Read the just-received data back out of the cache file
	    seek $var(cachefd) -$bytes end
	    set data [read -nonewline $var(cachefd)]
	}
	catch {set data "$var(putback_data)$data"; unset var(putback_data)}
	debug_puts [list returning data]
	return $data
    }

    if {[eof $var(fd)]} {
	# Determine content-length if not specified in headers
	if {$var(HDRcontent-length) == {}} {
	    set var(HDRcontent-length) $var(length)
	    Cache_set_content_length $var(url) $var(HDRcontent-length)
	}
	PRhandleEOF $state
    }

    debug_puts [list PRhttp_read_body finished]
    return $data
}

# Clean up a file event handler

proc PRhandleEOF {state} {
    upvar #0 $state var
    debug_puts [list PRhandleEOF $state]
    catch {close $var(fd)}
    catch {unset var(fd)}
    Cache_close $state
    set var(eof) 1
    return {}
}

### FTP protocol handler

# Supports both active and passive mode

PRregister_scheme ftp PRscheme_ftp PRparser_collapse 99

proc PRscheme_ftp {state method data put_data {proxy {}}} {
    global surfit PRtypes PRencodings
    upvar #0 $state var
    upvar 2 $data data_return

    if {$proxy != {}} {
	set host [lindex $proxy 0]
	set port [lindex $proxy 1]
	set service "proxy FTP server"
	upvar 3 $data data_return
    } elseif {[info exists surfit(CFGftpproxy)]} {
	set host $surfit(CFGftpproxy)
	catch {set port $surfit(CFGftpproxy-port)}
	set service "proxy FTP server"
    } else {
	# ftp: also allows user/password to be specified - not yet implemented
	if {[regexp {//([^/:]*)(:([^/]+))?(.*)} $var(path) all host optport port path]} {
	    if {$host == {}} {set host localhost}	;# Supply reasonable default
	    if {$port == {}} {set port 21}		;# FTP default service
	    if {$path == {}} {set path /}
	    array set var [list host $host port $port path $path]
	} else {
	    # Recover from malformed URL
	    if {![string match */* $var(path)]} {
		# Assume only the hostname:port has been given
		regexp {([^:]*)(:(.*))?} $var(path) all host optport port
		if {$host == {}} {set host localhost}
		if {$port == {}} {set port 21}
		set path /
		array set var [list host $host port $port path /]
	    } else {
		error "no hostname given in URL \"$var(url)\""
	    }
	}
	set service "host"
	set requestURI [URI_makePath $var(path) $var(query)]
    }

    # Default port for FTP
    if {$port == ""} {set port 21}

    if {[catch {socket -async $host $port} var(fd)]} {
	unset var(fd)
	error "cannot connect to $service $host on port $port"
    }

    fconfigure $var(fd) -blocking no

    array set var "
	read_handler PRftp_readConn HDRcontent-type {}
	HDRcontent-length {} HRcontent-encoding {} length 0
	ftpState madeConn listPreamble {} textData {}
    "

    set data_return $var(fd)
    return PRfd
}

proc PRftp_readConn {state} {
    global surfit
    upvar #0 $state var

    set var(readsize) 0

    # Read until we see a reply code
    for {set gotCode "-"} {$gotCode != " "} {} {
	if {[catch {gets $var(fd) line} retCnt]} {
	    PRhandleEOF $state
	    error "error occurred on FTP control connection to host $var(host), port $var(port)"
	}
	if {$retCnt < 0} {
	    if {[eof $var(fd)]} {return [PRhandleEOF $state]}  else {return {}}
	} elseif {!$retCnt} {
	    return {}
	}
	regexp "(\[0-9\]+)( |-)?(.*)\r?" $line all replyCode gotCode descr
	if {$gotCode == "-"} {append var(listPreamble) \n$descr}
    }

    switch $var(ftpState) {
	madeConn {
	    if {$replyCode == "220"} {
		puts $var(fd) "USER anonymous"
		flush $var(fd)
		set var(ftpState) userSent
	    } elseif {$replyCode == "120"} {
		return {}	;# expected delay. simply wait for the 220 reply
	    } else {
		PRhandleEOF $state
		error "FTP connection returned \"$descr\""
	    }
	}
	userSent {
	    if {$replyCode == "331"} {
		puts $var(fd) "PASS $surfit(CFGusername)@"
		flush $var(fd)
		set var(ftpState) passSent
	    } elseif {$replyCode == "230"} {
		puts $var(fd) "PASV"
		flush $var(fd)
		set var(ftpState) pasvSent
	    } else {
		PRhandleEOF $state
		error "ftp connection error \"$descr\""
	    }
	}
	passSent {
	    if {$replyCode == "230" || $replyCode == "202"} {
		puts $var(fd) "PASV"
		flush $var(fd)
		set var(ftpState) pasvSent
	    } else {
		PRhandleEOF $state
		error "ftp connection error \"$descr\""
	    }
	}
	pasvSent {
	    if {$replyCode == "227"} {
		if {[regexp {([0-9]+),([0-9]+),([0-9]+),([0-9]+),([0-9]+),([0-9]+)} $descr all a1 a2 a3 a4 p1 p2]} {
		    if {[catch {socket $a1.$a2.$a3.$a4 [expr $p1 * 256 + $p2]} var(listConnFd)]} {
			error "active mode FTP not implemented"
		    } else {
			fconfigure $var(listConnFd) -blocking no
			fileevent $var(listConnFd) r [list PRftp_readList $state]
		    }
		} else {error "active mode FTP not implemented"}
	    } else {error "active mode FTP not implemented"}
	    puts $var(fd) "TYPE I"
	    flush $var(fd)
	    set var(ftpState) typeSent
	}
	typeSent {
	    if {$replyCode == "200"} {
		puts $var(fd) "LIST $var(path)"
		flush $var(fd)
		set var(ftpState) listSent
	    } else {
		PRhandleEOF $state
		error "ftp connection error \"$descr\""
	    }
	}
	listSent {
	    if {$replyCode == "125" || $replyCode == "150"} {
		set var(ftpState) listStarted
		# Suspend this thread until all of the directory listing has been received
		set var(connHandler) [fileevent $var(fd) r]
		fileevent $var(fd) r {}
	    } else {
		PRhandleEOF $state
		catch {close $var(listConnFd)}
		unset var(listConnFd)
		error "ftp connection error \"$descr\""
	    }
	}
	listStarted {
	    if {$replyCode == "226" || $replyCode == "250"} {
		puts $var(fd) "PASV"
		flush $var(fd)
		set var(ftpState) pasv2Sent
	    } else {
		PRhandleEOF $state
		error "ftp connection error \"$descr\""
	    }
	}
	pasv2Sent {
	    if {$replyCode == "227"} {
		if {[regexp {([0-9]+),([0-9]+),([0-9]+),([0-9]+),([0-9]+),([0-9]+)} $descr all a1 a2 a3 a4 p1 p2]} {
		    if {[catch {socket $a1.$a2.$a3.$a4 [expr $p1 * 256 + $p2]} var(dataConnFd)]} {
			error "active mode FTP not implemented"
		    } else {
			fconfigure $var(dataConnFd) -translation lf
			fconfigure $var(cachefd) -translation lf
			set var(read_handler) PRftp_readData
			set var(base_handler) [list PRdelayed_handler $state]
			fcopy $var(dataConnFd) $var(cachefd) -size $var(chunkSize) -command [list PRfcopy_handler $state]
		    }
		} else {error "active mode FTP not implemented"}
	    } else {error "active mode FTP not implemented"}
	    puts $var(fd) "RETR $var(path)"
	    flush $var(fd)
	    set var(ftpState) retrSent
	}
	retrSent {
	    if {$replyCode == "125" || $replyCode == "150"} {
		if {[set var(HDRcontent-type) [PRfind_type_from_path $var(path)]] == {}} {
		    # Take a guess
		    set var(HDRcontent-type) text/plain
		}
		# Start caching the document
		# High level handlers normally take care of this, but we need
		# to force the cache file to be created since the data reader is called next.
		Cache_add_document PRfd $state
		catch {set var(HDRcontent-length) [expr [lindex $var(ftpList) 4] + 0]}
		set var(ftpState) dataStarted
		# Suspend this thread until all of the data has been received
		set var(connHandler) [fileevent $var(fd) r]
		fileevent $var(fd) r {}
	    } elseif {$replyCode == "550"} {
		close $var(dataConnFd)
		# Create a HTML document to display the contents
		set var(HDRcontent-type) text/html
		# Start caching the document
		Cache_add_document PRdata $state
		if {$var(forceType) != {}} {
		    # return data as-is
		    set var(readsize) [string length $var(ftpList)]
		    set var(length) $var(readsize)
		    set var(HDRcontent-length) $var(readsize)
		    PRhandleEOF $state
		    return $var(ftpList)
		}
		# Otherwise translate into HTML
		set doc [PRdirlist2html $state $var(listPreamble) $var(ftpList)]
		set var(readsize) [string length $doc]
		set var(length) $var(readsize)
		set var(HDRcontent-length) $var(readsize)
		PRhandleEOF $state
		return $doc
	    } else {
		PRhandleEOF $state
		error "ftp control connection error \"$descr\""
	    }
	}
	dataStarted {
	    if {$replyCode == "226" || $replyCode == "250"} {
		puts $var(fd) "QUIT"
		flush $var(fd)
		set var(ftpState) quitSent
	    } else {
		PRhandleEOF $state
		error "ftp control connection error \"$descr\""
	    }
	}
	quitSent {
	    PRhandleEOF $state
	    return $var(textData)
	}
	dataConnError {
	    PRhandleEOF $state
	    error "ftp data connection error"
	}
	default {
	    PRhandleEOF $state
	    error "internal error: invalid state \"$var(ftpState)\""
	}
    }
    return {}
}

proc PRftp_readList {state} {
    upvar #0 $state var
    if {[catch "read -nonewline $var(listConnFd)" data]} {
	close $var(listConnFd)
	set var(ftpState) dataConnError
	# Restore control connection fileevent handler
	if {[info exists var(fd)]} {
	    fileevent $var(fd) r $var(connHandler)
	    unset var(connHandler)
	}
	return {}
    }

    append var(ftpList) $data

    if {[eof $var(listConnFd)]} {
	close $var(listConnFd)
	unset var(listConnFd)
	# Restore control connection fileevent handler
	if {[info exists var(fd)]} {
	    fileevent $var(fd) r $var(connHandler)
	    unset var(connHandler)
	}
    }
    return {}
}

proc PRftp_readData {state bytes} {
    upvar #0 $state var

    # If the data should be text read it back in.

    set var(readsize) $bytes
    incr var(length) $bytes

    if {[eof $var(dataConnFd)]} {
	close $var(dataConnFd); unset var(dataConnFd)
	# Determine content-length if not specified in list
	if {$var(HDRcontent-length) == {}} {
	    set var(HDRcontent-length) $var(length)
	    Cache_set_content_length $var(url) $var(HDRcontent-length)
	}
	# Restore control connection fileevent handler
	if {[info exists var(fd)]} {
	    fileevent $var(fd) r $var(connHandler)
	    unset var(connHandler)
	}
	# Read data into memory if text
	if {[PRtextData $var(type)]} {
	    seek $var(cachefd) 0 start
	    set var(textData) [read -nonewline $var(cachefd)]
	}
    }
    return {}
}


set PRlistSpecs1 { 
    size      4 4    3 r "%s bytes" month     5 5    1 r "%s" day       6 6    1 r "%s" timeYr    7 7    1 r "%s" } 
set PRlistSpecs8  { 
    filename  8 end  3 l %s
}
set PRlistSpecs [concat $PRlistSpecs1 $PRlistSpecs8]

set PRlistParam {name f l sp just fmt}

proc PRdirlist2html {state preamble dirList} {
    upvar #0 $state var
    global PRlistSpecs PRlistSpecs1  PRlistSpecs8 PRlistParam
    debug_puts [list PRdirlist2html $state ...]

    # directory path title for local filesystems 
    if { $var(scheme) == "file" } {
	set dirTitle [PRpath_handler $var(scheme) canonicalise $var(path)]
	if { $dirTitle == {}} {set dirTitle "/"}
    } else {
	set dirTitle $var(url)
    }

    set doc "<html>
<head>
<title>Listing of Directory $dirTitle</title>
</head>
<body>
<h1>Listing of Directory $dirTitle</h1>
<pre>"
    if {$preamble != {}} {
	append doc "
$preamble
"
    }

    set contents [split $dirList \n]
    set skipLines 0
    if {[lindex [lindex $contents 0] 0] == "total"} { incr skipLines }
    if {[lindex [lindex $contents $skipLines] 8] == "."}  { incr skipLines }
    if {[lindex [lindex $contents $skipLines] 8] == ".."}  {
	if { $var(path) == "/" } { 
	    ;
	} else {
	    set parentURL [URI_make $var(scheme) [PRpath_handler $var(scheme) canonicalise [PRpath_handler $var(scheme) addcomponent $var(path) ..]] {}]
	    append doc "<li dingbat=\"parent\"><a href=\"$parentURL\">Up to a higher level</a>\n"
	}
	incr skipLines 
    }

    if { $skipLines } {
	set contents [lrange $contents $skipLines end]
    }

    foreach $PRlistParam $PRlistSpecs {
	set maxlen($name) 0
    }
    foreach line $contents {
	if {[string trim $line] == {}} continue;
        foreach $PRlistParam $PRlistSpecs  {
	    if { $fmt == {} } { set fmt "%s" }
	    set maxlen($name) [max $maxlen($name) [string length [format $fmt [lrange $line $f $l]]]]
	}
    }
    foreach $PRlistParam $PRlistSpecs {
	incr maxlen($name) $sp
    } 

    # need to abbreviate long names somewhere here ...

    foreach line $contents {
	if {[string compare {} [string trim $line]]} {
	    switch -- [string index [lindex $line 0] 0] {
		d {set dingbat folder}
		l {set dingbat next}
		- {
		    set dingbat unknown.document
		    global PRsuff_icon
		    catch {set dingbat $PRsuff_icon([file extension [lindex $line 8]])}
		}
		default {set dingbat unknown.document}
	    }

	    set text [format %-$maxlen(filename)s [set target [lrange $line 8 end]]]
	    append doc "<li dingbat=\"$dingbat\"><a href=\""
	    if {[regexp {(.*) -> (.*)} $target all src dst]} {
		append doc "[URI_make $var(scheme) [PRpath_handler $var(scheme) addcomponent $var(path) $src] {}]\">$text</a>"
	    } else {
		append doc "[URI_make $var(scheme) [PRpath_handler $var(scheme) addcomponent $var(path) $target] {}]\">$text</a>"
	    }

	    foreach $PRlistParam $PRlistSpecs1 {
		set text [format $fmt [lrange $line $f $l]]
		switch -- $just {
		    r {set ffmt "%$maxlen($name)s"}
		    l {set ffmt "%-$maxlen($name)s"}
		}
		append doc "[format $ffmt $text]"
	    }
	    if { [info exists dst] } { 
		append doc " -> $dst"
		unset dst
	    }
	}
    }
    append doc "
</pre>
</body>
</html>
"
    debug_puts finished
    return $doc
}


} \
cache.tcl #\ \$Id:\ cache.tcl,v\ 1.13\ 1996/08/06\ 11:54:23\ steve\ Exp\ \$\n#\n#\tCopyright\ (c)\ 1996,1997\ Australian\ National\ University\n#\n#--------\n#\n#\tPASTIME\ Project\n#\tCooperative\ Research\ Centre\ for\ Advanced\ Computational\ Systems\n#\tCOPYRIGHT\ NOTICE\ AND\ DISCLAIMER.\n#\n#\ Copyright\ (c)\ 1995,1996\ ANU\ and\ CSIRO\ on\ behalf\ of\ the\ participants\ in\ \n#\ the\ \ CRC\ for\ Advanced\ Computational\ Systems\ ('ACSys').\n#\ \n#\ ACSys\ makes\ this\ software\ and\ all\ associated\ data\ and\ documentation\n#\ ('Software')\ available\ free\ of\ charge\ for\ non-commercial\ purposes\ only.\ You\n#\ may\ make\ copies\ of\ the\ Software\ but\ you\ must\ include\ all\ of\ this\ notice\ on\n#\ any\ copy.\n#\ \n#\ The\ Software\ was\ developed\ for\ research\ purposes\ and\ ACSys\ does\ not\ warrant\n#\ that\ it\ is\ error\ free\ or\ fit\ for\ any\ purpose.\ \ ACSys\ disclaims\ any\n#\ liability\ for\ all\ claims,\ expenses,\ losses,\ damages\ and\ costs\ any\ user\ may\n#\ incur\ as\ a\ result\ of\ using,\ copying\ or\ modifying\ the\ Software.\n#\n###\n#\ cache.tcl\ provides\ routines\ for\ managing\ the\ cache\ of\ images\ and\ hyperdocs.\n#\ All\ procedures\ and\ global\ variables\ used\ in\ this\ module\ are\ prepended\ with\n#\ \"Cache\".\n#\n#\ There\ are\ three\ levels\ of\ cache:\ post-processed,\ in-memory\ and\ on-disk.\n#\n#\ The\ global\ array\ Cache_info\ is\ used\ to\ store\ information\ about\ documents\ in\n#\ the\ cache.\ \ It\ is\ indexed\ by\ URL.\ \ Elements\ are\ lists\ {level\ {attr}\ size\ info\ data\ cachefile}\n#\ where\ \"level\"\ indicates\ how\ the\ cache\ has\ the\ document\ stored\ (PRdata\ for\ in-memory\n#\ data,\ PRpost\ for\ post-processed\ and\ PRfile\ for\ on-disk),\ \"attr\"\ is\ a\ list\ containing\n#\ attribute,value\ pairs\ to\ be\ set\ in\ state\ variables,\ \"info\"\ is\ the\ data\ appropriate\n#\ to\ the\ level,\ \"data\"\ is\ the\ raw\ data\ (if\ any)\ and\ cachefile\ is\ the\ local\ file\n#\ containing\ the\ raw\ data.\n#\n#\ The\ cache\ is\ persistent,\ but\ is\ not\ intended\ to\ support\ use\ by\ multiple\ browsers.\n#\ The\ problem\ is\ locking\ the\ Index\ file\ for\ updates.\n#\n#\ The\ Index\ file\ in\ the\ cache\ directory\ contains\ meta-data\ (ie.\ HTTP\ headers)\ \n#\ for\ all\ files\ in\ the\ cache.\n\npackage\ provide\ cache\ \[lindex\ {\$Revision:\ 1.13\ \$}\ 1\]\n\n#\ Initialisation\n\n#\ See\ end\ of\ file\n\n###\n\n#\ Cache_get_document\ checks\ whether\ a\ document\ is\ in\ the\ cache.\ \ If\ so\ then\n#\ the\ supplied\ state\ array\ is\ initialised\ with\ the\ details\ of\ the\ document\n#\ and\ the\ appropriate\ return\ value\ passed\ back,\ see\ loadData.\n#\ Otherwise\ an\ empty\ string\ is\ returned.\n\nproc\ Cache_get_document\ {state\ data}\ {\n\ \ \ \ global\ Cache_info\ surfit\n\ \ \ \ upvar\ #0\ \$state\ var\n\ \ \ \ upvar\ 2\ \$data\ data_return\n\n\ \ \ \ if\ {\[set\ hit\ \[array\ get\ Cache_info\ \$var(url)\]\]\ !=\ {}}\ {\n\tset\ hit\ \[lindex\ \$hit\ 1\]\n\t#\ Check\ for\ redirection\n\tif\ {\[set\ idx\ \[lsearch\ \[lindex\ \$hit\ 1\]\ HDRlocation\]\]\ !=\ -1}\ {\n\t\ \ \ \ set\ newloc\ \[lindex\ \[lindex\ \$hit\ 1\]\ \[expr\ \$idx\ +\ 1\]\]\n\t\ \ \ \ #\ Re-initialise\ state\ var\ with\ new\ URL\n\t\ \ \ \ URI_parse\ \$newloc\ var(scheme)\ var(path)\ var(query)\ var(fragment)\n\t\ \ \ \ set\ var(url)\ \[URI_make\ \$var(scheme)\ \$var(path)\ \$var(query)\]\n\t\ \ \ \ return\ \[uplevel\ Cache_get_document\ \$state\ \$data\]\n\t}\n\tset\ data_return\ \[lindex\ \$hit\ 3\]\n\tarray\ set\ var\ \[lindex\ \$hit\ 1\]\n\tset\ var(readsize)\ \[lindex\ \$hit\ 2\]\n\tset\ var(eof)\ 1\n\tif\ {\[lindex\ \$hit\ 0\]\ ==\ \"PRfile\"}\ {\n\t\ \ \ \ #\ Override\ protocol\ handler\ to\ use\ file\n\t\ \ \ \ set\ var(read_handler)\ PRfile_read\n\t\ \ \ \ set\ var(file)\ \$data_return\n\t}\n\treturn\ \[lindex\ \$hit\ 0\]\n\ \ \ \ }\ else\ {return\ {}}\n}\n\n#\ Return\ the\ filename\ for\ the\ cache\ file\ for\ the\ given\ URL\n\nproc\ Cache_get_file\ {url}\ {\n\ \ \ \ global\ Cache_info\n\ \ \ \ return\ \[lindex\ \$Cache_info(\$url)\ 5\]\n}\n\n#\n#\ Add\ post-processed\ data\ to\ the\ cache\n#\nproc\ Cache_add_post\ {state\ post\ type\ {size\ 0}}\ {\n\ \ \ \ global\ Cache_info\n\ \ \ \ upvar\ #0\ \$state\ var\n\n\ \ \ \ if\ {!\[info\ exists\ Cache_info(\$var(url))\]}\ return\;\n\n\ \ \ \ set\ raw\ {}\n\ \ \ \ if\ {\[lindex\ \$Cache_info(\$var(url))\ 0\]\ ==\ \"PRdata\"}\ {\n\tset\ raw\ \[lindex\ \$Cache_info(\$var(url))\ 3\]\n\ \ \ \ }\n\ \ \ \ array\ set\ tmp\ \[lindex\ \$Cache_info(\$var(url))\ 1\]\n\ \ \ \ array\ set\ tmp\ \[list\ HDRcontent-type\ \$type\ HDRcontent-length\ \$size\]\n\ \ \ \ set\ Cache_info(\$var(url))\ \\\n\t\[lreplace\ \$Cache_info(\$var(url))\ 0\ 4\ PRpost\ \[array\ get\ tmp\]\ \$size\ \$post\ \$raw\]\n}\n\n#\n#\ Add\ a\ new\ document\ to\ the\ cache.\n#\nproc\ Cache_add_document\ {type\ state}\ {\n\ \ \ \ global\ Cache_info\ surfit\n\ \ \ \ upvar\ #0\ \$state\ var\n\n\ \ \ \ #\ Construct\ the\ list\ of\ attributes\ to\ store\ in\ the\ cache\n\ \ \ \ if\ {!\[info\ exists\ var(HDRcontent-length)\]}\ {\n\t#\ We\ need\ to\ know\ the\ size\ of\ the\ document\ for\ expiration\n\tset\ var(HDRcontent-length)\ 0\n\ \ \ \ }\n\ \ \ \ set\ attr\ \[array\ get\ var\ HDR*\]\n\ \ \ \ append\ attr\ \"\ \"\ \[array\ get\ var\ HTTP*\]\n\n\ \ \ \ switch\ \$type\ {\n\tPRfd\t{\n\t\ \ \ \ if\ {!\[info\ exists\ var(cachefile)\]}\ {\n\t\t#\ Open\ a\ file\ to\ receive\ the\ data\n\t\tset\ var(cachefile)\ \"\$surfit(CFGcachedir)\ \[incr\ surfit(cachecnt)\]\"\n\t\tif\ {!\[catch\ {open\ \[eval\ file\ join\ \$var(cachefile)\]\ w+}\ var(cachefd)\]}\ {\n\t\t\ \ \ \ set\ Cache_info(\$var(url))\ \[list\ PRfile\ \$attr\ \$var(HDRcontent-length)\ \$var(cachefile)\ {}\ \$var(cachefile)\]\n\t\t}\ else\ {unset\ var(cachefd)}\n\t\ \ \ \ }\n\t}\n\n\tPRfile\t{set\ Cache_info(\$var(url))\ \[list\ \$type\ \$attr\ \$var(HDRcontent-length)\ \$var(file)\ {}\ \$var(file)\]}\n\n\tPRdata\t{\n\t\ \ \ \ set\ filename\ {}\n\t\ \ \ \ if\ {\[info\ exists\ var(file)\]}\ {\n\t\tset\ filename\ \$var(file)\n\t\ \ \ \ }\ elseif\ {!\[info\ exists\ var(cachefile)\]}\ {\n\t\t#\ Open\ a\ file\ to\ receive\ the\ data\n\t\tset\ filename\ \[set\ var(cachefile)\ \"\$surfit(CFGcachedir)\ \[incr\ surfit(cachecnt)\]\"\]\n\t\tif\ {\[catch\ {open\ \[eval\ file\ join\ \$var(cachefile)\]\ w+}\ var(cachefd)\]}\ {\n\t\t\ \ \ \ unset\ var(cachefd)\n\t\t\ \ \ \ unset\ var(cachefile)\;\ set\ filename\ {}\n\t\t}\n\t\ \ \ \ }\ else\ {\n\t\tset\ filename\ \$var(cachefile)\n\t\ \ \ \ }\n\t\ \ \ \ set\ Cache_info(\$var(url))\ \[list\ \$type\ \$attr\ \$var(HDRcontent-length)\ {}\ {}\ \$filename\]\n\t}\n\n\tPRpost\t{Cache_add_post\ \$var(url)\ \$data\ \$type\ \$var(HDRcontent-length)}\n\n\tdefault\ {error\ \"improper\ data\ type\ \\\"\$type\\\"\"}\n\ \ \ \ }\n}\n\nproc\ Cache_append_post\ {state\ post}\ {\n\ \ \ \ global\ Cache_info\n\ \ \ \ upvar\ #0\ \$state\ var\n\n\ \ \ \ if\ {\[set\ hit\ \[array\ get\ Cache_info\ \$var(url)\]\]\ ==\ {}}\ {return}\ \;#\ document\ not\ in\ cache\n\ \ \ \ set\ hit\ \[lindex\ \$hit\ 1\]\n\n\ \ \ \ if\ {\[lindex\ \$hit\ 0\]\ !=\ \"PRpost\"}\ {error\ \"cached\ document\ is\ not\ post-processed\"}\n\n\ \ \ \ set\ new\ \[lindex\ \$hit\ 3\]\n\ \ \ \ append\ new\ \$post\n\ \ \ \ set\ Cache_info(\$var(url))\ \[lreplace\ \$hit\ 3\ 3\ \$new\]\n}\n\n#\ Data\ gets\ read\ straight\ into\ the\ cache\ file,\ and\ then\ read\ back\ out.\n#\ This\ procedure\ only\ appends\ the\ data\ to\ the\ in-memory\ copy.\n\nproc\ Cache_append_data\ {state\ data}\ {\n\ \ \ \ global\ Cache_info\n\ \ \ \ upvar\ #0\ \$state\ var\n\n\ \ \ \ if\ {\[set\ hit\ \[array\ get\ Cache_info\ \$var(url)\]\]\ ==\ {}}\ {return}\ \;#\ document\ not\ in\ cache\n\ \ \ \ set\ hit\ \[lindex\ \$hit\ 1\]\n\n\ \ \ \ #\ Add\ data\ to\ memory\ cache\n\ \ \ \ switch\ \[lindex\ \$hit\ 0\]\ {\n\tPRdata\t{\n\t\ \ \ \ set\ new\ \[lindex\ \$hit\ 3\]\n\t\ \ \ \ append\ new\ \$data\n\t\ \ \ \ set\ Cache_info(\$var(url))\ \[lreplace\ \$hit\ 2\ 3\ \ \[string\ length\ \$new\]\ \$new\]\n\t}\n\n\tPRfile\t{}\n\tPRpost\t{\n\t\ \ \ \ set\ rawData\ \[lindex\ \$hit\ 4\]\n\t\ \ \ \ append\ rawData\ \$data\n\t\ \ \ \ set\ Cache_info(\$var(url))\ \[lreplace\ \$hit\ 4\ 4\ \$rawData\]\n\t}\n\n\tPRfd\t{error\ \"can't\ append\ data\ to\ a\ file\ descriptor\"}\n\tdefault\ {error\ \"improper\ cache\ data\ type\ \\\"\[lindex\ \$hit\ 0\]\\\"\ value\"}\n\ \ \ \ }\n}\n\n#\ Change\ the\ attributes\ for\ a\ document\ in\ the\ cache\n\nproc\ Cache_set_header\ {url\ hdr\ value}\ {\n\ \ \ \ global\ Cache_info\n\n\ \ \ \ if\ {!\[info\ exists\ Cache_info(\$url)\]}\ {return}\ \;#\ Not\ there\n\ \ \ \ array\ set\ tmp\ \[lindex\ \$Cache_info(\$url)\ 1\]\n\ \ \ \ set\ tmp(HDR\$hdr)\ \$value\n\ \ \ \ set\ Cache_info(\$url)\ \[lreplace\ \$Cache_info(\$url)\ 1\ 1\ \[array\ get\ tmp\]\]\n}\n\nproc\ Cache_set_content_type\ {url\ type}\ {\n\ \ \ \ Cache_set_header\ \$url\ content-type\ \$type\n}\n\nproc\ Cache_set_content_length\ {url\ length}\ {\n\ \ \ \ Cache_set_header\ \$url\ content-length\ \$length\n}\n\n#\ Cache_close\ informs\ the\ cache\ that\ EOF\ has\ been\ detected\ while\n#\ downloading\ data\ for\ the\ document.\n\nproc\ Cache_close\ {state}\ {\n\ \ \ \ global\ Cache_info\ surfit\n\ \ \ \ upvar\ #0\ \$state\ var\n\n\ \ \ \ if\ {!\[info\ exists\ Cache_info(\$var(url))\]}\ {\n\t#\ No\ data\ for\ the\ document\ was\ ever\ received\n\t#\ Create\ an\ entry\ now\ with\ empty\ data\n\tCache_add_document\ PRdata\ \$state\n\ \ \ \ }\n\ \ \ \ if\ {\[set\ cid\ \[array\ get\ var\ cachefd\]\]\ !=\ {}}\ {\n\tclose\ \[lindex\ \$cid\ 1\]\n\tunset\ var(cachefd)\n\n\t#\ Append\ the\ meta\ data\ file\ to\ the\ cache\ index\n\t#\ Don't\ include\ data\n\t#\ TODO:\ If\ we\ lock\ the\ file,\ then\ multiple\ browsers\ could\ use\ the\ cache\n\tif\ {!\[catch\ {open\ \[eval\ file\ join\ \$surfit(CFGcachedir)\ Index\]\ a}\ filed\]}\ {\n\t\ \ \ \ puts\ \$filed\ \"array\ set\ Cache_info\ \\{\$var(url)\ \\{PRfile\ \[lrange\ \$Cache_info(\$var(url))\ 1\ 2\]\ \[list\ \[lindex\ \$Cache_info(\$var(url))\ 5\]\]\ {}\ \[list\ \[lindex\ \$Cache_info(\$var(url))\ 5\]\]\\}\\}\"\n\t\ \ \ \ puts\ \$filed\ \"set\ surfit(cachecnt)\ \$surfit(cachecnt)\"\t\;#\ Only\ need\ this\ once\ for\ the\ whole\ file\n\t\ \ \ \ close\ \$filed\n\t}\ else\ {\n\t\ \ \ \ puts\ stderr\ \"unable\ to\ append\ meta-data\ to\ cache\ index\"\n\t}\n\ \ \ \ }\n}\n\n#\ Given\ a\ cache\ entry\ with\ post-processed\ data,\ return\ the\ original\ raw\ data\n\nproc\ Cache_get_raw_data\ {url\ data}\ {\n\ \ \ \ global\ Cache_info\n\ \ \ \ upvar\ \$data\ data_return\n\ \ \ \ if\ {\[info\ exists\ Cache_info(\$url)\]}\ {\n\tif\ {\[lindex\ \$Cache_info(\$url)\ 0\]\ ==\ \"PRpost\"}\ {\n\t\ \ \ \ set\ data_return\ \[lindex\ \$Cache_info(\$url)\ 4\]\n\t\ \ \ \ return\ PRdata\t\;#\ May\ be\ wrong\n\t}\ else\ {\n\t\ \ \ \ set\ data_return\ \[lindex\ \$Cache_info(\$url)\ 3\]\n\t\ \ \ \ return\ \[lindex\ \$Cache_info(\$url)\ 0\]\n\t}\n\ \ \ \ }\ else\ {return\ {}}\n}\n\n#\ Cache_delete\ removes\ a\ document\ from\ the\ in-memory\ cache\n\nproc\ Cache_delete\ {url}\ {\n\ \ \ \ global\ Cache_info\n\ \ \ \ if\ {\[set\ hit\ \[array\ get\ Cache_info\ \$url\]\]\ !=\ {}}\ {\n\tset\ hit\ \[lindex\ \$hit\ 1\]\n\tif\ {\[lindex\ \$hit\ 0\]\ ==\ \"PRimage\"}\ {\n\t\ \ \ \ image\ delete\ \[lindex\ \$hit\ 3\]\n\t}\n\ \ \ \ }\n\ \ \ \ catch\ \"unset\ Cache_info(\$url)\"\n}\n\n#\ Flush\ all\ entries\ from\ the\ in-memory\ cache\n\nproc\ Cache_flush\ {}\ {\n\ \ \ \ global\ Cache_info\n\ \ \ \ foreach\ url\ \[array\ names\ Cache_info\]\ {Cache_delete\ \$url}\n}\n\n#\ Cache_expire\ manages\ the\ removal\ of\ cached\ documents\ from\ each\ of\ the\n#\ different\ cache\ levels\ to\ meet\ resource\ usage\ constraints.\ \n#\ The\ application\ will\ probably\ want\ to\ run\ this\ in\ the\ background,\n#\ so\ it\ should\ be\ careful\ not\ to\ do\ too\ much\ processing\ at\ a\ time.\n\nproc\ Cache_expire\ {}\ {\n\ \ \ \ error\ \"not\ yet\ implemented\"\n}\n\n#\ Vaporise\ all\ traces\ of\ cached\ documents\n\nproc\ Cache_expire_all\ {}\ {\n\ \ \ \ global\ surfit\n\n\ \ \ \ Cache_flush\n\ \ \ \ catch\ {foreach\ f\ \[glob\ \[eval\ file\ join\ \$surfit(CFGcachedir)\ *\]\]\ {File_delete\ \$f}}\n}\n\n#\ Cache_expire_document\ removes\ the\ given\ document\ from\ the\ cache\n\nproc\ Cache_expire_document\ {url}\ {\n\ \ \ \ global\ surfit\ Cache_info\n\ \ \ \ #\ Only\ remove\ file\ in\ the\ cache\n\ \ \ \ if\ {\[string\ match\ \"\$surfit(CFGcachedir)\ *\"\ \[lindex\ \$Cache_info(\$url)\ 5\]\]}\ {\n\tcatch\ {File_delete\ \[lindex\ \$Cache_info(\$url)\ 5\]}\n\ \ \ \ }\n\ \ \ \ Cache_delete\ \$url\n\ \ \ \ #\ Remove\ the\ entry\ from\ the\ Index\ file\ by\ rewriting\ the\ file.\n\ \ \ \ #\ Only\ write\ cached\ documents\ into\ the\ index.\n\ \ \ \ set\ filed\ \[open\ \[eval\ file\ join\ \$surfit(CFGcachedir)\ Index\]\ w\]\n\ \ \ \ foreach\ entry\ \[array\ names\ Cache_info\]\ {\n\tif\ {\[string\ match\ \"\$surfit(CFGcachedir)\ *\"\ \[lindex\ \$Cache_info(\$entry)\ 5\]\]}\ {\n\t\ \ \ \ puts\ \$filed\ \"array\ set\ Cache_info\ \\{\$entry\ \\{PRfile\ \[lrange\ \$Cache_info(\$entry)\ 1\ 2\]\ \[list\ \[lindex\ \$Cache_info(\$entry)\ 5\]\]\ {}\ \[list\ \[lindex\ \$Cache_info(\$entry)\ 5\]\]\\}\\}\"\n\t}\n\ \ \ \ }\n\ \ \ \ puts\ \$filed\ \"set\ surfit(cachecnt)\ \$surfit(cachecnt)\"\n\ \ \ \ close\ \$filed\n}\n\n###\ Cache\ initialisation\n\nproc\ Cache_init\ {}\ {\n\ \ \ \ global\ surfit\ Cache_info\ tcl_platform\n\ \ \ \ catch\ {unset\ Cache_info}\n\n\ \ \ \ #\ Set\ the\ various\ cache\ defaults\n\ \ \ \ #\ For\ resource\ limits,\ a\ value\ of\ -1\ implies\ no\ limit\ and\ a\ value\ of\ 0\n\ \ \ \ #\ implies\ no\ caching\ (or\ just\ immediate\ expiration?).\n\n\ \ \ \ if\ {!\[info\ exists\ surfit(CFGcachedir)\]}\ {\n\tswitch\ \$tcl_platform(platform)\ {\n\t\ \ \ \ unix\ {\n\t\tif\ {\[info\ exists\ \[eval\ file\ join\ surfit(userdir)\]\]}\ {\n\t\t\ \ \ \ set\ surfit(CFGcachedir)\ \"\$surfit(userdir)\ Cache\"\n\t\t}\ else\ {\n\t\t\ \ \ \ puts\ stderr\ \"unable\ to\ find\ cache\ directory.\ \ Set\ surfit(CFGcachedir)\"\n\t\t}\n\t\ \ \ \ }\n\t\ \ \ \ windows\ {\n\t\tif\ {\[info\ exists\ surfit(userdir)\]}\ {\n\t\t\ \ \ \ set\ surfit(CFGcachedir)\ \"\$surfit(userdir)\ Cache\"\n\t\t}\ elseif\ {\[info\ exists\ env(TMP)\]}\ {\n\t\t\ \ \ \ set\ surfit(CFGcachedir)\ \"\[file\ split\ \[File_normalizeFile\ \$env(TMP)\]\]\ SurfIt\"\n\t\t}\ elseif\ {\[info\ exists\ env(TEMP)\]}\ {\n\t\t\ \ \ \ set\ surfit(CFGcachedir)\ \"\[file\ split\ \[File_normalizeFile\ \$env(TEMP)\]\]\ SurfIt\"\n\t\t}\ elseif\ {!\[file\ exists\ \[file\ join\ \[file\ split\ C:/SURFITCA.CHE\]\]\]}\ {\n\t\t\ \ \ \ set\ surfit(CFGcachedir)\ \[file\ split\ C:/SURFITCA.CHE\]\n\t\t}\ else\ {\n\t\t\ \ \ \ puts\ stderr\ \"unable\ to\ find\ cache\ directory.\ \ Set\ surfit(CFGcachedir)\"\n\t\t}\n\t\ \ \ \ }\n\t\ \ \ \ macintosh\ {\n\t\tif\ {\[info\ exists\ surfit(userdir)\]}\ {\n\t\t\ \ \ \ set\ surfit(CFGcachedir)\ \"\$surfit(userdir)\ Cache\"\n\t\t}\ else\ {\n\t\t\ \ \ \ puts\ stderr\ \"unable\ to\ find\ cache\ directory.\ \ Set\ Cache\ Directory\ in\ preferences\"\n\t\t}\n\t\ \ \ \ }\n\t}\n\ \ \ \ }\n\n\ \ \ \ #\ Create\ cache\ directory\ if\ it\ doesn't\ already\ exist\n\ \ \ \ if\ {!\[file\ isdirectory\ \[eval\ file\ join\ \$surfit(CFGcachedir)\]\]}\ {\n\ \ \ \ \ \ \ \ if\ {\[file\ exists\ \[eval\ file\ join\ \$surfit(CFGcachedir)\]\]}\ {\n\t\ \ \ \ #\ Find\ an\ unused\ name\n\t\ \ \ \ for\ {set\ counter\ 0}\ {\[file\ exists\ \[eval\ file\ join\ \$surfit(CFGcachedir)\$counter\]\]}\ {incr\ counter}\ {}\n\t\ \ \ \ append\ surfit(CFGcachedir)\ \$counter\n\t}\n\tif\ {\[catch\ {File_mkdir\ \$surfit(CFGcachedir)}\]}\ {\n\t\ \ \ \ puts\ stderr\ \"error:\ unable\ to\ create\ cache\ directory\ \\\"\[eval\ file\ join\ \$surfit(CFGcachedir)\]\\\"\"\n\t\ \ \ \ #\ Should\ disable\ caching\ of\ binary\ data\ documents\n\t}\n\ \ \ \ }\n\n\ \ \ \ #\ Read\ in\ the\ Index\ file\ which\ describes\ the\ contents\ of\ the\ cache\n\n\ \ \ \ if\ {\[file\ readable\ \[eval\ file\ join\ \$surfit(CFGcachedir)\ Index\]\]}\ {\n\ \ \ \ \ \ \ \ if\ {\[catch\ {source\ \[eval\ file\ join\ \$surfit(CFGcachedir)\ Index\]}\]}\ {\n\ \ \ \ \ \ \ \ \ \ \ \ puts\ stderr\ \"unable\ to\ read\ cache\ index\ \\\"\[eval\ file\ join\ \$surfit(CFGcachedir)\ Index\]\\\"\"\n\ \ \ \ \ \ \ \ \ \ \ \ puts\ stderr\ \"clearing\ cache\"\n\ \ \ \ \ \ \ \ \ \ \ \ Cache_expire_all\n\t\ \ \ \ }\n\ \ \ \ }\ else\ {\n\t\ \ \ \ #\ Clean\ out\ the\ cache\ directory\n\ \ \ \ \ \ \ \ Cache_expire_all\n\ \ \ \ }\n\n\ \ \ \ array\ set\ surfit\ {\n\tcachedisklimit\ -1\ cachedatalimit\ -1\ cacheimagelimit\ -1\n\ \ \ \ }\n\ \ \ \ if\ {!\[info\ exists\ surfit(cachecnt)\]}\ {\n\tset\ surfit(cachecnt)\ 0\n\ \ \ \ }\n}\n\n \
htmllib.tcl {# $Id: htmllib.tcl,v 1.11 1996/06/26 04:32:02 steve Exp $
#
#	Copyright (c) 1996,1997 Australian National University
#
#--------
#
#	PASTIME Project
#	Cooperative Research Centre for Advanced Computational Systems
#	COPYRIGHT NOTICE AND DISCLAIMER.
#
# Copyright (c) 1995,1996 ANU and CSIRO on behalf of the participants in 
# the  CRC for Advanced Computational Systems ('ACSys').
# 
# ACSys makes this software and all associated data and documentation
# ('Software') available free of charge for non-commercial purposes only. You
# may make copies of the Software but you must include all of this notice on
# any copy.
# 
# The Software was developed for research purposes and ACSys does not warrant
# that it is error free or fit for any purpose.  ACSys disclaims any
# liability for all claims, expenses, losses, damages and costs any user may
# incur as a result of using, copying or modifying the Software.
#
###  Notice from html_library.tcl:
#
# Simple HTML display library by Stephen Uhler (stephen.uhler@sun.com)
# Copyright (c) 1995 by Sun Microsystems
# Version 0.3 Fri Sep  1 10:47:17 PDT 1995
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
###
#
# To use this package, see http://tcltk.anu.edu.au/eng/

package provide html_library [lindex {$Revision: 1.11 $} 1]

proc HMlibrary_init {} {

    # There is a bug in the tcl library focus routines that prevents focus
    # from every reaching an un-viewable window.  Use our *own*
    # version of the library routine, until the bug is fixed, make sure we
    # over-ride the library version, and not the otherway around

    auto_load tkFocusOK
    proc tkFocusOK w {
	set code [catch {$w cget -takefocus} value]
	if {($code == 0) && ($value != "")} {
	    if {$value == 0} {
		return 0
	    } elseif {$value == 1} {
		return 1
	    } else {
		set value [uplevel #0 $value $w]
		if {$value != ""} {
		    return $value
		}
	    }
	}
	set code [catch {$w cget -state} value]
	if {($code == 0) && ($value == "disabled")} {
	    return 0
	}
	regexp Key|Focus "[bind $w] [bind [winfo class $w]]"
    }

    HMoptimize
}

##################################################################
############################################
# mapping of html tags to text tag properties
# properties beginning with "T" map directly to text tags

# These are Defined in HTML 3.0

# HMtag_map contains mapping that apply to the hyperwindow

array set HMtag_map {
	address	{style i}
	b      {weight bold}
	blockquote	{style i}
	bq		{style i}
	cite   {style i}
	code   {family courier}
	dfn    {style i}	
	em     {style i}
	h1     {size 24 weight bold}
	h2     {size 22}		
	h3     {size 20}	
	h4     {size 18}
	h5     {size 16}
	h6     {style i}
	i      {style i}
	kbd    {family courier weight bold}
	pre    {family courier}
	samp   {family courier}		
	strong {weight bold}		
	tt     {family courier}
	u	 {Tunderline underline}
	var    {style i}	
}

# These are defined in html3.2

array set HMtag_map {
	center {Tcenter center}
	strike {Tstrike strike}
	big    {size 18}
	small  {size 10}
	sub    {size 8 Tsubscript subscript}
	sup    {size 8 Tsuperscript superscript}
}

# HMwintag_map contains values that are only set for a window

array set HMwintag_map {
	blockquote	{Trindent rindent indent 1}
	bq		{Trindent rindent indent 1}
	dir	{indent 1}
	dl	{indent 1}
	dd	{indent 1}
	menu	{indent 1}
	ol	{indent 1}
	pre	{fill 0 Tnowrap nowrap pre 1}
	ul     {indent 1}
}

# initial values

set HMtag_map(hmstart) {
	family times   weight medium   style r   size 14
	Tcenter ""   Tlink ""   Tunderline ""   list list
	counter 0 adjust 0
}
set HMwintag_map(hmstart) {
	Tnowrap "" fill 1 pre 0
}

# Left alignment is the default, so no need to explicitly set a tag
# (kludge to fix proper stacking of alignment)

array set HMalign_map {
	left {}
	center {Tcenter center}
	right {Tright right}
}

# Map HTML standard colours that are not X standard colours

array set HMcolour_map {
	aqua aquamarine  fuchsia hotpink lime limegreen  olive olivedrab  silver #d9d9d9 teal lightseagreen
}
proc HMmap_colour {colour} {
	global HMcolour_map
	if {[regexp {^#.*} $colour]} {return $colour}
	catch {set colour $HMcolour_map($colour)}
	return $colour
}

# html tags that insert white space
# A better algorithm to insert white space is to only add
# the first newline if the insertion point is not at the
# beginning of the line.  Normalise the array HMinsert_map
# so that every entry is assumed to cause a newline if the
# above condition holds.

array set HMinsert_map {
	address	"\n"
	blockquote "\n" /blockquote ""
	br	""
	dd	"" /dd	""
		   /dl	"\n"
	dt	"\n"
	form ""	/form ""
	h1	"\n"	/h1	"\n"
	h2	"\n"	/h2	"\n"
	h3	"\n"	/h3	""
	h4	""	/h4	""
	h5	""	/h5	""
	h6	""	/h6	""
	li   ""
	/dir ""
	/ul ""
	/ol ""
	/menu ""
	p	"\n"	/p	""
	center	""	/center	""
	pre ""	/pre ""
}

# tags that are list elements, that support "compact" rendering

array set HMlist_elements {
	ol 1   ul 1   menu 1   dl 1   dir 1
}
############################################
# initialize the window and stack state

proc HMinit_hyperwin {win} {
	HMinit_state $win
	HMpush_win HM$win $win
}

############################################
# initialize a window and push it onto the widget stack

proc HMpush_win {state win} {
	upvar #0 $state var
	upvar #0 HN$win winvar

	# Push the window
	set var(current) $win
	set var(stack) [linsert $var(stack) 0 $win]

	# Initialise its state
	HMinit_win_state $win
	$win tag configure underline -underline 1
	$win tag configure center -justify center
	$win tag configure left -justify left
	$win tag configure right -justify right
	$win tag configure nowrap -wrap none
	$win tag configure rindent -rmargin $var(S_tab)c
	$win tag configure strike -overstrike 1
	$win tag configure subscript -offset -4
	$win tag configure superscript -offset 4
	$win tag configure mark -foreground [HMmap_colour red]		;# list markers
	$win tag configure list -spacing1 3p -spacing3 3p		;# regular lists
	$win tag configure compact -spacing1 0p		;# compact lists
	if {[catch {$win tag configure bullet -font [HMx_font zapfdingbats 12 medium r]}]} {
		# Default to same as hmstart
		$win tag configure bullet -font [HMx_font times 14 medium r]
		set var(S_symbols) O*=+-o\xd7\xb0>:\xb7
	}
	$win tag configure link -borderwidth 2 -foreground [HMmap_colour $var(linkfg)]	;# hypertext links
	# Handle B&W displays
	if {[winfo depth $win] == 1} {
		$win tag configure link -underline 1
	}
	HMset_indent $win $var(S_tab)
	$win configure -wrap word
	catch {$win configure -background [HMmap_colour $var(bgcolor)]}
	catch {$win configure -foreground [HMmap_colour $var(fgcolor)]}

	# configure the text insertion point
	$win mark set $var(S_insert) 1.0

	# for horizontal rules
	if {[catch {$win tag configure thin -font [HMx_font times 2 medium r]}]} {
		# Default to standard size
		$win tag configure thin -font [HMx_font times 8 medium r]
	}
	$win tag configure hr -relief sunken -borderwidth 2 -wrap none  -tabs [winfo width $win]
	bind $win <Configure> {
		%W tag configure hr -tabs %w
		%W tag configure last -spacing3 %h
	}

	# generic link enter callback

	$win tag bind link <ButtonRelease-1> "HMlink_hit $state $win %x %y"
}

proc HMpop_win {state} {
	upvar #0 $state var
	set var(stack) [lreplace $var(stack) 0 0]
	set var(current) [lindex $var(stack) 0]
}

# set the indent spacing (in cm) for lists
# TK uses a "weird" tabbing model that causes \t to insert a single
# space if the current line position is past the tab setting

proc HMset_indent {win cm} {
	set tabs [expr $cm / 2.0]
	$win configure -tabs ${tabs}c
	foreach i {1 2 3 4 5 6 7 8 9} {
		set tab [expr $i * $cm]
		$win tag configure indent$i -lmargin1 ${tab}c -lmargin2 ${tab}c  -tabs "[expr $tab + $tabs]c [expr $tab + 2*$tabs]c"
	}
}

# reset the state of window - get ready for the next page
# remove all but the font tags for the main window, 
# and remove all form state

proc HMreset_win {win} {
	upvar #0 HM$win var
	regsub -all { +[^L ][^ ]*} " [$win tag names] " {} tags
	catch "$win tag delete $tags"
	eval $win mark unset [$win mark names]
	$win delete 0.0 end
	$win tag configure hr -tabs [winfo width $win]

	# configure the text insertion point
	$win mark set $var(S_insert) 1.0

	# remove form state.  If any check/radio buttons still exists, 
	# their variables will be magically re-created, and never get
	# cleaned up.
	catch {eval unset [info globals HM$win.form*]}

	# Remove previous window state variables
	foreach w [info globals HN$win*] {
		upvar #0 $w winvar
		unset winvar
	}

	HMinit_state $win
	HMpush_win HM$win $win
	return HM$win
}

# initialize the window's state array
# two (or more) variables are used to store the state.
# HM$win (upvar'd to var) stores state associated with the document as a whole
# $var(current) is the window currently being rendered
# $var(stack) is a list of the current stack of nested windows
# ([lindex $var(stack) 0] == $var(currwin))
# HN$var(current) (upvar'd to winvar) stores state associated 
# with rendering into that window
# (NB. the use of the "HN" prefix rather than "HM" to avoid clashes)

# Parameters beginning with S_ are NOT reset
#  adjust_size:		global font size adjuster
#  unknown:		character to use for unknown entities
#  tab:			tab stop (in cm)
#  symbols:		Symbols to use on un-ordered lists
#  current:		Text widget to render
#  stack:		the widget stack

proc HMinit_state {win} {
	upvar #0 HM$win var
	array set tmp [array get var S_*]
	catch {unset var}
	array set var {
		S_adjust_size 0
		S_tab 1.0
		S_unknown \xb7
		S_symbols O*=+-o\xd7\xb0>:\xb7
		S_insert Insert
		current {}
		stack {}
		tables {}
		tags 0
		linkfg blue
		basefontSize 3
	}
	set var(basewin) $win
	set var(S_bgcolor) [$win cget -background]
	set var(S_fgcolor) [$win cget -foreground]
	array set var [array get tmp]
	set var(bgcolor) $var(S_bgcolor)
	set var(fgcolor) $var(S_fgcolor)
}

proc HMinit_win_state {win} {
	upvar #0 HN$win winvar
	catch {unset winvar}
	array set winvar {
		fill 1
		list list
		indent ""
	}
}

# alter the parameters of the text state
# this allows an application to over-ride the default settings
# it is called as: HMset_state -param value -param value ...

array set HMparam_map {
	-tab S_tab
	-unknown S_unknown
	-size S_adjust_size
	-symbols S_symbols
    -insert S_insert
    -foreground S_fgcolor
    -background S_bgcolor
}

proc HMset_state {win args} {
	upvar #0 HM$win var
	global HMparam_map
	set bad 0
	if {[catch {array set params $args}]} {return 0}
	foreach i [array names params] {
		incr bad [catch {set var($HMparam_map($i)) $params($i)}]
	}
	return [expr $bad == 0]
}

############################################
# manage the display of html

# HMrender gets called for every html tag
#   cookie:   The name of the main text widget to render into
#   tag:   The html tag (in arbitrary case)
#   not:   a "/" or the empty string
#   param: The un-interpreted parameter list
#   text:  The plain text until the next html tag

proc HMrender {cookie tag not param text} {
	global HMtag_map HMwintag_map HMinsert_map HMlist_elements
	upvar #0 HM$cookie var
	if {![catch {$var(noProcessCmd) HM$cookie $var(current) tag not param text} result] && !$result} return;
	upvar #0 HN$var(current) winvar

	set tag [string tolower $tag]
	set text [HMmap_esc $text]

	# manage compact rendering of lists
	if {[info exists HMlist_elements($tag)]} {
		set list "list [expr {[HMextract_param $param compact] ? "compact" : "list"}]"
	} else {
		set list ""
	}

	# adjust (push or pop) tag state
	catch {HMstack HM$cookie $not $HMtag_map($tag)} err
	catch {HMstack HN$var(current) $not "$HMwintag_map($tag) $list"} err

	# insert white space (with current font)
	# adding white space can get a bit tricky.  This isn't quite right
	if {[info exists HMinsert_map($not$tag)]} {
		if {[lindex [split [$var(current) index $var(S_insert)] .] 1] != 0} {
			$var(current) insert $var(S_insert) \n "space $winvar(font)"
		}
		$var(current) insert $var(S_insert) $HMinsert_map($not$tag) "space $winvar(font)"
		if {[lindex $winvar(fill) end]} {
			set text [string trimleft $text]
		}
	}

	# to fill or not to fill
	if {[lindex $winvar(fill) end]} {
		set text [HMzap_white $text]
	} else {
		set text [HMtranslate_cr $text]
	}

	# generic mark hook
	catch {HMmark $not$tag $win $param text}

	# do any special tag processing
	catch {HMtag_$not$tag HM$cookie $var(current) $param text}
	#puts "HMtag_$not$tag returned \"$err\""
	# NB. current window may have changed, so winvar may now be invalid

	# add the text with proper tags
	set tags [HMcurrent_tags HM$cookie $var(current)]
	$var(current) insert $var(S_insert) $text $tags
}

proc HMgetUniqueNum {state} {
	upvar #0 $state var
	return [incr var(tags)]
}

# Empty html tags - all tags are defined here to improve performance,
# but may be redefined later

proc HMmark {args} {}
foreach tag {html /html head /head body /body  title /title isindex base style script meta link banner /banner  h1 h2 h3 h4 h5 h6 /h1 /h2 /h3 /h4 /h5 /h6  address /address map /map  p /p pre /pre ul /ul ol /ol li dl /dl dt dd div /div  blockquote /blockquote center /center hr br  form /form input select /select textarea /textarea option  table /table caption /caption tr td th  b /b i /i u /u big /big small /small sub /sub sup /sup font /font tt /tt strike /strike  em /em strong /strong dfn /dfn code /code samp /samp kbd /kbd var /var cite /cite  a /a img applet /applet param} {
    proc HMtag_$tag {args} {}
}

# html tags requiring special processing
# Procs of the form HMtag_<tag> or HMtag_</tag> get called just before
# the text for this tag is displayed.  These procs are called inside a 
# "catch" so it is OK to fail.
#   state: The name of the hyperdocument state array
#   win:   The name of the text widget to render into
#   param: The un-interpreted parameter list
#   text:  A pass-by-reference name of the plain text until the next html tag
#          Tag commands may change this to affect what text will be inserted
#          next.

# A pair of pseudo tags are added automatically as the 1st and last html
# tags in the document.  The default is <HMstart> and </HMstart>.
# Append enough blank space at the end of the text widget while
# rendering so HMgoto can place the target near the top of the page,
# then remove the extra space when done rendering.

proc HMtag_hmstart {state win param text} {
	upvar #0 $state var
	$win mark gravity $var(S_insert) left
	$win insert end "\n " last
	$win mark gravity $var(S_insert) right
}

proc HMtag_/hmstart {state win param text} {
	$win delete last.first end
}

# Comment processing.  The HTML 2.0 spec seems to imply that tags cannot be
# nested within comments, but it is common practice - even W3C pages!
# This breaks the HTML parser (which is extremely simplistic), so we have to
# do alot of munging to fix things up.

proc HMtag_!-- {state win param text} {
	debug_puts [list HTML start comment]
	upvar #0 $state var
	upvar $text data
	if {![regexp -- {--$} $param]} {
		if {[regexp -- {-->(.*)} $data all remains]} {
			set data $remains
		} else {
			# Nested tag(s)
			set data {}
			set var(noProcessCmd) HMcomment
		}
	}
}

# Consume tags until we find '-->' (the end of the comment)

proc HMcomment {state win tag not param text} {
	upvar #0 $state var
	upvar $param elparam
	upvar $text data
	set remains $data
	debug_puts [list in comment, param $elparam, text $data]
	if {[regexp -- {--$} $elparam] || [regexp -- {-->(.*)} $data all remains]} {
		debug_puts [list found end]
		upvar $tag eltag
		upvar $not elnot
		set data $remains
		set elparam {}
		set eltag {}
		set elnot {}
		unset var(noProcessCmd)
		return 1
	} else {debug_puts [list still going]; return 0}
}

# put the document title in the window banner, and remove the title text
# from the document

proc HMtag_title {state win param text} {
	upvar $text data
	wm title [winfo toplevel $win] $data
	wm iconname [winfo toplevel $win] $data
	set data ""
}

# put enclosed text in the hyperdocument banner - a nonscrolled
# region.  There can only be one banner.

proc HMtag_banner {state win param text} {
	upvar #0 $state var
	# preserve the Insertion point
	set index [[set banner [HMget_banner $win]] index $var(S_insert)]
	# push banner onto the state
	HMpush_win $state $banner
	$banner mark set $var(S_insert) $index
}

proc HMtag_/banner {state win param text} {
	# Make the banner fit its text
	HMpop_win $state
}

# Links may be used to include an icon in the document "toolbar".
# In this case we use the hyperdocument banner as the toolbar.
# The HREF attribute is mandatory.

# draft-ietf-html-relrev-00 proposes these relationships:
#	MADE/AUTHOR COPYRIGHT DISCLAIMER META PUBLISHER TRADEMARK
#	TOC/CONTENTS INDEX NAVIGATOR 
#	CHILD PARENT SIBLING TOP/ORIGIN
#	NEXT PREV/PREVIOUS BEGIN/FIRST END/LAST
#	BIBLIOENTRY BIBLIOGRAPHY CITATION DEFINITION GLOSSARY FOOTNOTE 
# these are reserved
#	BACK FORWARD HOME

proc HMtag_link {state win param text} {
	upvar #0 $state var
	if {([HMextract_param $param rev link] || [HMextract_param $param rel link]) &&
	     [HMextract_param $param href]} {
		if {[regexp [set link [string tolower $link]] {back forward home}]} return;	# Illegal values
		set banner [HMget_banner $win]
		set item [button $banner.[HMgetUniqueNum $state]  -text [string toupper [string index $link 0]][string range $link 1 end]  -foreground orange  -command "HMlink_callback $state $banner $href"]
		HMlink_window_setup $item $href
		HMset_icon $state $item $link
		$banner window create $var(S_insert) -window $item -align center
	}
}

# Application must override this

proc HMget_banner {win} {
	return $win
}

# HTML 3.2 introduces various attributes for the BODY tag.
# BACKGROUND (Tk 4.1 cannot support this), BGCOLOR, TEXT, LINK, VLINK, ALINK

proc HMtag_body {state win param text} {
	upvar #0 $state var
	# Only valid at the toplevel
	if {$var(basewin) != $win} return;
	if {[HMextract_param $param bgcolor]} {
		set bgcolor [string tolower $bgcolor]
		if {![catch {$win configure -background [HMmap_colour $bgcolor]}]} {
			set var(bgcolor) [HMmap_colour $bgcolor]
		}
	}
	# GOTCHA: HMextract_param clobbers text argument
	if {[HMextract_param $param text]} {
		set text [string tolower $text]
		if {![catch {$win configure -foreground [HMmap_colour $text]}]} {
			set var(fgcolor) [HMmap_colour $text]
		}
	}
	if {[HMextract_param $param link]} {
		set link [string tolower $link]
		if {![catch {$win tag configure link -foreground [HMmap_colour $link]}]} {
			set var(linkfg) [HMmap_colour $link]
		}
	}
}

proc HMtag_hr {state win param text} {
	upvar #0 $state var
	$win insert $var(S_insert) "\n" space "\n" thin "\t" "thin hr" "\n" thin
}

array set HMfont_size_map {
    1 8 2 10 3 14 4 16 5 18 6 22 7 24
}
proc HMtag_font {state win param text} {
    global HMfont_size_map
    upvar #0 $state var
    set fontsize [lindex $var(size) end]
    if {[HMextract_param $param size]} {
	debug_puts [list got size param $size]
	debug_puts [list current font size $fontsize]
	if {[regexp {(\+|-)[0-9]+} $size]} {
	    # Relative to basefont
	    catch {set fontsize $HMfont_size_map([expr $var(basefontSize) + $size])}
	} else {
	    # Absolute font size
	    catch {set fontsize $HMfont_size_map($size)}
	}
    }
    debug_puts [list setting font size to $fontsize]
    HMstack $state "" [list size $fontsize]
    if {[HMextract_param $param color]} {
	set color [string tolower $color]
	$win tag configure colour$color -foreground [HMmap_colour $color]
	HMstack $state "" "Tcolour colour$color"
    } else {
	HMstack $state "" [list Tcolour {}]
    }
}

# Pop font size & colour stacks and pop tag stack as appropriate

proc HMtag_/font {state win param text} {
	upvar #0 $state var
	HMstack $state "/" [list size {}]
	HMstack $state "/" [list Tcolour {}]
}

proc HMtag_basefont {state win param text} {
    upvar #0 $state var
    if {[HMextract_param $param size]} {
	set var(basefontSize) $size
    }
}

proc HMtag_address {state win param text} {
	if {[HMextract_param $param nowrap]} {
		HMstack HN$win "" "Tnowrap nowrap"
	}
}

proc HMtag_/address {state win param text} {
	upvar #0 HN$win winvar
	if {[info exists winvar(Tnowrap)]} {
		unset winvar(Tnowrap)
		HMstack HN$win / "Tnowrap nowrap"
	}
}

proc HMtag_p {state win param text} {
	global HMalign_map
	upvar #0 HN$win winvar

	# Paragraphs acts as line breaks inside preformatted text
	if {$winvar(pre)} {
		HMrender $win br "" "" ""
	}

	# Implicitly end the previous alignment
	if {[info exists winvar(curralign)]} {HMtag_/p $state $win {} {}}

	set align [string tolower [HMget_alignment $param]]
	HMstack $state "" "$HMalign_map($align)"
	set winvar(curralign) $align

	if {[HMextract_param $param nowrap]} {
		HMstack HN$win "" "Tnowrap nowrap"
	}
}

proc HMtag_/p {state win param text} {
	upvar #0 $state var
	upvar #0 HN$win winvar
	global HMalign_map
	if {$winvar(pre)} return
	catch {HMstack $state / "$HMalign_map($winvar(curralign))"}
	catch {unset winvar(curralign)}
	if {[info exists winvar(Tnowrap)]} {
		unset winvar(Tnowrap)
		HMstack HN$win / "Tnowrap nowrap"
	}
}

# Process attributes common to all headings

proc HMheading {state win param} {
	global HMalign_map
	upvar #0 HN$win winvar

	if {[HMextract_param $param src]} {
		HMtag_img $state $win "src=$src" ""
	}

	# Positioning the SRC image before the DINGBAT image
	# is an arbitrary choice
	if {[HMextract_param $param dingbat]} {
		HMload_icon $state $win [string tolower $dingbat]
	}

	set align [string tolower [HMget_alignment $param]]
	HMstack $state "" "$HMalign_map($align)"
	set winvar(curralign) $align

	if {[HMextract_param $param nowrap]} {
		HMstack HN$win "" "Tnowrap nowrap"
	}
}

# Remove any alignment and nowrap attributes.
# Ending paragraphs does the same, so use that code.

proc HMendHeading {state win param} { 
	HMtag_/p $state $win $param {}
}

proc HMtag_h1 {state win param text} {
	HMheading $state $win $param
}
proc HMtag_h2 {state win param text} {
	HMheading $state $win $param
}
proc HMtag_h3 {state win param text} {
	HMheading $state $win $param
}
proc HMtag_h4 {state win param text} {
	HMheading $state $win $param
}
proc HMtag_h5 {state win param text} {
	HMheading $state $win $param
}
proc HMtag_h6 {state win param text} {
	HMheading $state $win $param
}

proc HMtag_/h1 {state win param text} {
	HMendHeading $state $win $param
}
proc HMtag_/h2 {state win param text} {
	HMendHeading $state $win $param
}
proc HMtag_/h3 {state win param text} {
	HMendHeading $state $win $param
}
proc HMtag_/h4 {state win param text} {
	HMendHeading $state $win $param
}
proc HMtag_/h5 {state win param text} {
	HMendHeading $state $win $param
}
proc HMtag_/h6 {state win param text} {
	HMendHeading $state $win $param
}

# Standard HTML v3.0 icon entity names are:

# ftp gopher telnet archive filing.cabinet folder fixed.disk disk.drive document unknown.document
# text.document binary.document binhex.document audio film image map form mail parent
# next previous home toc glossary index summary calculator caution clock compressed.data
# diskette display fax mail.in mail.out mouse printer tn3270 trash uuencoded.document

# Returns the widget pathname which has the icon image loaded

# Application should override these
proc HMload_icon {state win icon {href {}}} {
	puts "==> load icon \"$icon\""
}

proc HMset_icon {state win icon} {
	puts "==> set icon \"$icon\""
}

# list element tags

proc HMtag_ol {state win param text} {
	upvar #0 HN$win winvar
	set winvar(count$winvar(level)) 0
}

proc HMtag_ul {state win param text} {
	upvar #0 HN$win winvar
	catch {unset winvar(count$winvar(level))}
}

proc HMtag_menu {state win param text} {
	upvar #0 HN$win winvar
	set winvar(menu) ->
	set winvar(compact) 1
}

proc HMtag_/menu {state win param text} {
	upvar #0 HN$win winvar
	catch {unset winvar(menu)}
	catch {unset winvar(compact)}
}

# Definition lists are managed by dl & dd pushing the indent
# stack and dt popping it.

proc HMtag_dt {state win param text} {
	upvar #0 $state var
	if {![info exists var(defnTerm)]} {
		HMstack HN$win "/" "indent \"\""
		set var(defnTerm) 1
	}
}

# Sometimes people don't balance their <dt>'s and <dd>'s
# If we get a <dd> without a <dt> then pop the stack

proc HMtag_dd {state win param text} {
	upvar #0 $state var
	if {![info exists var(defnTerm)]} {
		HMstack HN$win "/" "indent \"\""
	}
	catch {unset var(defnTerm)}
}

proc HMtag_/dl {state win param text} {
	upvar #0 $state var
	if {[info exists var(defnTerm)]} {
		catch {HMstack HN$win "/" "indent \"\""}
	}
	catch {unset var(defnTerm)} err
}

proc HMtag_li {state win param text} {
	upvar #0 $state var
	upvar #0 HN$win winvar
	set level $winvar(level)
	incr level -1

	if {[HMextract_param $param dingbat]} {
		$win insert $var(S_insert) "\t" "mark [lindex $winvar(list) end] indent$level"
		HMload_icon $state $win [string tolower $dingbat]
		$win insert $var(S_insert) "\t"
	} else {
		set x [string index $var(S_symbols)+-+-+-+- $level]
		set b bullet
		catch {set x [incr winvar(count$level)]; set b ""}
		catch {set x $winvar(menu); set b ""}
		$win insert $var(S_insert) \t$x\t "mark $b [lindex $winvar(list) end] indent$level $winvar(font)"
	}
}

# Manage hypertext "anchor" links.  A link can be either a source (href)
# a destination (name) or both.  If its a source, register it via a callback,
# and set its default behavior.  If its a destination, check to see if we need
# to go there now, as a result of a previous HMgoto request.  If so, schedule
# it to happen with the closing </a> tag, so we can highlight the text up to
# the </a>.  We also need to remember which window the destination occurred in.

proc HMtag_a {state win param text} {
	upvar #0 $state var
	upvar #0 HN$win winvar

	# Check for missing end tag on previous anchor
	if {[info exists var(Tref)] || [info exists var(Tname)]} {
		# Synthesize the end tag
		HMtag_/a $state $win {} {}
	}

	# a source

	if {[HMextract_param $param href]} {
		# check the relation
		set rel {}	;# default is hypertext link
		HMextract_param $param rel
		if {[string tolower $rel] == "embed"} {
			# assume it's an applet
			HMload_applet $state $win $href hyperpage
		} else {
			set var(Tref) [list L:$href]
			HMstack $state "" "Tlink link"
			HMlink_setup $state $win $href
		}
	}

	# a destination

	if {[HMextract_param $param name]} {
		set var(Tname) [list N:$name]
		set var(N:$name) $win
		HMstack $state "" "Tanchor anchor"
		$win mark set N:$name "$var(S_insert) - 1 chars"
		$win mark gravity N:$name left
		if {[info exists var(goto)] && $var(goto) == $name} {
			unset var(goto)
			set var(going) $name
		}
	}
}

# The application should call here with the fragment name
# to cause the display to go to this spot.
# If the target exists, go there (and do the callback),
# otherwise schedule the goto to happen when we see the reference.
# In this case win is the main text window

proc HMgoto {state where {callback HMwent_to}} {
	upvar #0 $state var
	if {[info exists var(N:$where)]} {
		$var(N:$where) see N:$where
		update
		eval $callback $var(N:$where) [list $where]
		return 1
	} else {
		set var(goto) $where
		return 0
	}
}

# We actually got to the spot, so highlight it!
# This should/could be replaced by the application
# We'll flash it orange a couple of times.
# win is the window in which the element occurred, but
# this implementation has problems where the element
# is split over multiple windows.

proc HMwent_to {win where {count 0} {color orange}} {
	if {$count > 5} return
	catch {$win tag configure N:$where -foreground $color}
	update
	after 200 [list HMwent_to $win $where [incr count]  [expr {$color=="orange" ? "" : "orange"}]]
}

proc HMtag_/a {state win param text} {
	upvar #0 $state var
	if {[info exists var(Tref)]} {
		unset var(Tref)
		HMstack $state / "Tlink link"
	}

	# goto this link, then invoke the call-back.

	if {[info exists var(going)]} {
		$win yview N:$var(going)
		update
		HMwent_to $win $var(going)
		unset var(going)
	}

	if {[info exists var(Tname)]} {
		unset var(Tname)
		HMstack $state / "Tanchor anchor"
	}
}

#           Inline Images
# This interface is subject to change
# Most of the work is getting around a limitation of TK that prevents
# setting the size of a label to a widthxheight in pixels
#
# Images have the following parameters:
#    align:  top,middle,bottom
#    alt:    alternate text
#    ismap:  A clickable image map
#    src:    The URL link
# Netscape supports (and so do we)
#    width:  A width hint (in pixels)
#    height:  A height hint (in pixels)
#    border: The size of the window border

proc HMtag_img {state win param text} {
	global HMinsert_map HMalign_map
	upvar #0 $state var
	upvar #0 HN$win winvar

	set align [HMget_alignment $param bottom]		;# The spec isn't clear what the default should be
	set imgalign {}
	if {$align == "left"} {
		# Force a line break, text flows to right of graphic
		$win insert $var(S_insert) $HMinsert_map(br) "space $winvar(font)"
		set align bottom
		set imgalign left
	} elseif {$align == "right"} {
		# Right-align the graphic, text flows around to the left (not currently possible)
		HMstack $state "" "$HMalign_map(right)"
		set align bottom
		set imgalign right
	}

	# get alternate text
	set alt "<image>"
	HMextract_param $param alt
	set alt [HMmap_esc $alt]

	# get the border width
	set border 1
	HMextract_param $param border

	set tags [HMcurrent_tags $state $win]
	set link {}
	if {[set link [lindex $tags [lsearch -glob $tags L:*]]] != {}} {
		regsub L: $link {} link
		# This image is within an anchor
		set c button
		set sourceanchor 1
	} else {
		set c label
		set sourceanchor 0
	}

	# see if we have an image size hint
	# If so, make a frame the "hint" size to put the label in
	# otherwise just make the label
	set item $win.[HMgetUniqueNum $state]
	# catch {destroy $item}
	if {[HMextract_param $param width] && [HMextract_param $param height]} {
		frame $item -width $width -height $height -bd 0
		pack propagate $item 0
		set label $item.label
		$c $label
		pack $label -expand 1 -fill both
	} else {
		set label $item
		$c $label 
	}

	$label configure -fg orange -text $alt -bg $var(bgcolor) -highlightthickness 0
	catch {$label configure -bd $border} err
	$win window create $var(S_insert) -align $align -window $item -pady 2 -padx 2

	# add in all the current tags (this is overkill)
	foreach tag $tags {
		$win tag add $tag $item
	}

	# Pop the right-align tag if necessary
	if {$imgalign == "right"} {
		catch {HMstack $state / "$HMalign_map(right)"}
	}

	# set imagemap callbacks
	if {[HMextract_param $param ismap]} {
		global HMevents
		regsub -all {%} $link {%%} link2
		foreach i [array names HMevents] {
			bind $label <$i> "catch \{%W configure $HMevents($i)\}"
		}
		bind $label <ButtonRelease-1> "+HMlink_callback $state $win $link2?%x,%y"
	} elseif {$sourceanchor} {
		# The image will be in a button
		$label configure -command "HMlink_callback $state $win $link"
		# Callback to the application
		HMlink_window_setup $label $link
	}

	# now callback to the application
	set src ""
	HMextract_param $param src
	HMset_image $state $win $label $src
	return $label	;# used by the forms package for input_image types
}

proc HMget_alignment {param {default left}} {
	array set align_map {left left	top top    middle center    bottom baseline}
	set align $default
	HMextract_param $param align
	catch {set align $align_map([string tolower $align])}
	return $align
}

# The app needs to supply one of these
proc HMset_image {state win handle src} {
	HMgot_image $handle "can't get\n$src"
}
# and one of these
proc HMlink_window_setup {b link} {
	puts "link \"$link\" for button $b"
}

# When the image is available, the application should call back here.
# If we have the image, put it in the label, otherwise display the error
# message.  If we don't get a callback, the "alt" text remains.
# if we have a clickable image, arrange for a callback

proc HMgot_image {win image_error} {
	# if we're in a frame turn on geometry propogation
	if {[winfo name $win] == "label"} {
		pack propagate [winfo parent $win] 1
	}
	if {[catch {$win configure -image $image_error}]} {
		$win configure -image {}
		$win configure -text $image_error
	}
}

# Sample hypertext link callback routine - should be replaced by app
# This proc is called once for each <A> tag.
# Applications can overwrite this procedure, as required, or
# replace the HMevents array
#   win:   The name of the text widget to render into
#   href:  The HREF link for this <a> tag.

array set HMevents {
	Enter	{-borderwidth 2 -relief raised }
	Leave	{-borderwidth 2 -relief flat }
	1		{-borderwidth 2 -relief sunken}
	ButtonRelease-1	{-borderwidth 2 -relief raised}
}

# We need to escape any %'s in the href tag name so the bind command
# doesn't try to substitute them.

proc HMlink_setup {state win href} {
	global HMevents
	regsub -all {%} $href {%%} href2
	foreach i [array names HMevents] {
		eval {$win tag bind  L:$href <$i>}  \{$win tag configure \{L:$href2\} $HMevents($i)\}

#		eval $win tag bind L:$href <$i>  #			\{$win tag configure \{L:$href2\} $HMevents($i)\}
#		$win tag bind L:$href <$i>  #			"$win tag configure \{L:$href2\} $HMevents($i)"
	}
}

# generic link-hit callback
# This gets called upon button hits on hypertext links
# Applications are expected to supply ther own HMlink_callback routine
#   state: The global array for the hyperwindow
#   win:   The name of the text widget to render into
#   x,y:   The cursor position at the "click"

proc HMlink_hit {state win x y} {
	set tags [$win tag names @$x,$y]
	set link [lindex $tags [lsearch -glob $tags L:*]]
	# regsub -all {[^L]*L:([^ ]*).*}  $tags {\1} link
	regsub L: $link {} link
	if {$link == {}} return;
	HMlink_callback $state $win $link
}

# replace this!
#   win:   The name of the text widget to render into
#   href:  The HREF link for this <a> tag.

proc HMlink_callback {state win href} {
	puts "Got hit on $win, link $href"
}

#
# Table handling
#
# Each table cell is rendered in its own Text widget which has its own,
# separate state.  A stack is maintained of nested tables (the window stack
# is not sufficient).  The Text widget cannot size itself properly, so
# once the table has been finished the cells are resized.

# Simple port to the grid geometry manager from blt_table

proc HMtag_table {state win param text} {
	debug_puts [list HMtag_table $state $win $param $text]
	catch {HMrealtag_table $state $win $param $text} err
	debug_puts [list HMrealtag_table returned $err]
}

proc HMrealtag_table {state win param text} {
	upvar #0 $state var
	upvar #0 HN$win winvar

	if {[HMextract_param $param border]} {
		if {[info exists border] && $border == "0"} {
			set bdopt {}
		} else {set bdopt {-borderwidth 2 -relief groove}}
	} else {set bdopt {}}

	# Create a frame for sizing the table
	set tableframe $win.table[HMgetUniqueNum $state]
	catch {destroy $tableframe}
	eval frame $tableframe $bdopt
	HMwin_install $state $win $tableframe

	# Create frame for containing the table and push onto table stack
	set table [frame $tableframe.f -borderwidth 0 -background $var(bgcolor)]
	set parent [lindex $var(tables) end]
	lappend var(tables) $table
	upvar #0 HN$table tblvar
	pack $table
	pack propagate $tableframe 0

	if {[HMextract_param $param cols]} {
		# Use fixed layout algorithm.  Columns get equal width by default
		array set tblvar [list columns $cols layout fixed]
		for {set c 0} {$c < $cols} {incr c} {
			grid columnconfigure $table $c -weight 1
		}
	} else {set tblvar(layout) auto}

	# Initialise row and column counters, and other variables
	array set tblvar [list parent $parent rowcntr 1 colcntr 0 colspec 0 occluded {} in_table_cell 0 bdopt $bdopt]

	# Set width as specified
	set width 100%
	HMextract_param $param width
	set tblvar(tablewidth) $width
	if {[regexp {(.+)%} $width all w]} {
		# BUG: doesn't take right margin into account
		set winwidth [expr [winfo width $win] - 2 * [$win cget -borderwidth] - 4 - [winfo pixels $win $winvar(level)c]]
		debug_puts [list width specified as percentage - $w]
		debug_puts [list avail width $winwidth, set $tableframe width to [expr $winwidth * $w / 100] pixels]
		$tableframe configure -width [expr $winwidth * $w / 100]
	} else {
		debug_puts [list width specified as $width]
		catch {$tableframe configure -width [HMtable_stdSize $width]}
	}
	pack configure $table -fill both -expand true
}

# Column groups not yet implemented
proc HMtag_colgroup {state win param text} {}

proc HMtag_col {state win param text} {
	debug_puts [list HMtag_col $state $win $param $text]
	upvar #0 $state var
	upvar #0 HN[set table [lindex $var(tables) end]] tblvar

	set span 1
	HMextract_param $param span

	if {[HMextract_param $param width]} {
		if {[regexp {(.*)\*$} $width all val]} {
			set opt -weight
			set cmd2 HMnoop; set opt2 {}
		} else {
			set opt -minsize 
			set val [HMtable_stdSize $val]
			set cmd2 {grid columnconfigure}
			set opt2 [list -weight 0]
		}
		for {} {$span} {incr tblvar(colspec); incr span -1} {
			eval $cmd2 $table $tblvar(colspec) $opt2
			grid columnconfigure $table $tblvar(colspec) $opt $val
		}
	} else {incr tblvar(colspec) $span}
}

# Convert HTML standard sizes to Tk screen measurements
# NB. Tk <= 4.2 does not provide a convenient way to get the default height
# of a font, so we assume 12 pixels

proc HMtable_stdSize {size} {
	regexp {([0-9]+)([^0-9]*)} $size all measure units
	switch $units {
		pt {return [expr $measure / 72.27]i}
		pi {return [expr $measure / (72.27 * 12)]i}
		in {return ${measure}i}
		cm {return ${measure}c}
		mm {return ${measure}m}
		em {return [expr $measure * 12]}
		px -
		default {return $measure}
	}
}

proc HMtag_/table {state win param text} {
	debug_puts [list HMtag_/table $state $win $param $text]
	upvar #0 $state var
	upvar #0 HN[set table [lindex $var(tables) end]] tblvar

	# If the previous cell is still "open" then close it
	if {$tblvar(in_table_cell)} {
		HMtag_/td $state $win $param $text
	}

	# Make the caption span all columns
	lassign [grid size $table] cols rows
	if {[winfo exists $table.caption]} {
		grid configure $table.caption -columnspan $cols
	}

	# Pop table stack
	set var(tables) [lrange $var(tables) 0 [expr [llength $var(tables)] - 2]]

	# When the entire table has been rendered, fix up the cell geometries
	if {$tblvar(layout) == "auto"} {
		# Initialise columns to equal weighting
		for {set i $tblvar(colspec)} {$i < $cols} {incr i} {
			grid columnconfigure $table $i -weight 1
		}
	}
	if {$var(tables) == {}} {HMtable_setup_geom $state $table}

	# Can't remove this yet, since resizing code uses it
	#unset tblvar
}

proc HMtable_setup_geom {state table} {
	debug_puts [list HMtable_setup_geom $state $table]
	foreach cell [winfo children $table] {
		HMcell_setupGeom $state $cell.c $table
		regsub -all "($cell\\.c\\.\[^t\]\[^ \]*)" [winfo children $cell.c] {} tables
		set tables [string trim $tables]
		foreach leaf $tables {HMtable_setup_nested $state $leaf.f}
	}
}

# Set table width according to available space for nested tables

proc HMtable_setup_nested {state table} {
	debug_puts [list HMtable_setup_nested $state $table]
	catch {HMrealtable_setup_nested $state $table} err
	debug_puts [list HMrealtable_setup_nested returned $err]
}

proc HMrealtable_setup_nested {state table} {
	upvar #0 HN$table tblvar
	update	;# Necessary to make 'winfo width' accurate
	set tableframe [winfo parent $table]
	if {[regexp {(.+)%} $tblvar(tablewidth) all w]} {
		# BUG: doesn't take margins into account
		set win [winfo parent [winfo parent $table]]
		set winwidth [expr [winfo width $win] - 2 * [$win cget -borderwidth] - 4]
		debug_puts [list width specified as percentage - $w]
		debug_puts [list $win avail width $winwidth, set $tableframe width to [expr $winwidth * $w / 100] pixels]
		$tableframe configure -width [expr $winwidth * $w / 100]
	} else {
		debug_puts [list width specified as $width]
		catch {$tableframe configure -width [HMtable_stdSize $width]}
	}

	HMtable_setup_geom $state $table
}

proc HMcell_setupGeom {state win table} {
	debug_puts [list HMcell_setupGeom $state $win $table]
	catch {HMrealcell_setupGeom $state $win $table} err
	debug_puts [list HMrealcell_setupGeom returned $err]
}

proc HMrealcell_setupGeom {state win table} {
	upvar #0 $state var
	upvar #0 HN$win winvar
	upvar #0 HN$table tblvar

	set resize "
	    HMtable_setGeom [list $state] [list $table]
	    UTIL_resetFitText [list $win]
	    UTIL_fitTextVertical [list $win] 1 [list [list HMtable_setGeom $state $table]]
	"

	if {$tblvar(layout) == "fixed"} {
		debug_puts [list fixed layout for table $table]
	} else {
		debug_puts [list auto layout for table $table]

	        # Ignore this cell if it spans multiple columns
	        array set gridopts [grid info [winfo parent $win]]
	        if {$gridopts(-columnspan) > 1} {
		    eval $resize
		    return
	        }

		# Work out min/max values
		# Assume that a character is 8 pixels wide
		set min 0
		set max 0; set line 0
		foreach {key value index} [$win dump -text -window 1.0 end] {
			switch -- $key {
				text {
					if {$value == "\n"} {
						set max [max $max $line]
						set line 0
					} else {
					        regsub -all {("|{|}|\\)} $value X value
						incr line [expr [string length $value] * 8]
						foreach word $value {
							set min [max $min [expr [string length $word] * 8]]
						}
					}
				}
				window {
					# Ignore nested tables in calc.
					if {![regexp {table.*} [winfo name [winfo parent $value]]]} {
						set min [max $min [winfo reqwidth $value]]
						incr line [winfo reqwidth $value]
					}
				}
			}
		}
		set max [max $max $line]	;# catch the last line
		debug_puts [list min is $min, max is $max]
		regexp {cell_([0-9]+)_([0-9]+)} [winfo name [winfo parent $win]] all row col
		debug_puts [list window $win is in column $col]
		if {$min > [grid columnconfigure $table $col -minsize]} {
			debug_puts [list setting column $col minsize to $min]
			grid columnconfigure $table $col -minsize $min
		}
		if {$min > 0 && [expr double($max) / double($min)] > [grid columnconfigure $table $col -weight]} {
			debug_puts [list setting column $col weighting to [expr double($max) / double($min)]]
			grid columnconfigure $table $col -weight [expr $max / $min]
		}
	}
	eval $resize
}

proc HMtable_setGeom {state table} {
	debug_puts [list HMtable_setGeom $state $table]
	upvar #0 HN$table tblvar

	# Set the height appropriately for this cell
	[winfo parent $table] configure -height [winfo reqheight $table]
	debug_puts [list [winfo parent $table] configure -height [winfo reqheight $table]]
}

# The caption is rendered into its own text widget, which is then considered
# to be on its own row

proc HMtag_caption {state win param text} {
	debug_puts [list HMtag_caption $state $win $param $text]
	upvar #0 $state var
	upvar #0 HN[set table [lindex $var(tables) end]] tblvar

	set caption $table.caption
	catch {destroy $caption}
	eval text $caption -takefocus 0 -width 1 -height 1 $tblvar(bdopt) -highlightthickness 0

	# Enter into the table
	grid $caption -row 0 -sticky news

	# push caption onto the state
	HMpush_win $state $caption
}

proc HMtag_/caption {state win param text} {
	debug_puts [list HMtag_/caption $state $win $param $text]
	$win configure -wrap none
	UTIL_resetFitText $win
	UTIL_fitTextVertical $win 1 {}

	# Pop state
	HMpop_win $state
}

proc HMtag_tr {state win param text} {
	upvar #0 $state var
	upvar #0 HN[lindex $var(tables) end] tblvar

	# If the previous cell is still "open" then close it
	if {$tblvar(in_table_cell)} {
		HMtag_/td $state $win $param $text
	}

	# Increment the row counter
	incr tblvar(rowcntr)
	# Reset the column counter
	set tblvar(colcntr) 0
}

proc HMtag_th {state win param text} {
	global HMtag_map
	HMtable_cell $state $win $param $text
	catch {HMstack $state $cell {} "$HMtag_map(center)"}
}

proc HMtag_td {state win param text} {
	HMtable_cell $state $win $param $text
}

proc HMtable_cell {state win param text} {
	upvar #0 $state var
	upvar #0 HN[set table [lindex $var(tables) end]] tblvar
	debug_puts [list HMtable_cell $state $win $param $text in table $table]

	# If the previous cell is still "open" then close it
	if {$tblvar(in_table_cell)} {
		HMtag_/td $state $win $param $text
	}
	set tblvar(in_table_cell) 1

	# Determine various parameters
	set rowspan 1
	HMextract_param $param rowspan
	set colspan 1
	HMextract_param $param colspan

	# Put the cell in a frame to control size pixel-wise
	set cell $table.cell_$tblvar(rowcntr)_$tblvar(colcntr)
	catch {destroy $cell}
	eval frame $cell $tblvar(bdopt)
	eval text $cell.c -takefocus 0 -width 1 -height 1 -borderwidth 0 -highlightthickness 0
	pack $cell.c -fill both -expand 1

	# Enter into the table
	grid $cell -row $tblvar(rowcntr) -column $tblvar(colcntr)  -sticky news -rowspan $rowspan -columnspan $colspan

	# Advance the column counter
	incr tblvar(colcntr) $colspan

	# push cell onto the state
	HMpush_win $state $cell.c
	# Initialise it
	HMrender $var(basewin) hmstart {} {} {}
}

# there isn't really a </TD> tag!

# The cell's width is fixed after its height has been
# determined.

proc HMtag_/td {state win param text} {
	debug_puts [list HMtag_/td $state $win $param $text]
	upvar #0 $state var
	upvar #0 HN[set table [lindex $var(tables) end]] tblvar

	HMrender $var(basewin) hmstart / {} {}

	# Pop window stack
	HMpop_win $state
	# Popping window stack does not pop table stack
	set tblvar(in_table_cell) 0
}

# extract a value from parameter list (this needs a re-do)
# returns "1" if the keyword is found, "0" otherwise
#   param:  A parameter list.  It should alredy have been processed to
#           remove any entity references
#   key:    The parameter name
#   val:    The variable to put the value into (use key as default)

proc HMextract_param {param key {val ""}} {

	if {$val == ""} {
		upvar $key result
	} else {
		upvar $val result
	}
    set ws "    \n\r"
 
    # look for name=value combinations.  Either (') or (") are valid delimeters
    if {
      [regsub -nocase [format {.*%s[%s]*=[%s]*"([^"]*).*} $key $ws $ws] $param {\1} value] ||
      [regsub -nocase [format {.*%s[%s]*=[%s]*'([^']*).*} $key $ws $ws] $param {\1} value] ||
      [regsub -nocase [format {.*%s[%s]*=[%s]*([^%s]+).*} $key $ws $ws $ws] $param {\1} value] } {
        set result $value
        return 1
    }

	# now look for valueless names
	# I should strip out name=value pairs, so we don't end up with "name"
	# inside the "value" part of some other key word - some day
	
	set bad \[^a-zA-Z\]+
	if {[regexp -nocase  "$bad$key$bad" -$param-]} {
		return 1
	} else {
		return 0
	}
}

# These next three routines manage the display state of the page.

# Push or pop tags to/from stack.
# Each orthogonal text property has its own stack, stored as a list.
# The current (most recent) tag is the last item on the list.
# Push is {} for pushing and {/} for popping
# Properties may belong to either the hyperwindow or to a window.

proc HMstack {state push list} {
	upvar #0 $state var
	array set tags $list
	if {$push == ""} {
		foreach tag [array names tags] {
			lappend var($tag) $tags($tag)
		}
	} else {
		foreach tag [array names tags] {
			# set cnt [regsub { *[^ ]+$} $var($tag) {} var($tag)]
			set var($tag) [lreplace $var($tag) end end]
		}
	}
}

# extract set of current text tags
# tags starting with T map directly to text tags, all others are
# handled specially.  There is an application callback, HMset_font
# to allow the application to do font error handling
# This is complicated by some tags being managed in the hyperwindow
# state, and some in the window state

proc HMcurrent_tags {state win} {
	upvar #0 $state var
	upvar #0 HN$win winvar

	set font font
	foreach i {family size weight style} {
		set $i [lindex $var($i) end]
		append font :[set $i]
	}
	set xfont [HMx_font $family $size $weight $style $var(S_adjust_size)]
	HMset_font $win $font $xfont
	set indent [llength $winvar(indent)]
	#incr indent -1
	lappend tags $font indent$indent
	foreach tag [array names var T*] {
		lappend tags [lindex $var($tag) end]	;# test
	}
	foreach tag [array names winvar T*] {
		lappend tags [lindex $winvar($tag) end]
	}
	set winvar(font) $font
	set winvar(xfont) [$win tag cget $font -font]
	set winvar(level) $indent
	return $tags
}

# allow the application to do do better font management
# by overriding this procedure

proc HMset_font {win tag font} {
	catch {$win tag configure $tag -font $font; $win tag lower $tag mark}
	catch {$win tag configure $tag -kanjifont [HMfont2kfont $font]}
}

# generate an X font name
proc HMx_font {family size weight style {adjust_size 0}} {
	catch {incr size $adjust_size}
	return "-*-$family-$weight-$style-normal-*-*-${size}0-*-*-*-*-*-*"
}

proc HMfont2kfont {font} {
    set jis_size [lindex [split $font {-}] 8]
    incr jis_size -10
    return "-*-*-medium-r-normal-*-*-$jis_size-75-75-*-*-jisx0208.*-*"
}

# Optimize HMrender (hee hee)
# This is experimental

proc HMoptimize {} {
	regsub -all "\n\[ 	\]*#\[^\n\]*" [info body HMrender] {} body
	regsub -all ";\[ 	\]*#\[^\n]*" $body {} body
	regsub -all "\n\n+" $body \n body
	proc HMrender {cookie tag not param text} $body
}
############################################
# Turn HTML into TCL commands
#   html    A string containing an html document
#   incomplete	A variable to store leftover HTML
#   cmd		A command to run for each html tag found
#   start	The name of the dummy html start/stop tags

# The routine populates the given array with the HTML elements, indexed by a 
# number.  Retrieving elements from an array will be faster than a list.
# A sentinel value, "\win\" is included as a placeholder for application-
# specific arguments.
# The routine returns the last index used.

proc HMparse_html {html parsedArray startElement numTags incomplete {cmd HMtest_parse} {start hmstart}} {
	upvar #0 $parsedArray var
	upvar $incomplete incvar

	regsub -all \{ $html {\&ob;} html
	regsub -all \} $html {\&cb;} html
	regsub -all \\\\ $html {\&bsl;} html
	set w " \t\r\n"	;# white space
	proc HMcl x {return "\[$x\]"}
	set exp <(/?)([HMcl ^$w>]+)[HMcl $w]*([HMcl ^>]*)>
	set sub "\}\n[list $cmd] {\\2} {\\1} {\\3} \{"
	regsub -all $exp $html $sub html
	# Check for incomplete final tag
	set incvar {}
	if {[regexp {[^<]*(<.*)} [lindex "\{$html\}" end] all incvar]} {
		set html [string range $html 0 [expr [string last "<" $html] - 1]]
	}

	for {set i 0; set group {}} {$i < $numTags} {incr i} {
		append group "a$i c$i d$i e$i f$i\n"
		regsub -all {(a[0-9]+)} $group {$\1 \\\\win\\\\} subgroup
		regsub -all {([b-z]+[0-9]+)} $subgroup {{$\1}} subgroup
	}
	foreach $group "[list $cmd] {$start} {} {} \{ $html \}" {
		set var([incr startElement]) [subst $subgroup]
	}
	return $startElement
}

# Just consume arguments and do nothing
proc HMnoop {args} {}

proc HMtest_parse {command tag slash text_after_tag} {
	puts "==> $command $tag $slash $text_after_tag"
}

# Convert multiple white space into a single space

proc HMzap_white {data} {
	regsub -all "\[ \t\r\n\]+" $data " " data
	return $data
}

proc HMtranslate_cr {data} {
	regsub -all "\r+" $data "\n" data
	return $data
}

# find HTML escape characters of the form &xxx;

proc HMmap_esc {text} {
	if {![regexp & $text]} {return $text}
	regsub -all {([][$\\])} $text {\\\1} new
	regsub -all {&#([0-9][0-9]?[0-9]?);?}  $new {[format %c [scan \1 %d tmp;set tmp]]} new
	regsub -all {&([a-zA-Z]+);?} $new {[HMdo_map \1]} new
	return [subst $new]
}

# convert an HTML escape sequence into character

proc HMdo_map {text {unknown ?}} {
	global HMesc_map
	set result $unknown
	catch {set result $HMesc_map($text)}
	return $result
}

# table of escape characters (ISO latin-1 esc's are in a different table)

array set HMesc_map {
   lt <   gt >   amp &   quot \"   copy \xa9
   reg \xae   ob \x7b   cb \x7d   nbsp \xa0
   bsl \\
}
#############################################################
# ISO Latin-1 escape codes

array set HMesc_map {
	nbsp \xa0 iexcl \xa1 cent \xa2 pound \xa3 curren \xa4
	yen \xa5 brvbar \xa6 sect \xa7 uml \xa8 copy \xa9
	ordf \xaa laquo \xab not \xac shy \xad reg \xae
	hibar \xaf deg \xb0 plusmn \xb1 sup2 \xb2 sup3 \xb3
	acute \xb4 micro \xb5 para \xb6 middot \xb7 cedil \xb8
	sup1 \xb9 ordm \xba raquo \xbb frac14 \xbc frac12 \xbd
	frac34 \xbe iquest \xbf Agrave \xc0 Aacute \xc1 Acirc \xc2
	Atilde \xc3 Auml \xc4 Aring \xc5 AElig \xc6 Ccedil \xc7
	Egrave \xc8 Eacute \xc9 Ecirc \xca Euml \xcb Igrave \xcc
	Iacute \xcd Icirc \xce Iuml \xcf ETH \xd0 Ntilde \xd1
	Ograve \xd2 Oacute \xd3 Ocirc \xd4 Otilde \xd5 Ouml \xd6
	times \xd7 Oslash \xd8 Ugrave \xd9 Uacute \xda Ucirc \xdb
	Uuml \xdc Yacute \xdd THORN \xde szlig \xdf agrave \xe0
	aacute \xe1 acirc \xe2 atilde \xe3 auml \xe4 aring \xe5
	aelig \xe6 ccedil \xe7 egrave \xe8 eacute \xe9 ecirc \xea
	euml \xeb igrave \xec iacute \xed icirc \xee iuml \xef
	eth \xf0 ntilde \xf1 ograve \xf2 oacute \xf3 ocirc \xf4
	otilde \xf5 ouml \xf6 divide \xf7 oslash \xf8 ugrave \xf9
	uacute \xfa ucirc \xfb uuml \xfc yacute \xfd thorn \xfe
	yuml \xff
}

##########################################################
# html forms management commands

# As each form element is located, it is created and rendered.  Additional
# state is stored in a form specific global variable to be processed at
# the end of the form, including the "reset" and "submit" options.
# Remember, there can be multiple forms existing on multiple pages.  When
# HTML tables are added, a single form could be spread out over multiple
# text widgets, which makes it impractical to hang the form state off the
# HN$win structure.  We don't need to check for the existance of required
# parameters, we just "fail" and get caught in HMrender

# An applet may be attached to the form.  The applet is given the starting
# index of the form.  The form_id is stored for the applet (but not visible
# to the applet).  The applet is given the widget pathname of each element
# in the form, and is informed when the </form> is seen.  There are callins
# after the form is reset and just before it is submitted.

# This causes line breaks to be preserved in the inital values
# of text areas
array set HMwintag_map {
	textarea    {fill 0}
}

##########################################################
# html isindex tag.  Although not strictly forms, they're close enough
# to be in this file

# is-index forms
# make a frame with a label, entry, and submit button

proc HMtag_isindex {state win param text} {
	upvar #0 $state var
	upvar #0 HN$win winvar

	set item $win.[HMgetUniqueNum $state]
	if {[winfo exists $item]} {
		destroy $item
	}
	frame $item -relief ridge -bd 3 -background $var(bgcolor)
	set prompt "Enter search keywords here"
	HMextract_param $param prompt
	label $item.label -text [HMmap_esc $prompt] -font $winvar(xfont)  -background $var(bgcolor) -foreground $var(fgcolor)
	entry $item.entry -background $var(bgcolor) -foreground $var(fgcolor)
	bind $item.entry <Return> "$item.submit invoke"
	button $item.submit -text search -font $winvar(xfont) -command  [format {HMsubmit_index %s %s {%s} [HMmap_reply [%s get]]}  $state $win $param $item.entry]  -background $var(bgcolor) -foreground $var(fgcolor)
	pack $item.label -side top
	pack $item.entry $item.submit -side left

	# insert window into text widget

	$win insert $var(S_insert) \n isindex
	HMwin_install $state $win $item
	$win insert $var(S_insert) \n isindex
	bind $item <Visibility> {focus %W.entry}
}

# This is called when the isindex form is submitted.
# The default version calls HMlink_callback.  Isindex tags should either
# be deprecated, or fully supported (e.g. they need an href parameter)

proc HMsubmit_index {state win param text} {
	HMlink_callback $state $win ?$text
}

# initialize form state.  All of the state for this form is kept
# in a global array whose name is stored in the form_id field of
# the main window array.
# Parameters: ACTION, METHOD, ENCTYPE

proc HMtag_form {state win param text} {
	upvar #0 $state var

	# create a global array for the form
	set id HM$win.form[HMgetUniqueNum $state]
	upvar #0 $id form

	# missing /form tag, simulate it
	if {[info exists var(form_id)]} {
		# puts "Missing end-form tag !!!! $var(form_id)"
		HMtag_/form $state $win {} {}
	}
	catch {unset form}
	set var(form_id) $id

	set form(param) $param		;# form initial parameter list
	set form(reset) ""			;# command to reset the form
	set form(reset_button) ""	;# list of all reset buttons
	set form(submit) ""			;# command to submit the form
	set form(submit_button) ""	;# list of all submit buttons

	# If the SCRIPT attribute is specified, load the applet and attach
	# it to the form

	if {[HMextract_param $param script]} {
		HMload_applet $state $win $script form $id
	}
}

# Application should override this procedure
proc HMload_applet {state win script level {formID {}}} {
	puts "==> HMload_applet $state $win $script $level $formID"
}

# Where we're done try to get all of the state into the widgets so
# we can free up the form structure here.  Unfortunately, we can't!

proc HMtag_/form {state win param text} {
	upvar #0 $state var
	upvar #0 $var(form_id) form

	# make submit button entries for all radio buttons
	foreach name [array names form radio_*] {
		regsub radio_ $name {} name
		lappend form(submit) [list $name \$form(radio_$name)]
	}

	# process the reset button(s)

	append form(reset) "; HMreset_form $state $win $var(form_id)"	;# Allow applet callins
	foreach item $form(reset_button) {
		$item configure -command $form(reset)
	}

	# no submit button - add one
	if {$form(submit_button) == ""} {
		HMinput_submit $state $win {}
	}

	# process the "submit" command(s)
	# each submit button could have its own name,value pair

	foreach item $form(submit_button) {
		set submit $form(submit)
		catch {lappend submit $form(submit_$item)}
		$item configure -command   [list HMsubmit_button $state $win $var(form_id) $form(param)  $submit]
	}

	# Callin to the applet to let it know we're done.
	HMapplet_/form $state $win $var(form_id)

	# unset all unused fields here
	unset form(reset) form(submit) form(reset_button) form(submit_button)
	unset var(form_id)
}

# Application should override this
proc HMapplet_/form {state win form_id} {
	puts "==> end of form, state $state, window $win, form $form_id,  at index [$win index {end - 1 char}]"
}

###################################################################
# handle form input items
# each item type is handled in a separate procedure
# Each "type" procedure needs to:
# - create the window
# - initialize it
# - add the "submit" and "reset" commands onto the proper Q's
#   "submit" is subst'd
#   "reset" is eval'd
# - return the name of the new window

proc HMtag_input {state win param text} {
	upvar #0 $state var

	set type text	;# the default
	HMextract_param $param type
	set type [string tolower $type]
	set name {}; HMextract_param $param name
	set value {}; HMextract_param $param value
	# Pass name & value to proc to avoid need to reparse attributes
	if {[catch {HMinput_$type $state $win $param} err]} {
		#puts stderr $err
	} else {
		# Inform the applet of the new window
		HMapplet_item $state $win $var(form_id) $type $name $value $err
	}
}

# input type=text
# parameters NAME (reqd), MAXLENGTH, SIZE, VALUE

proc HMinput_text {state win param {show {}}} {
	upvar #0 $state var
	upvar #0 $var(form_id) form

	# make the entry
	HMextract_param $param name		;# required
	set item $win.input_text,[HMgetUniqueNum $state]
	set size 20; HMextract_param $param size
	set maxlength 0; HMextract_param $param maxlength
	entry $item -width $size -show $show -background $var(bgcolor) -foreground $var(fgcolor)
	catch {$item configure -kanjifont k14 -font a14; bind $item <Tab> [list kanjInput end $item]}

	# set the initial value
	set value ""; HMextract_param $param value
	$item insert 0 $value
		
	# insert the entry
	HMwin_install $state $win $item

	# set the "reset" and "submit" commands
	append form(reset) ";$item delete 0 end;$item insert 0 [list $value]"
	lappend form(submit) [list $name "\[$item get]"]

	# handle the maximum length (broken - no way to cleanup bindtags state)
	if {$maxlength} {
		bindtags $item "[bindtags $item] max$maxlength"
		bind max$maxlength <KeyPress> "%W delete $maxlength end"
	}

	return $item
}

# password fields - same as text, only don't show data
# parameters NAME (reqd), MAXLENGTH, SIZE, VALUE

proc HMinput_password {state win param} {
	HMinput_text $state $win $param *
}

# checkbuttons are missing a "get" option, so we must use a global
# variable to store the value.
# Parameters NAME, VALUE, (reqd), CHECKED

proc HMinput_checkbox {state win param} {
	upvar #0 $state var
	upvar #0 $var(form_id) form

	HMextract_param $param name
	HMextract_param $param value

	# Set the global variable, don't use the "form" alias as it is not
	# defined in the global scope of the button
	set uniqnum [HMgetUniqueNum $state]
	set variable $var(form_id)(check_$uniqnum)	
	set item $win.input_checkbutton,$uniqnum
	checkbutton $item -variable $variable -off {} -on $value -text "  "  -background $var(bgcolor) -foreground $var(fgcolor)
	if {[HMextract_param $param checked]} {
		$item select
		append form(reset) ";$item select"
	} else {
		append form(reset) ";$item deselect"
	}

	HMwin_install $state $win $item
	lappend form(submit) [list $name \$form(check_$uniqnum)]
	return $item
}

# radio buttons.  These are like check buttons, but only one can be selected

proc HMinput_radio {state win param} {
	upvar #0 $state var
	upvar #0 $var(form_id) form

	HMextract_param $param name
	HMextract_param $param value

	set first [expr ![info exists form(radio_$name)]]
	set variable $var(form_id)(radio_$name)
	set variable $var(form_id)(radio_$name)
	set item $win.input_radiobutton,[HMgetUniqueNum $state]
	radiobutton $item -variable $variable -value $value -text " " -background $var(bgcolor) -foreground $var(fgcolor)

	HMwin_install $state $win $item

	if {$first || [HMextract_param $param checked]} {
		$item select
		append form(reset) ";$item select"
	} else {
		append form(reset) ";$item deselect"
	}

	# do the "submit" actions in /form so we only end up with 1 per button grouping
	# contributing to the submission

	return $item
}

# hidden fields, just append to the "submit" data
# params: NAME, VALUE (reqd)

proc HMinput_hidden {state win param} {
	upvar #0 $state var
	upvar #0 $var(form_id) form
	HMextract_param $param name
	set value ""; HMextract_param $param value
	lappend form(submit) [list $name $value]
	return {}
}

# handle input images.  The spec isn't very clear on these, so I'm not
# sure its quite right
# Use std image tag, only set up our own callbacks
#  (e.g. make sure ismap isn't set)
# params: NAME, SRC (reqd) ALIGN

proc HMinput_image {state win param} {
	upvar #0 $state var
	upvar #0 $var(form_id) form
	HMextract_param $param name
	set name		;# barf if no name is specified
	set item [HMtag_img $state $win $param {}]
	$item configure -relief raised -bd 2 -bg blue

	# make a dummy "submit" button, and invoke it to send the form.
	# We have to get the %x,%y in the value somehow, so calculate it during
	# binding, and save it in the form array for later processing

	set submit $win.dummy_submit,[HMgetUniqueNum $state]
	if {[winfo exists $submit]} {
		destroy $submit
	}
	button $submit	-takefocus 0;# this never gets mapped!
	lappend form(submit_button) $submit
	set form(submit_$submit) [list $name $name.\$form(X).\$form(Y)]
	
	$item configure -takefocus 1
	bind $item <FocusIn> "catch \{$win see $item\}"
	bind $item <1> "$item configure -relief sunken"
	bind $item <Return> "
		set $var(form_id)(X) 0
		set $var(form_id)(Y) 0
		$submit invoke	
	"
	bind $item <ButtonRelease-1> "
		set $var(form_id)(X) %x
		set $var(form_id)(Y) %y
		$item configure -relief raised
		$submit invoke	
	"
	return $item
}

# Set up the reset button.  Wait for the /form to attach
# the -command option.  There could be more that 1 reset button
# params VALUE

proc HMinput_reset {state win param} {
	upvar #0 $state var
	upvar #0 $var(form_id) form

	set value reset
	HMextract_param $param value

	set item $win.input_reset,[HMgetUniqueNum $state]
	button $item -text [HMmap_esc $value]  -background $var(bgcolor) -foreground $var(fgcolor)
	HMwin_install $state $win $item
	lappend form(reset_button) $item
	return $item
}

# Set up the submit button.  Wait for the /form to attach
# the -command option.  There could be more that 1 submit button
# params: NAME, VALUE

proc HMinput_submit {state win param} {
	upvar #0 $state var
	upvar #0 $var(form_id) form

	HMextract_param $param name
	set value submit
	HMextract_param $param value
	set item $win.input_submit,[HMgetUniqueNum $state]
	button $item -text [HMmap_esc $value]  -background $var(bgcolor) -foreground $var(linkfg)
	HMwin_install $state $win $item
	lappend form(submit_button) $item
	# need to tie the "name=value" to this button
	# save the pair and do it when we finish the submit button
	catch {set form(submit_$item) [list $name $value]}
	return $item
}

#########################################################################
# selection items
# They all go into a list box.  We don't what to do with the listbox until
# we know how many items end up in it.  Gather up the data for the "options"
# and finish up in the /select tag
# params: NAME (reqd), MULTIPLE, SIZE 

proc HMtag_select {state win param text} {
	upvar #0 $state var
	upvar #0 $var(form_id) form

	HMextract_param $param name
	set size 5;  HMextract_param $param size
	set form(select_size) $size
	set form(select_name) $name
	set form(select_values) ""		;# list of values to submit
	if {[HMextract_param $param multiple]} {
		set mode multiple
	} else {
		set mode single
	}
	set item $win.select,[HMgetUniqueNum $state]
    frame $item -background $var(bgcolor)
    set form(select_frame) $item
	listbox $item.list -selectmode $mode -width 0 -exportselection 0  -background $var(bgcolor) -foreground $var(linkfg)
	HMwin_install $state $win $item
}

# select options
# The values returned in the query may be different from those
# displayed in the listbox, so we need to keep a separate list of
# query values.
#  form(select_default) - contains the default query value
#  form(select_frame) - name of the listbox's containing frame
#  form(select_values)  - list of query values
# params: VALUE, SELECTED

proc HMtag_option {state win param text} {
	upvar #0 $state var
	upvar #0 $var(form_id) form
	upvar $text data
	set frame $form(select_frame)

	# set default option (or options)
	if {[HMextract_param $param selected]} {
        lappend form(select_default) [$form(select_frame).list size]
    }
    set value [string trimright $data " \n"]
    $frame.list insert end $value
	HMextract_param $param value
	lappend form(select_values) $value
	set data ""
}
 
# do most of the work here!
# if SIZE>1, make the listbox.  Otherwise make a "drop-down"
# listbox with a label in it
# If the # of items > size, add a scroll bar
# This should probably be broken up into callbacks to make it
# easier to override the "look".

proc HMtag_/select {state win param text} {
	upvar #0 $state var
	upvar #0 $var(form_id) form
	set frame $form(select_frame)
	set size $form(select_size)
	set items [$frame.list size]

	# set the defaults and reset button
	append form(reset) ";$frame.list selection clear 0  $items"
	if {[info exists form(select_default)]} {
		foreach i $form(select_default) {
			$frame.list selection set $i
			append form(reset) ";$frame.list selection set $i"
		}
	} else {
		$frame.list selection set 0
		append form(reset) ";$frame.list selection set 0"
	}

	# set up the submit button. This is the general case.  For single
	# selections we could be smarter

	for {set i 0} {$i < $size} {incr i} {
		set value [format {[expr {[%s selection includes %s] ? {%s} : {}}]}  $frame.list $i [lindex $form(select_values) $i]]
		lappend form(submit) [list $form(select_name) $value]
	}
	
	# show the listbox - no scroll bar

	if {$size > 1 && $items <= $size} {
		$frame.list configure -height $items
		pack $frame.list

	# Listbox with scrollbar

	} elseif {$size > 1} {
		scrollbar $frame.scroll -command "$frame.list yview"   -orient v -takefocus 0
		$frame.list configure -height $size  -yscrollcommand "$frame.scroll set"
		pack $frame.list $frame.scroll -side right -fill y

	# This is a joke!

	} else {
		scrollbar $frame.scroll -command "$frame.list yview"   -orient h -takefocus 0
		$frame.list configure -height 1  -yscrollcommand "$frame.scroll set"
		pack $frame.list $frame.scroll -side top -fill x
	}

	# Inform the applet
	HMapplet_item $state $win $var(form_id) select $form(select_name) {} $frame

	# cleanup

	foreach i [array names form select_*] {
		unset form($i)
	}
}

# do a text area (multi-line text)
# params: COLS, NAME, ROWS (all reqd, but default rows and cols anyway)

proc HMtag_textarea {state win param text} {
	upvar #0 $state var
	upvar #0 $var(form_id) form
	upvar $text data

	set rows 5; HMextract_param $param rows
	set cols 30; HMextract_param $param cols
	HMextract_param $param name
	set item $win.textarea,[HMgetUniqueNum $state]
	frame $item -background $var(bgcolor)
	text $item.text -width $cols -height $rows -wrap none  -yscrollcommand "$item.scroll set" -padx 3 -pady 3  -background $var(bgcolor) -foreground $var(linkfg)
	catch {$item.text conf -kanjifont k14 -font a14; bind $item.text <Escape> [list kanjiInput end $item.text]}
	scrollbar $item.scroll -command "$item.text yview"  -orient v
	$item.text insert 1.0 $data
	HMwin_install $state $win $item
	pack $item.text $item.scroll -side right -fill y
	lappend form(submit) [list $name "\[$item.text get 0.0 end]"]
	append form(reset) ";$item.text delete 1.0 end;  $item.text insert 1.0 [list $data]"
	set data ""

	# Inform the applet
	HMapplet_item $state $win $var(form_id) textarea $name {} $item
}

# procedure to install windows into the text widget
# - state: name of hyperwindow state array
# - win:  name of the text widget
# - item: name of widget to install

proc HMwin_install {state win item} {
	upvar #0 $state var
	upvar #0 HN$win winvar
	$win window create $var(S_insert) -window $item -align bottom
	$win tag add indent$winvar(level) $item
	set focus [expr {[winfo class $item] != "Frame"}]
	$item configure -takefocus $focus
	bind $item <FocusIn> "$win see $item"
}

#####################################################################
# Assemble and submit the query
# each list element in "stuff" is a name/value pair
# - The names are the NAME parameters of the various fields
# - The values get run through "subst" to extract the values
# - We do the user callback with the list of name value pairs

proc HMsubmit_button {state win form_id param stuff} {
	upvar #0 $state var
	upvar #0 $form_id form
	set query ""
	foreach pair $stuff {
		set value [subst [lindex $pair 1]]
		if {$value != ""} {
			set item [lindex $pair 0]
			lappend query $item $value
		}
	}
	# this is the user callback.
	HMsubmit_form $state $win $form_id $param $query
}

# sample user callback for form submission
# should be replaced by the application
# Sample version generates a string suitable for http
# The application may wish to have the applet process
# the query first

proc HMsubmit_form {state win form_id param query} {
	set result ""
	set sep ""
	foreach i $query {
		append result  $sep [HMmap_reply $i]
		if {$sep != "="} {set sep =} {set sep &}
	}
	puts $result
}

# sample user callback for form reset
# should be replaced by the application
# The application should call the form's applet

proc HMreset_form {state win form_id} {
	puts "==> reset form $state $win $form_id"
}
# do x-www-urlencoded character mapping
# The spec says: "non-alphanumeric characters are replaced by '%HH'"
 
set HMalphanumeric	a-zA-Z0-9	;# definition of alphanumeric character class
for {set i 1} {$i <= 256} {incr i} {
    set c [format %c $i]
    if {![string match \[$HMalphanumeric\] $c]} {
        set HMform_map($c) %[format %.2x $i]
    }
}

# These are handled specially
array set HMform_map {
    " " +   \n %0d%0a
}

# 1 leave alphanumerics characters alone
# 2 Convert every other character to an array lookup
# 3 Escape constructs that are "special" to the tcl parser
# 4 "subst" the result, doing all the array substitutions
 
proc HMmap_reply {string} {
    global HMform_map HMalphanumeric
    regsub -all \[^$HMalphanumeric\] $string {$HMform_map(&)} string
    regsub -all \n $string {\\n} string
    regsub -all \t $string {\\t} string
    regsub -all {[][{})\\]\)} $string {\\&} string
    return [subst $string]
}

# convert a x-www-urlencoded string int a a list of name/value pairs

# 1  convert a=b&c=d... to {a} {b} {c} {d}...
# 2, convert + to  " "
# 3, convert %xx to char equiv

proc HMcgiDecode {data} {
	set data [split $data "&="]
	foreach i $data {
		lappend result [cgiMap $i]
	}
	return $result
}

proc HMcgiMap {data} {
	regsub -all {\+} $data " " data
	
	if {[regexp % $data]} {
		regsub -all {([][$\\])} $data {\\\1} data
		regsub -all {%([0-9a-fA-F][0-9a-fA-F])} $data  {[format %c 0x\1]} data
		return [subst $data]
	} else {
		return $data
	}
}



} \
safe-tk.tcl {# $Id: safe-tk.tcl,v 1.8 1996/08/06 11:54:33 steve Exp $
#
#	Copyright (c) 1996,1997 Australian National University.
#
#--------
#
#	PASTIME Project
#	Cooperative Research Centre for Advanced Computational Systems
#	COPYRIGHT NOTICE AND DISCLAIMER.
#
# Copyright (c) 1995,1996 ANU and CSIRO on behalf of the participants in 
# the  CRC for Advanced Computational Systems ('ACSys').
# 
# ACSys makes this software and all associated data and documentation
# ('Software') available free of charge for non-commercial purposes only. You
# may make copies of the Software but you must include all of this notice on
# any copy.
# 
# The Software was developed for research purposes and ACSys does not warrant
# that it is error free or fit for any purpose.  ACSys disclaims any
# liability for all claims, expenses, losses, damages and costs any user may
# incur as a result of using, copying or modifying the Software.
#
###
# Utility module for creating Safe interpreters with Tk
#
# Credits:
#	Tom Tromey, various fixes and enhancements
#	colin@field.medicine.adelaide.edu.au, security fixes

# All procedures and global variables use the prefix "interp"

package provide safe-tk [lindex {$Revision: 1.8 $} 1]

# This is a dummy procedure to allow the package to be autoloaded

proc interp_init {} {}

proc interp_create_safe_tk {{name {}}} {
    global interp_tk

    # Create a safe interpreter, w/o Tk
    set slave [eval interp create -safe $name]

    # Create a top level for this interpreter

    set top [toplevel .$slave -width 200 -height 200]
    wm title $top "Applet $name"
    bind interp$top <Destroy> "if {\"%W\" == \"$top\"} {interp_delete_slave $slave}"
    bindtags $top "interp$top [bindtags $top]"

    # Make sure any previous interp_tk array elements are removed
    foreach e [array get interp_tk $slave*] {catch {unset interp_tk($e)}}

    # Add in aliases for Tk commands

    # The initial window
    $slave alias . interp_tk_method $slave .

    interp_add_toplevel $slave $top

    # Widget creation commands
    foreach cmd {
	button 
	canvas checkbutton 
	entry 
	frame 
	label listbox 
	menubutton message 
	radiobutton 
	scale scrollbar 
	text } {
	$slave alias $cmd interp_tk_create $slave $cmd
    }
    # Exceptions: toplevels need some further restriction
    $slave alias toplevel interp_tk_create_toplevel $slave
    # menu items take a -command argument which must be trapped
    $slave alias menu interp_tk_create_menu $slave
    # these have different formal parameters
    $slave alias tk_dialog interp_tk_dialog $slave
    $slave alias tk_popup interp_tk_popup $slave
    $slave alias image interp_tk_image $slave

    # These commands will need restriction in some way
    foreach cmd {
	after 
	bell bind bindtags 
	destroy 
	focus tk_focusNext tk_focusPrev tk_focusFollowsMouse 
	grab grid 
	option 
	pack place 
	tkerror 
	update 
	winfo wm} {
	$slave alias $cmd interp_tk_$cmd $slave
    }
    # These two share the same implementation
    $slave alias lower interp_tk_order lower $slave
    $slave alias raise interp_tk_order raise $slave

    # Notable exceptions.  These commands would allow the system to be
    # compromised
    # clipboard fileevent palette selection send tk tkwait 
    # Note that it may be possible to implement the tkwait command
    # clipboard could be implemented in a restricted fashion

    # These commands are useless since they require access to variables
    # inside the slave interpreter
    # tk_optionMenu 

    return $slave
}

# Convenience routine for sending arguments for evaluation in the slave
# Just to get out of "quoting hell".  Courtesy Jacob Levy, Sun Labs.
proc interp_collect {args} {list $args}

# Add a toplevel widget to the list of exposed heirarchies

proc interp_add_toplevel {slave top} {
    global interp_tk
    lappend interp_tk($slave,toplevels) $top
}

proc interp_delete {slave} {
    destroy .$slave
}

proc interp_delete_slave {slave} {
    global interp_tk

    if {[array get interp_tk $slave,destroy] != {}} {
	return; # Already been here
    }

    foreach e [array get interp_tk $slave*] {catch {unset interp_tk($e)}}
    set interp_tk($slave,destroy) 1
    after idle interp_really_delete $slave
}

# Also remove images

proc interp_really_delete {slave} {
    foreach a [interp aliases $slave] {$slave alias $a {}}
    foreach i [image names] {
	if {[regexp "${slave}_" $i]} {image delete $i}
    }
    interp delete $slave
}

###
### Map slave widget paths to master widget paths, and vice versa
###

# By convention, slave pathnames starting with "master.<top>" are
# mapped to the pathname ".<top>".

proc interp_get_win {slave win} {
    global interp_tk

    if {[regexp {^master(\..+)} $win all win]} {
	# Find matching toplevel
	if {[regsub "([join [lrange $interp_tk($slave,toplevels) 1 end] |])" $win {} dummy]} {
	    return $win
	} else {error "bad window path name for \"$win\""}
    } else {
	if {[string index $win 0] != "."} {error "bad window path name \"$win\""}
	if {$win == "."} {
	    return [lindex $interp_tk($slave,toplevels) 0]
	} else {return [lindex $interp_tk($slave,toplevels) 0]$win}
    }
}

# Ditto for a line full of references

proc interp_fix_pathnames {top cmd} {
    # Take care of pathnames
    regsub -all [format {(^|[ 	[\{"]+)master%s(\.[^ 	]+)?} $top] $cmd "\\1$top\\2 " cmd
    return $cmd
}

proc interp_fix_all_pathnames {slave cmd} {
    global interp_tk

    # Do it for the applet's heirarchy
    set top [lindex $interp_tk($slave,toplevels) 0]
    regsub -all {(^|[ 	[\{"]+)(\.[^ 	]+)} $cmd "\\1$top\\2 " cmd
    # Take care of "."
    regsub -all {(^|[ 	[\{"]+)\.([ 	]+|$)} $cmd "\\1$top " cmd

    foreach top $interp_tk($slave,toplevels) {
	set cmd [interp_fix_pathnames $top $cmd]
    }
    return [string trim $cmd]
}

proc interp_unfix_all_pathnames {slave cmd} {
    global interp_tk

    # Do it for the applet's heirarchy
    regsub -all "(^|\[ 	\]+)[lindex $interp_tk($slave,toplevels) 0](\\.\[^ 	\]*)"  $cmd { \2 } cmd
    # Take care of "."
    regsub -all "(^|\[  \]+)[lindex $interp_tk($slave,toplevels) 0](\[ 	\]+|\$)"  $cmd { . } cmd

    foreach top [lrange $interp_tk($slave,toplevels) 1 end] {
	regsub -all [format {(^|[ 	]+)%s(\.[^ 	]*)?} $top]  $cmd " master$top\\2 " cmd
    }
    return [string trim $cmd]
}

###
### Creation, method and configuration alias targets
###

proc interp_tk_create {slave cmd win args} {
    # Map into master pathname & also check validity of operation
    set path [interp_get_win $slave $win]
    set extraOptions {}
    # Special case: checkbuttons and radiobuttons have globals
    # associated with them.  Unset these before proceeding.
    switch -glob $cmd {
	ch* {
	    set name [lindex [split $path .] end]
	    global $name
	    catch {unset $name}
	}
	rad* {
	    # Make the widget use a non-default variable
	    set extraOptions "-variable selectedButton$slave"
	}
    }
    # Create the widget
    eval $cmd $path $extraOptions
    if {[llength $args] > 0} {
	interp_tk_configure $slave $path $args
    }
    $slave alias $win interp_tk_method $slave $win
    return $win
}

# Same as above, only trap to menu-specific procedure

proc interp_tk_create_menu {slave win args} {
    # Map into master pathname & also check validity of operation
    set path [interp_get_win $slave $win]
    # Create the widget
    menu $path
    if {[llength $args] > 0} {
	interp_tk_configure $slave $path $args
    }
    $slave alias $win interp_tk_method_menu $slave $win
    return $win
}

# Image namespace mapped for slaves

global interp_tk_image_num 0
proc interp_tk_image {slave method args} {
    global interp_tk_image_num
    switch -glob $method {
	c* {
	    # create
	    set type [lindex $args 0]
	    if {[string index [lindex $args 1] 0] != "-"} {
		set name [lindex $args 1]
		set args [lrange $args 2 end]
	    } else {
		set name [incr interp_tk_image_num]
		set args [lrange $args 1 end]
	    }
	    eval image create $type ${slave}_$name $args
	    $slave alias ${slave}_$name interp_tk_method $slave ${slave}_$name
	    return $name
	}
	d* {
	    foreach i $args {
		image delete $slave_$i
	    }
	    return {}
	}
	n* {
	    regsub -all "${slave}_(.*)" [image names] {\\1} ret
	    return $ret
	}
	types {return [image $method]}
	type -
	h* -
	w* {return [image $method $slave_$args]}
	default {return [interp_evaluate_list image $args]}
    }
}

proc interp_tk_create_toplevel {slave win args} {
    # Map into master pathname & also check validity of operation
    set path [interp_get_win $slave $win]
    # Create the widget
    toplevel $path
    if {[llength $args] > 0} {
	interp_tk_configure $slave $path $args
    }
    $slave alias $win interp_tk_method $slave $win
    wm title $path "Applet: $slave; [wm title $path]"
    return $win
}

proc interp_tk_dialog {slave win title text bitmap default string args} {
    interp_evaluate_list tk_dialog  "[interp_get_win $slave $win] {$title} {$text} {$bitmap} {$default} {$string} $args"
}

proc interp_tk_popup {slave menu x y {entry {}}} {
    tk_popup [interp_get_win $slave $menu] $x $y $entry
}

proc interp_tk_method {slave win args} {
    # Map into master pathname & also check validity of operation
    set path [interp_get_win $slave $win]
    if {[string first conf [lindex $args 0]] == 0} {
	# Configure method
	return [interp_tk_configure $slave $path [lrange $args 1 end]]
    } else {
	interp_evaluate_list $path [interp_fix_all_pathnames $slave $args]
    }
}

# For menu items we need to trap the -command option
# and disallow the -variable option

proc interp_tk_method_menu {slave win args} {
    # Map into master pathname & also check validity of operation
    set path [interp_get_win $slave $win]
    set method [lindex $args 0]
    set command {}
    set entryidx {}
    switch -glob $method {
	ad* {
	    if {[lsearch $args "-var*"] != -1} {
		error "cannot access variables inside a slave interpreter"
	    } elseif {[set cidx [lsearch $args "-co*"]] != -1} {
		set command [lindex $args [expr $cidx + 1]]
		set args [lreplace $args $cidx [expr $cidx + 1]]
	    }
	}
	ins* {
	    if {[lsearch $args "-var*"] != -1} {
		error "cannot access variables inside a slave interpreter"
	    } elseif {[set cidx [lsearch $args "-co*"]] != -1} {
		set command [lindex $args [expr $cidx + 1]]
		set args [lreplace $args $cidx [expr $cidx + 1]]
		set entryidx [lindex $args 1]
	    }
	}
	entryco* {
	    if {[lsearch [lrange $args 2 end] "-var*"] != -1} {
		error "cannot access variables inside a slave interpreter"
	    } elseif {[set cidx [lsearch $args "-co*"]] != -1} {
		set command [lindex $args [expr $cidx + 1]]
		set args [lreplace $args $cidx [expr $cidx + 1]]
		set entryidx [lindex $args 1]
	    }
	}
    }

    set result {}
    # similar to interp_tk_method
    if {$args != ""} {
	if {[string first conf [lindex $args 0]] == 0} {
	    return [interp_tk_configure $slave $path [lrange $args 1 end]]
	} else {
	    set result [interp_evaluate_list $path [interp_fix_all_pathnames $slave $args]]
	}
	if {$entryidx == {}} {
	    set entryidx [$path index end]	;# New menu items are created on the end
	}
    }

    # correctly configure -command option
    if {$command != {}} {
	$path entryconfigure $entryidx -command "$slave eval \[interp_unfix_all_pathnames $slave \{$command\}\]"
    }
    return $result
}

# interp_tk_mincom - unfortunately such things as -xscrollcommand
# can be abbreviated (and to different extents, depending on the widget)
array set interp_tk_comconf {
    Button	{-co}
    Canvas	{-xscrollc -yscrollc}
    Checkbutton	{-co}
    Entry	{-x -y}
    Frame	{}
    Label	{}
    Listbox	{-x -y}
    Menu	{-tearoffc -p}
    Menubutton	{}
    Message	{}
    Radiobutton	{-co}
    Scale	{-co}
    Scrollbar	{-co}
    Text	{-x -y}
}

array set interp_tk_varconf {
    Button	{-textv}
    Canvas	{}
    Checkbutton	{-v -textv}
    Entry	{-text}
    Frame	{}
    Label	{-textv}
    Listbox	{}
    Menu	{}
    Menubutton	{-textv}
    Message	{}
    Radiobutton	{-textv -var}
    Scale	{-v}
    Scrollbar	{}
    Text	{}
}

# We need to be careful here and intercept any options that specify
# scripts.  These scripts should be evaluated in the slave interpreter.

# References to images must be mapped to the slave image namespace

# Also, -variable and -textvariable variables should be made to refer to 
# a variable in the slave interpreter.  This is not currently possible.

# We need to prevent the Applet from covering the entire screen.  This is
# difficult to do in general by examining the -width & -height options
# given to each widget.  Rather, we will allow widgets to size themelves
# but then check the final width & height of the toplevel once we are done.

proc interp_tk_configure {slave win cfg} {
    global interp_tk_comconf interp_tk_varconf

    if {[llength $cfg] == 0} {
	return [$win configure]
    } elseif {[llength $cfg] == 1} {
	return [$win configure $cfg]
    }

    if {[expr [llength $cfg] % 2] == 1} {
	error "value for \"[lindex $cfg end]\" missing"
    }

    # iterate and sanitise each {option,value} pair in the configuration
    foreach {opt value} $cfg {
	set value [interp_fix_all_pathnames $slave [string trim $value]]

	# check for configurations which contain commands
	if {![catch {set match $interp_tk_comconf([winfo class $win])}]} {
	    foreach m $match {
		if {[string first $m $opt] == 0} {
		    set value "$slave eval \[interp_unfix_all_pathnames $slave \{$value\}\]"
		    break
		}
	    }
	}

	# check for configurations which contain variable references
	if {![catch {set match $interp_tk_varconf([winfo class $win])}]} {
	    foreach m $match {
		if {[string first $m $opt] == 0} {
		    set value $slave.$value
		    break
		}
	    }
	}

	if {[string first $opt "-image"] == 0} {
	    set value ${slave}_$value
	} 
    
	# Looks harmless enough (now :)
	$win configure $opt $value
    }

    # Check that the screen hasn't been covered (include a safety margin)
    set top [winfo toplevel $win]
    if {[winfo reqheight $top] > [expr [winfo screenheight .] - 10]} {
	$top configure -height [expr [winfo screenheight .] - 10]
    }
    if {[winfo reqwidth $top] > [expr [winfo screenwidth .] - 10]} {
	$top configure -width [expr [winfo screenwidth .] - 10]
    }

    return {}
}

###
### Tk commands
###

# Make sure that any scripts are evaluated in the slave interpreter

proc interp_tk_after {slave args} {
    if {$args == {}} {after} ;# Cause the error
    switch -glob [lindex $args 0] {
	c* {
	    # cancel
	    if {[llength $args] == 2} {
		# Is arg #3 an id or a script?  Do gross hack...
		if {[regexp {^after#} [lindex $args 1]]} {
		    after cancel [lindex $args 1]
		    break
		}
	    }
	    set args "cancel $slave eval [lrange $args 1 end]"
	    interp_evaluate_list after $args
	}
	i* {
	    # idle
	    return [after idle catch \{$slave eval [lrange $args 1 end]\}]
	}
	default {
	    if {[llength $args] == 1} {error "safe interpreters may not sleep"}
	    return [after [lindex $args 0] catch \{$slave eval [lrange $args 1 end]\}]
	}
    }
    return {}
}

# Restrict to ringing the bell of the slave's toplevel window
# interp_bell_quota specifies how many times the bell may be rung at a time
# interp_bell_freq specifies how often, in milliseconds, the bell may be rung

# By default the bell may only be rung three times each second.
array set interp_tk {
	interp_bell_quota	3
	interp_bell_freq	1000
}
proc interp_tk_bell {slave args} {
    global interp_tk

    # If the handler routine hasn't been scheduled, 
    # then ring the bell immediately and schedule handler
    if {[array get interp_tk $slave,bell_handler] == {}} {
	# should check -displayof option...
	bell -displayof [lindex $interp_tk($slave,toplevels) 0]
	set interp_tk($slave,bell_handler) [after $interp_tk(interp_bell_freq) interp_bell_handler $slave]
    } else {
	# Add a bell request to the request queue
	if {[array get interp_tk $slave,bell_cnt] == {}} {
	    set interp_tk($slave,bell_cnt) 1
	} else {incr interp_tk($slave,bell_cnt) 1}
    }
}

proc interp_bell_handler {slave} {
    global interp_tk

    if {[array get interp_tk $slave,bell_cnt] != {}} {
	for {set i 0} {$i < $interp_tk($slave,bell_cnt) && $i < $interp_tk(interp_bell_quota)} {incr i} {
	    bell -displayof [lindex $interp_tk($slave,toplevels) 0]
	}
    }
    catch {unset interp_tk($slave,bell_handler)}
    catch {unset interp_tk($slave,bell_cnt)}
}

# The thing to watch here, of course, is to make sure that scripts get 
# evaluated in the slave interpreter.
# Also, bindings added to windows not created by the applet must be
# removed when the slave interpreter is deleted (this doesn't currently happen)

proc interp_tk_bind {slave tag args} {
    if {[regexp {^master\.} $tag] || [string index $tag 0] == "."} {
	set slavetag [interp_get_win $slave $tag]
    } else {set slavetag $tag}

    if {[llength $args] == 0} {
	return [bind $slavetag]
    } elseif {[llength $args] == 1} {
	return [bind $slavetag $args]
    } elseif {[llength $args] > 2} {
	# Cause an error message to be sent back
	return {[bind $slavetag <1> too many arguments]}
    } elseif {[string index [lindex $args 1] 0] == "+"} {
	bind $slavetag [lindex $args 0]  "+$slave eval \[interp_unfix_all_pathnames $slave \{[string range [lindex $args 1] 1 end]\}\]"
    } else {
	bind $slavetag [lindex $args 0]  "$slave eval \[interp_unfix_all_pathnames $slave \{[lindex $args 1]\}\]"
    }
    return {}
}

# We need to store the bindtags added to windows not created by the
# applet and remove these when the applet is deleted

proc interp_tk_bindtags {slave args} {
    if {[llength $args] == 1} {
	return [interp_unfix_all_pathnames $slave  [bindtags [interp_get_win $slave [lindex $args 0]]]]
    } else {
	return [interp_evaluate_list bindtags [interp_fix_all_pathnames $slave $args]]
    }
}

proc interp_tk_destroy {slave args} {
    foreach win $args {
	set win [interp_get_win $slave $win]
	destroy $win
    }
}

# Disallow the "force" method - too dangerous

proc interp_tk_focus {slave args} {
    if {$args == {}} {
	return [focus displayof $top]
    } elseif {[llength $args] == 1} {
	return [focus [interp_get_win $slave $args]]
    } elseif {[llength $args] > 2} {
	# Cause the error message to be returned
	return [focus [lindex $args 0] [lindex $args 1] [lindex $args 2]]
    } elseif {[lindex $args 0] != "force"} {
	return [focus [lindex $args 0] [interp_get_win $slave [lindex $args 1]]]
    } else {
	error "method \"force\" is not permitted for safe interpreters"
    }
}

proc interp_tk_tk_focusNext {slave win} {
    return tk_focusNext [interp_get_win $slave $win]
}

proc interp_tk_tk_focusPrev {slave win} {
    return tk_focusPrev [interp_get_win $slave $win]
}

# Don't let the slave interpreter affect other windows

proc interp_tk_tk_focusFollowsMouse {slave} {
    error "safe interpreters may not change the focus model"
}

# Only allow local grabs
# Window pathnames returned should be normalized for the slave interpreter's
# context

proc interp_tk_grab {slave args} {
    if {$args == {}} {grab}; # Cause the error
    switch -glob [lindex $args 0] {
	-g* {error "global grabs not permitted for safe interpreters"}
	c* {
	    # current
	    if {[llength $args] == 1} {
		return [grab current]
	    } elseif {[lindex $args] > 2} {
		grab current [lindex $args 1] [lindex $args 2]; # Cause the error
	    } else {
		return [grab current [interp_get_win $slave [lindex $args 1]]]
	    }
	}
	r* {
	    # release
	    if {[llength $args] > 2} {
		grab release [lindex $args 1] [lindex $args 2]; # Cause the error
	    }
	    grab release [interp_get_win $slave [lindex $args 1]]
	}
	se* {
	    # set
	    if {[llength $args] == 1} {grab set}; # Cause the error
	    if {[lindex $args 1] == "-global"} {error "global grabs not permitted for safe interpreters"}
	    if {[[length $args] > 2} {
		grab set [lindex $args 1] [lindex $args 2]; # Cause the error
	    }
	    grab set [interp_get_win $slave [lindex $args 1]]
	}
	st* {
	    # status
	    if {[llength $args] == 1} {grab status}; # Cause the error
	    if {[[length $args] > 2} {
		grab status [lindex $args 1] [lindex $args 2]; # Cause the error
	    }
	    return [grab status [interp_get_win $slave [lindex $args 1]]]
	}
	default {grab [lindex $args 0] [lrange $args 1 end]}; # Cause the error
    }
    return {}
}

proc interp_tk_grid {slave args} {
    return [interp_evaluate_list grid [interp_fix_all_pathnames $slave $args]]
}

proc interp_tk_order {dir slave top win args} {
    set win [interp_get_win $slave $win]
    if {$args == {}} {
	$dir $win
    } elseif {[llength $args > 1]} {
	$dir $win [lindex $args 1] [lindex $args 2]; # Cause the error
    } else {
	$dir $win [interp_get_win $slave $args]
    }
}

# Only the get method is permitted

proc interp_tk_option {slave method args} {
    if {$method != "get"} {error "safe interpreters may not change options"}
    if {[llength $args] != 3} {error "wrong # arguments"}
    return [option get [interp_get_win $slave [lindex $args 0]] [lindex $args 1] [lindex $args 2]]
}

proc interp_tk_pack {slave args} {
    return [interp_evaluate_list pack [interp_fix_all_pathnames $slave $args]]
}

proc interp_tk_place {slave args} {
    return [interp_evaluate_list place [interp_fix_all_pathnames $slave $args]]
}

proc interp_tk_update {slave args} {
    if {$args == {}} {update} else {update $args}
}

# Disallow interps method.  Munge parent, pathname, toplevel return value

proc interp_tk_winfo {slave args} {
    if {$args == {}} {winfo}; # Cause the error
    set args [interp_fix_all_pathnames $slave $args]

    switch -glob [lindex $args 0] {
	in* {
	    # interps
	    error "\"interps\" method disallowed for safe interpreters"
	}
	par* -
	pat* -
	t* {
	    # parent, pathname, toplevel
	    return [interp_unfix_all_pathnames $slave  [interp_evaluate_list winfo $args]]
	}
	default {
	    return [interp_evaluate_list winfo $args]
	}
    }
}

# Play it safe with the potentially dangerous methods for the moment
# Simply retrieving values is probably safe though.

proc interp_tk_wm {slave args} {
    if {$args == {}} {wm}; # Cause the error
    set args [interp_fix_all_pathnames $slave $args]

    switch [lindex $args 0] {
	cl* -
	co* -
	fo* -
	pr* -
	tr* {
	    # client, colormapwindows, command, focusmodel, protocol, transient
	    error "method \"[lindex $args 0]\" disallowed for safe interpreters"}
	ti* {
	    # title
	    if {[llength $args] == 3} {
		wm title [lindex $args 1] "Applet: $slave; [lindex $args 2]"
	    } else {
		return [wm title [lindex $args 1]]
	    }
	}
	default {
	    return [interp_evaluate_list wm $args]
	}
    }
}

# Experimental version.  Courtesy Tom Tromey (tromey@cygnus.com)

proc interp_evaluate_list {cmd args} {
  set args [lindex $args 0]
  return [eval [list $cmd] $args]
}
} \
applets.tcl {# $Id: applets.tcl,v 1.9 1996/08/06 11:54:21 steve Exp $
#
#	Copyright (c) 1996,1997 Australian National University
#
#--------
#
#	PASTIME Project
#	Cooperative Research Centre for Advanced Computational Systems
#	COPYRIGHT NOTICE AND DISCLAIMER.
#
# Copyright (c) 1995,1996 ANU and CSIRO on behalf of the participants in 
# the  CRC for Advanced Computational Systems ('ACSys').
# 
# ACSys makes this software and all associated data and documentation
# ('Software') available free of charge for non-commercial purposes only. You
# may make copies of the Software but you must include all of this notice on
# any copy.
# 
# The Software was developed for research purposes and ACSys does not warrant
# that it is error free or fit for any purpose.  ACSys disclaims any
# liability for all claims, expenses, losses, damages and costs any user may
# incur as a result of using, copying or modifying the Software.
#
###
# applets.tcl creates a content-type handler for application/x-tcl documents
#
# We must be extremely careful not to perform substitutions or evaluate the 
# Applet code in the main (trusted) interpreter or otherwise the security of 
# the browser (and the user's entire environment) may be compromised.
#
# Facilities are also provided for Applets to interact with the main interpreter
# in a safe (restricted) manner.
#
# By default, applets are local to a hyperpage (unless they are the target
# of a <FORM SCRIPT=applet> element, in which case they are local to that form).
# All applets local to a hyperpage are destroyed when that page is reset.
# Applets may declare themselves to be local to a window, or global to the browser.
#
# All procedures and variables in this module are prefixed by "App"
# Application call-ins:
#	Applet_created <dataState> <renderState> <slave>
# 	Applet_levelChange <slave> <old level> <new level>
#	Applet_new_page <base> <win>
#	Applet_browserversion
#	Applet_loadURI <URI> <win> <mime-type>

package provide applets [lindex {$Revision: 1.9 $} 1]

proc Applet_init {} {

    ### Augment HTML renderer window installation procedure so that
    ### all applets attached to this hyperpage can manipulate the window
    ### and its children

    if {[info commands Applet_win_install_original] == {}} {
	rename HMwin_install Applet_win_install_original
    }

    proc HMwin_install {state win item} {
	Applet_win_install_original $state $win $item

	# If an applet is attached to the hyperpage define a target alias 
	# in the slave for this window
	upvar #0 $state var
	upvar #0 AppPage$var(basewin) app

	foreach a [array names app] {
	    if {[interp exists $a]} {
		Applet_install_win_interp $a $item
	    }
	}
    }

    # Form management routines.
    # A race condition exists wrt to managing forms.
    # The applet may not have been created when these 
    # procedures are called.  To overcome this, the
    # parameters are stored and the procedures re-invoked
    # when the applet is created.

    # This is suboptimal - the user can interact with the form
    # (eg. enter data, submit it) before the applet has a chance to
    # manage it.  Form display should probably be suspended
    # until the applet is loaded.  However, we currently don't
    # have a mechanism for suspending HTML display.

    # When the form is finished, let the applet know the index
    # of the end of the form.

    proc HMapplet_/form {state win form_id} {
	upvar #0 $state var
	upvar #0 AppFormsMap$var(basewin) formmap

	if {[catch {set formmap($form_id)} slave]} {
	    upvar #0 $form_id form
	    # HMend_form will be invoked later
	    set form(Applet_formready) 1
	    return
	}

	upvar #0 AppState$slave appstate
	set appstate(formendindex) [$win index {end - 1 char}]

	# Call into slave so that it knows that the form is ready
	catch {$slave eval HMend_form}
    }

    # HMapplet_item is called for each input element in the form.

    proc HMapplet_item {state win form_id type name value item} {
	debug_puts [list HMapplet_item $state $win $form_id $type $name $value $item]
	upvar #0 $state var
	upvar #0 AppFormsMap$var(basewin) formmap
	upvar #0 $form_id form

	if {[catch {set formmap($form_id)} slave]} {
	    # HMapplet_item will be invoked later
	    debug_puts [list storing form item for later notification]
	    lappend form(Applet_formitems) [list $type $name $value $item]
	    return
	}

	# Define a target alias in the slave for this window and all children
	if {$item != {}} {Applet_install_win_interp $slave $item}

	upvar #0 AppState$slave appstate
	lappend appstate(formitems) [list $type $name $value $item]
	catch {$slave eval HMapplet_item \{$type\} \{$name\} \{$value\} \{[interp_unfix_all_pathnames $slave $item]\}} err
	debug_puts [list slave notification returned $err]
    }
}

# Problem: disambiguating applet inclusions: eg. 
# <html><body><a rel=embed href="applet.tcl"><a rel=embed href="applet.tcl"></body</html>
# The same applet source is used in two separate applet instances.
# We solve this by assigning a unique identifier to every document downloaded,
# and using that in creating applets.

# Saving the script text shouldn't be necessary -
# should just get it out of the cache

proc Applet_handler {dataState renderState data} {
    upvar #0 $dataState var

    # Store the Applet code in a variable for later evaluation
    # in a slave interpreter

    append var(AppScript) $data
    return {}
}

proc Applet_evaluate {dataState renderState} {
    upvar #0 $dataState var
    upvar #0 $renderState ren

    # Check that the applet script has been received
    if {![info exists var(AppScript)] || $var(AppScript) == {}} return;

    # Check that win really refers to a window
    if {[info exists ren(AppWin)]} {
	set win $ren(AppWin)
    } else {set win $var(cookie)}

    upvar #0 AppPage$win page
    upvar #0 AppForms$win forms
    upvar #0 AppFormsMap$win formmap

    # Create a slave interpreter for this applet
    # Safe-Tcl doesn't support Tk within safe interpreters (yet)
    # so we need to add in this functionality via aliases
    set slave [interp_create_safe_tk $var(id)]

    # Configure the slave interpreter

    $slave alias applet Applet_command $slave $win
    $slave alias exit Applet_destroy $slave

    $slave alias HMmap_reply HMmap_reply

    # Just for debugging
    $slave alias puts Applet_puts $slave

    # Expose other commands to the slave
    $slave alias random random

    # Allow the applet access to the hyperpage from which it was loaded
    interp_add_toplevel $slave [SurfIt_hyperpage $win]
    $slave alias master[SurfIt_hyperpage $win] interp_tk_method $slave master[SurfIt_hyperpage $win]

    # Attach the applet
    if {[info exists ren(form)]} {
	# Attach at the form level, and record which form
	set forms($slave) $ren(form)
	set formmap($ren(form)) $slave	;# Map from form ID to slave
    } else {
	# Attach at the hyperpage level
	set page($slave) {}
    }

    # Initialise the global array for this Applet's state
    upvar #0 AppState$slave appstate
    set appstate(loadID) 0
    if {[info exists ren(index)]} {
	set appstate(embedindex) $ren(index)
    } else {
	set appstate(embedindex) [$win index {end - 1 char}]
    }
    set appstate(url) $var(url)

    # Callback to the application so that it can set up
    # controls for this applet
    catch {Applet_created $dataState $renderState $slave}

    # Send the applet code to the interpreter for evaluation
    if {[catch {$slave eval $var(AppScript)} apperr]} {
	puts stderr "Applet execution resulted in error:\n$apperr"
    }
    unset var(AppScript)	;# no longer needed

    # The applet loader (eg. HMload_applet) must supply the link between
    # 
    if {[info exists ren(form)]} {
	debug_puts [list applet is attached to form $ren(form)]
	upvar #0 $ren(form) formstate
	# If this applet is attached to a form and form items have already
	# been processed then invoke HMapplet_item for each item
	if {[info exists formstate(Applet_formitems)]} {
	    debug_puts [list notifying applet of items $formstate(Applet_formitems)]
	    catch {foreach item $formstate(Applet_formitems) {
		debug_puts "eval HMapplet_item $ren(renderState) $win $forms($slave) $item"
		eval HMapplet_item $ren(renderState) $win $ren(form) $item
	    }} err
	    debug_puts [list notifying returned $err]
	    unset formstate(Applet_formitems)
	} else {
	    debug_puts [list no pending form items]
	}
	# If this applet is attached to a form and the </FORM> has already
	# been seen then invoke HMend_form
	if {[info exists formstate(Applet_formready)]} {
	    HMapplet_/form $ren(renderState) $win $forms($slave)
	    debug_puts [list notifying applet of </FORM>]
	}
    }
}

# Register the handler for application/x-tcl content-type

catch {
    PRregister_type application/x-tcl .tcl PRpassNpost Applet_handler Applet_evaluate 99
}

# Destroy all applets associated with a window

proc Applet_destroy_window {win} {
    upvar #0 AppWin$win app

    Applet_destroy_page $win	;# Also destroy hyperpage & form applets

    foreach a [array names app] {
	if {[interp exists $a]} {
	    Applet_destroy $a
	}
    }
    catch {unset app}
}

# Destroy all applets associated with a hyperpage

proc Applet_destroy_page {win} {
    upvar #0 AppPage$win app
    upvar #0 AppForms$win forms
    upvar #0 AppFormsMap$win formmap

    foreach a [array names app] {
	if {[interp exists $a]} {
	    Applet_destroy $a
	}
    }
    catch {unset app}

    foreach a [array names forms] {
	if {[interp exists $a]} {
	    Applet_destroy $a
	}
    }
    catch {unset forms}
    catch {unset formmap}
}

# Destroy an applet attached to a form

proc Applet_destroy_form {win form_id} {
    upvar #0 AppForms$win forms
    upvar #0 AppFormsMap$win formmap

    if {[info exists formmap($form_id)]} {
	Applet_destroy $formmap($form_id)
	unset form($formmap($form_id))
	unset formmap($form_id)
    }
}

### Application API

# Notify all applets on a hyperpage that an anchor has been activated.
# Should applets attached to forms be notified? Yes, but probably only
# of anchors activated within their form (not implemented)

proc Applet_anchor_activation {win url} {
    upvar #0 AppPage$win page
    upvar #0 AppWin$win app

    foreach a [array names app] {
	if {[interp exists $a]} {
	    catch {$a eval anchor_activation $url}
	}
    }
    foreach a [array names page] {
	if {[interp exists $a]} {
	    catch {$a eval anchor_activation $url}
	}
    }
}

# Notify all applets when a new page has been loaded
# NB. Only hyperwindow level applets need be informed, since
# lower level applets attached to the previous page have 
# been terminated

proc Applet_pageloaded {win url} {
    upvar #0 AppWin$win app

    foreach a [array names app] {
	if {[interp exists $a]} {
	    catch {$a eval pageloaded $url}
	}
    }
}

### API for form management

proc Applet_install_win_interp {slave win} {
    set winlist $win
    while {$winlist != {}} {
	# Remove the head of the list, and then expand the list with the
	# children of the (just removed) head
	set path [lindex $winlist 0]
	set winlist [lrange $winlist 1 end]
	append winlist [winfo children $path]

	set path [interp_unfix_all_pathnames $slave $path]
	$slave alias $path interp_tk_method $slave $path
    }
}

###
### Slave alias target procedures
###

#
# The "applet" command allows the applet to interact 
# with the master interpreter
#

proc Applet_command {slave win method args} {
    upvar #0 AppState$slave appstate

    set hyperpage [SurfIt_hyperpage $win]
    switch $method {
	browserversion {
	    if {$args != {}} {error "too many arguments"}

	    if {[catch {Applet_browserversion} version]} {
		return {}
	    } else {return $version}
	}
	embedindex {
	    if {$args != {}} {error "too many arguments"}

	    # The index at which the applet was embedded
	    return $appstate(embedindex)
	}
	embedwindow {
	    if {$args != {}} {error "too many arguments"}

	    # The Text widget in which the applet is embedded
	    return master$win
	}
	flush {
	    if {[llength $args] != 1} {error "wrong # arguments"}

	    Cache_expire_document $args
	    return {}
	}
	formendindex {
	    if {$args != {}} {error "too many arguments"}
	    upvar #0 AppForms$win forms
	    if {![info exists forms($slave)]} {error "slave is not attached to a form"}
	    if {[info exists appstate(formendindex)]} {
		return $appstate(formendindex)
	    } else {
		# Hasn't been set yet
		return {}
	    }
	}
	formitems {
	    if {$args != {}} {error "too many arguments"}
	    upvar #0 AppForms$win forms
	    if {![info exists forms($slave)]} {error "slave is not attached to a form"}
	    if {[info exists appstate(formitems)]} {
		# Map master widget pathnames to slave pathnames
		return [interp_unfix_all_pathnames $slave $appstate(formitems)]
	    } else {
		# Hasn't been set yet
		return {}
	    }
	}
	level {
	    if {[llength $args] > 1} {error "too many arguments"}

	    # Report or change at which level the applet is attached

	    upvar #0 AppForms$win forms
	    upvar #0 AppPage$hyperpage page
	    upvar #0 AppWin$hyperpage appwin
	    global AppBrowser

	    # Find where the applet is currently attached
	    if {[info exists forms($slave)]} {
		set former form
	    } elseif {[info exists page($slave)]} {
		set former hyperpage
	    } elseif {[info exists appwin($slave)]} {
		set former hyperwindow
	    } elseif {[info exists AppBrowser($slave)]} {
		set former browser
	    } else {error "*** internal error: cannot find slave interpreter"}

	    if {$args == {}} {return $former}

	    ### Change applet level

	    if {$former == "form"} {error "cannot reattach from form level"}
	    switch $args {
		form {error "cannot demote applet"}
		hyperpage {error "cannot demote applet"}
		hyperwindow {if {$former == "browser"} {error "cannot demote applet"}}
		browser {}
		default {error "unknown level \"$args\""}
	    }
	    if {$former == $args} return	;# No change

	    # Remove applet from previous level
	    if {$former == "hyperpage"} {unset page($slave)} else {unset appwin($slave)}

	    # Reattach at the new level
	    if {$args == "hyperwindow"} {set appwin($slave) {}} else {set AppBrowser($slave) {}}

	    # Callback to application to signal that things have changed
	    catch {Applet_levelChange $slave $former $args}

	    return {}
	}
	loaddata {
	    set method get; set put_data {}
	    lassign $args url data callback method put_data
	    Applet_loadData $win $slave $url $data $callback $method $put_data
	}
	loaduri {
	    # Args should be uri [type]
	    if {[llength $args] == 1} {
		set type {}
	    } elseif {[llength $args] != 2} {
		error "wrong # arguments"
	    } else {set type [lindex $args 1]}

	    Applet_loadURI [lindex $args 0] $win $type

	    return {}
	}
	newpage {
	    if {$args != {}} {error "too many arguments"}

	    # Callback to application to clear the hyperpage
	    catch {Applet_new_page $appstate(url) $win}

	    return {}
	}
	parsehtml {
	    catch {Applet_parse_html AppletParse$slave [lindex $args 0] "HMrender $win"}

	    return {}
	}
	default {error "unknown method \"$method\""}
    }
}

#
# Self destruct
#

proc Applet_destroy {slave args} {
    catch {$slave eval terminate}

    # This is wrong, since the Applet is still on the 
    # execution stack.
    #interp delete $slave

    # Instead, arrange for the slave interpeter to be removed later
    after idle interp_delete $slave
}

# Interim data load handler: We want to allow nonblocking I/O, but
# we'll batch up the document until it has been downloaded and then
# deliver it to the applet.  These procedures currently only support text data.
# 'callback' is the procedure to invoke when the data has been downloaded.
# 'data' is the (global) variable which is to receive the data.

proc Applet_loadData {win slave url data callback {method get} {put_data {}}} {
    global surfit_feedback
    # Allocate global variable for this download
    upvar #0 AppData$slave appstate

    # Provide visual cues for the user
    set surfit_feedback([SurfIt_hyperpage $win]) "Applet $slave is downloading \"$url\""
    SurfIt_start_download $win
    update

    array set appstate [list win $win slave $slave url $url closed 0 var $data callback $callback]
    PRloadDocument $url $slave AppData$slave Applet_dataStart surfit/appdata $method $put_data
}

catch {
    PRregister_type surfit/appdata {} Applet_dataProcessor Applet_dataRender Applet_dataClose
}

proc Applet_dataStart {dataState renderState} {
    global surfit_pageStatus surfit_feedback
    upvar #0 $dataState var
    upvar #0 $renderState ren

    set hyperpage [SurfIt_hyperpage $ren(win)]
    incr surfit_pageStatus($hyperpage,numdocs)
    catch {incr surfit_pageStatus($hyperpage,total) $var(HDRcontent-length)}
    set surfit_feedback($hyperpage) "Applet $ren(slave) is downloading \"$ren(url)\""
    update idletasks
    SurfIt_setFeedback $hyperpage
}

# Map images to the slave image namespace
# This is done by copying the image

proc Applet_dataProcessor {dataState dataType data} {
    global surfit_feedback
    upvar #0 $dataState var
    upvar #0 $var(renderState) ren

    if {[string match image/* $var(HDRcontent-type)]} {
	set img [Image_process $dataState $dataType $data]
	if {$img != {}} {
	    set slaveimg [image create [image type $img] $ren(slave)_$img]
	    $slaveimg configure -width [image width $img] -height [image height $img]
	    switch [image type $img] {
		photo {$slaveimg copy $img}
		bitmap {$slaveimg configure -data [$img cget -data]}
	    }
	}
	return $img
    } else {
	return $data
    }
}

proc Applet_dataRender {dataState renderState data} {
    global surfit_feedback
    upvar #0 $dataState var
    upvar #0 $renderState ren

    set surfit_feedback([SurfIt_hyperpage $ren(win)]) "Applet $ren(slave) is downloading \"$ren(url)\""
    update idletasks

    if {[PRtextData $var(HDRcontent-type)]} {
	if {$var(dataType) != "PRpost"} {
	    $ren(slave) eval append $ren(var) [list $data]
	} else {
	    # send the raw data
	    switch [set type [Cache_get_raw_data $ren(url) data]] {
		PRdata {$ren(slave) eval set $ren(var) [list $data]}
		default {error "cannot deliver data of type $type to applet}
	    }
	}
	$ren(slave) eval $ren(callback) [list [list content-type $var(HDRcontent-type)]] [list $data]
    } elseif {[string match image/* $var(HDRcontent-type)]} {
	$ren(slave) eval set $ren(var) $data
	$ren(slave) eval $ren(callback) [list [list content-type $var(HDRcontent-type)]] $data
    } ;# else {applet can't access other forms of binary data}
}

proc Applet_dataClose {dataState renderState} {
    global surfit_feedback
    upvar #0 $dataState var
    upvar #0 $renderState ren

    if {!$var(eof) || $ren(closed)} return;

    if {[string match image/* $var(HDRcontent-type)]} {
	Image_close $dataState $renderState
    }

    $ren(slave) eval $ren(callback) eof

    set surfit_feedback([SurfIt_hyperpage $ren(win)]) {}
    set ren(closed) 1
}

proc Applet_blt_table {slave args} {
    interp_evaluate_list blt_table [interp_fix_all_pathnames $slave $args]
}

proc Applet_puts {slave args} {
    if {[llength $args] > 1} {
	set filed [lindex $args 0]
	set text [lindex $args 1]
	if {$filed == "stdout" || $filed == "stderr"} {
	    puts $filed "Applet $slave: $text"
	} else {error "applet not permitted to use file descriptor \"$filed\""}
    } else {puts "Applet $slave: [lindex $args 0]"}
}
} \
handlers.tcl {# $Id: handlers.tcl,v 1.11 1996/08/06 11:54:26 steve Exp $
#
#	Copyright (c) 1996 Australian National University
#
#--------
#
#	PASTIME Project
#	Cooperative Research Centre for Advanced Computational Systems
#	COPYRIGHT NOTICE AND DISCLAIMER.
#
# Copyright (c) 1995,1996 ANU and CSIRO on behalf of the participants in 
# the  CRC for Advanced Computational Systems ('ACSys').
# 
# ACSys makes this software and all associated data and documentation
# ('Software') available free of charge for non-commercial purposes only. You
# may make copies of the Software but you must include all of this notice on
# any copy.
# 
# The Software was developed for research purposes and ACSys does not warrant
# that it is error free or fit for any purpose.  ACSys disclaims any
# liability for all claims, expenses, losses, damages and costs any user may
# incur as a result of using, copying or modifying the Software.
#
###
# Content-type handlers for sundry types, plus renderers for Text widget.

package provide handlers [lindex {$Revision: 1.11 $} 1]

proc handlers_init {} {}

# The Text handling procedures are all prepended by "Text"
# The image handling procedures are all prepended by "Image"
# The Helper application procedures are all prepended by "Helper"

proc Text_render {dataState renderState data} {
    upvar #0 $dataState var
    $var(cookie) tag configure Tnormal -font [HMx_font courier 14 medium r]
    catch {$var(cookie) tag configure Tnormal -kanjifont k14}
    $var(cookie) insert end $data Tnormal
}

# Register this handler for text/plain, text/text and text/*

foreach t {text/text text/plain text/*} {
    catch {
	PRregister_type $t .txt PRpassNpost Text_render {} 99 text.document
    }
}

#
# Image type handlers - image/x-xbitmap, image/gif, image/x-portable-pixmap
# (and later, image/jpeg)
# These can't actually do anything until the data has been read into the
# cache file.  At that time, load the image data into the image type using
# the -file switch.
#

proc Image_process {dataState dataType data} {
    upvar #0 $dataState var

    if {![info exists var(image)]} {
	set var(image) [Image_create $dataState]
	Cache_add_post $dataState $var(image) $var(HDRcontent-type)
    }
    switch $dataType {
	PRfile {
	    catch {$var(image) configure -file [eval file join $var(file)]}
	    # Override normal protocol handling as required
	    if {![string compare file $var(scheme)]} {
		set var(eof) 1
		catch {close $var(fd); unset var(fd)}
	    }
	}
	PRfd {}
	default {
	    error "images cannot be produced from a \"$dataType\" data source"
	}
    }
    return $var(image)
}

proc Image_close {dataState renderState} {
    upvar #0 $dataState var

    if {[info exists var(cachefile)]} {
	catch {$var(image) configure -file [eval file join $var(cachefile)]}
    }
}

proc Image_renderer {dataState renderState img} {
    upvar #0 $dataState var

    if {!$var(eof)} return;	# Wait until image has been downloaded

    set win $var(cookie)
    if {[winfo class $win] == "Text"} {
	# We may have been here before
	catch {$win window create end -window [label $win.img -image $img]}
    } elseif {[winfo class $win] == "Label" ||  [winfo class $win] == "Button"} {
	if {[winfo name $win] == "label"} {
	    pack propagate [winfo parent $win] 1
	}
	$win configure -image $img
    }
}

proc Image_create {state} {
    upvar #0 $state var

    switch $var(HDRcontent-type) {
	image/x-xbitmap {return [image create bitmap]}
	image/x-portable-pixmap {return [image create photo -format ppm]}
	image/gif {return [image create photo -format gif]}
	image/jpeg {return [image create photo -format jpeg]}
	image/tiff {return [image create photo -format tiff]}
	image/png {return [image create photo -format png]}
	default {error "unknown image type \"$var(HDRcontent-type)\""}
    }
}

foreach {type suffixes} {
	image/x-xbitmap .xbm 
	image/x-portable-pixmap {.ppm .pnm .pbm}
	image/gif .gif
} {
    catch {
	PRregister_type $type $suffixes Image_process Image_renderer Image_close 99 image
    }
}

# Use Jan Nijtman's Img package if it is available

if {[lsearch [package names] Img] >= 0} {
    foreach {type suffixes} {
	image/jpeg {.jpeg .jpg}
	image/tiff .tiff
	image/png .png
    } {
	catch {
	    PRregister_type $type $suffixes Image_process Image_renderer Image_close 99 image
	}
    }
}

# A catch-all content-type handler.

# This module uses the prefix "Generic" for all
# procedure names and global variables.

catch {PRregister_type */* {} PRnoop PRnoop Generic_close 99 unknown.document}

proc Generic_close {dataState renderState} {
    upvar #0 $dataState var

    set state generic_$dataState
    upvar #0 $state gen
    toplevel .$state
    wm title .$state "Save Downloaded File"
    label .$state.msg -text "Please choose a file for downloaded document:"
    pack .$state.msg -side top -anchor nw -fill x
    array set gen [list win .$state.fsbox top .$state ignoreMissing 1  exit Generic_Exit open Generic_OpenFile  flist [list [list Everything *]]  ]
    Fsbox $state
    pack $gen(win) -side top -fill both -expand yes
    grab .$state
}

proc Generic_Exit {state} {
    upvar #0 $state var
    grab release $var(top)
    destroy $var(top)
    unset var
}

proc Generic_OpenFile {state} {
    upvar #0 $state var
    Generic_Exit $state
}

# Handler to pass data to "helper applications"
# It will be easy to implement passing data as it comes
# off the network, but for the moment we'll just point the
# helper app at the cache file

proc Helper_register {type ext cmd} {
   global Helpers

    set Helpers($type) $cmd
    catch {
	PRregister_type $type $ext {} {} Helper_close 40
    }
}

# Nothing to do (for now).  A future version might pipe the
# data to the help app.

proc Helper_close {dataState renderState} {
    global Helpers tcl_platform surfit_feedback
    upvar #0 $dataState var

    if {!$var(eof)} return; # Wait until all data is loaded

    # Find the file that the data has been written to
    # by the cache, and then start the helper app
    set cmd [format $Helpers($var(HDRcontent-type)) [eval file join [Cache_get_file $var(url)]]]
    # Escape special characters
    regsub -all {\\} $cmd {\\\\} cmd
    regsub -all {\$} $cmd {\\$} cmd
    regsub -all {"} $cmd {\\"} cmd
    regsub -all {\{} $cmd {\\\{} cmd
    regsub -all {\}} $cmd {\\\}} cmd
    regsub -all {;} $cmd {\\;} cmd
    # Shell constructs
    regsub -all {'} $cmd {"} cmd
    catch {eval exec $cmd &}
    set surfit_feedback($var(cookie)) "Started helper application [lindex $cmd 0]"
}

#
# Register helper apps specified in the user's .mailcap
# BUG: This will only work under Unix.
# Will have to be handled internally on Windows, Mac
#

if {![catch {open $env(HOME)/.mailcap r} mailcapfd]} {
    set mailcap [read -nonewline $mailcapfd]
    close $mailcapfd
    foreach line [split $mailcap \n] {
	if {[regexp {([^;]*);[ 	]*(.*)} $line all type cmd]} {
	    Helper_register $type {} $cmd
	}
    }
    unset mailcapfd
    unset mailcap
}


} \
hypetool.tcl {# $Id: hypetool.tcl,v 1.8 1996/08/06 11:54:27 steve Exp $
#
#	Copyright (c) 1996 Australian National University
#
#--------
#
#	PASTIME Project
#	Cooperative Research Centre for Advanced Computational Systems
#	COPYRIGHT NOTICE AND DISCLAIMER.
#
# Copyright (c) 1995,1996 ANU and CSIRO on behalf of the participants in 
# the  CRC for Advanced Computational Systems ('ACSys').
# 
# ACSys makes this software and all associated data and documentation
# ('Software') available free of charge for non-commercial purposes only. You
# may make copies of the Software but you must include all of this notice on
# any copy.
# 
# The Software was developed for research purposes and ACSys does not warrant
# that it is error free or fit for any purpose.  ACSys disclaims any
# liability for all claims, expenses, losses, damages and costs any user may
# incur as a result of using, copying or modifying the Software.
#

### This module provides support for interacting with hypertools.
### Several classes of tools are defined, each one having its own
### protocol.  The protocol may specify which parts are to be exposed
### to applets (if any).  Hypertools communicate by means of the Tk send command.
### Note that SurfIt! itself is a hypertool.

# All procedures in this module are prefixed by "Hypertool"
# All global variables are prefixed by "HT"

package provide hypertools [lindex {$Revision: 1.8 $} 1]

proc HT_init {} {
    CM_init
    NEWS_init
}

# The Continuous Media Hypertool manager prefixes all procedures and global
# variables with CM

# A program which implements a particular class of hypertool
# may register itself or may be pre-registered.  We probably need
# some sort of hypertool broker.

# The array "HThypertools" holds the registration of hypertools,

proc Hypertool_register {class type tool program init} {
    global HThypertools HTtypes

    set HThypertools($class) [list $tool $program $init]
    foreach t $type {
	set HTtypes($t) $class
    }
}

proc Hypertool_remove {class} {
    global HThypertools HTtypes
    if {[catch {unset HThypertools($class)}]} {error "hypertool class \"$class\" not registered"}
    foreach {k e} [array get HTtypes] {
	if {$e == $class} {unset HTtypes($k)}
    }
}

# Start a hypertool, optionally sending it a command once started

proc Hypertool {class {cmd {}}} {
    global HThypertools

    if {[info exists HThypertools($class)]} {
	if {[lsearch [winfo interps] [lindex $HThypertools($class) 0]] == -1} {
	    # We need to startup the tool
	    exec [lindex $HThypertools($class) 1] &
	    # Wait for the application to start
	    after idle Hypertool_check [lindex $HThypertools($class) 0]
	    tkwait variable HT[lindex $HThypertools($class) 0]
	    # Send the initialisation code
	    if {[lindex $HThypertools($class) 2] != {}} {
		send [lindex $HThypertools($class) 0] [lindex $HThypertools($class) 2]
	    }
	    # Send pending command
	    if {$cmd != {}} {
		send [lindex $HThypertools($class) 0] $cmd
	    }
	}
	return [lindex $HThypertools($class) 0]
    } else {error "hypertool \"$class\" class is not defined"}
}

proc Hypertool_check {tool} {
    global HTwait
    upvar #0 HT$tool var

    # Is the application responding?
    if {[catch {send $tool info tclversion}]} {
	if {![info exists HTwait($tool)]} {
	    set HTwait($tool) 9	;# wait up to 2 seconds
	}
	incr HTwait($tool) -1
	if {$HTwait($tool)} {
	    after 250 Hypertool_check $tool
	} else {
	    puts stderr "unable to contact hypertool for $class application"
	    catch {unset HTwait($tool)}
	    set var 0	;# terminates pending tkwait
	}
    } else {
	catch {unset HTwait($tool)}
	set var 1	;# terminates pending tkwait
    }
}

# Wrapper to send commands to hypertools

proc Hypertool_send {class cmd} {
    global HThypertools

    # Try straight-forward send first,
    # if unsuccessful may need to (re-)start tool
    if {[catch {send [lindex $HThypertools($class) 0] $cmd}]} {
	if {[catch {Hypertool $class $cmd} err]} {
	    puts "send to hypertool failed: \"$err\""
	}
    }
}

###
### Continuous Media hypertool - Berkeley's CMT cmplayer
###

# Here we define a class of hypertools for handling continuous media.
# The Berkeley Plateau Multimedia Research Group's Continuous Media
# Toolkit contains an application, cmplayer, which is representative
# of this class.

proc CM_init {} {
    global surfit

if {[info exists surfit(CFGcm_handler)] && [string match *cmplayer* $surfit(CFGcm_handler)]} {
    Hypertool_register CM  {video/mpeg video/mjpg audio/basic audio/audiofile application/x-cmt-script movie/cmt}  cmplayer cmplayer {

	proc Hypertool_NewFile {file {action {}}} {
	    global currentScript Tags lts;

	    ResetPlayer;
	    Extern_SetActionCmd $action

	    set Tags(title) [file tail $file]
	    set currentScript $file;

	    catch {grab .status}
	    setCursor watch;

	    PutDisp "Contacting media server..."

	    if { [catch {CMApp_ReadURL $file} script] == 0 } {
		if { [catch {CMApp_ResolveMovie $script} script] == 0 } {
		    SetupScript $script;
		    SetupForPlay;
		    grab release .status;
		    setCursor {};
		} else {
		    ResetPlayer
		    grab release .status
		    setCursor {}
		    error [list Can't play script $script]
		}
	    } else {
		ResetPlayer;
		grab release .status;
		setCursor {};
		error "Can't play $file.\nNot a valid filename or URL";
	    }
	}

	# Simulates functionality of CMApp_GetsClean
	proc Hypertool_CleanInput {script} {
	    set out {}
	    foreach line $script {
		if {$line != {} && ![string match #* [lindex $line 0]]} {lappend out $line}
	    }
	    return $out
	}

	# Simulates functionality of CMApp_ReadMovie
	# Input is a CMT Movie Script, output a PlayList

	proc Hypertool_ReadScript {script} {
	    global kCMMovieHeader kCMMovie_Types kCMMovie_Version

	    set script [split $script \n]
	    if {![string match $kCMMovieHeader  [lindex $script 0]]} {
		error "Not a $kCMMovie_Version script"
	    }
	    set script [Hypertool_CleanInput [lrange $script 1 end]]
	    set line [lindex $script 0]; set script [lrange $script 1 end]
	    if {[lindex $line 0] != "movie"} {
		error "First movie line not 'movie' command"
	    }
	    set line [CMApp_ArgsAndSwitchList {{tags {}}} [lrange $line 1 end]]
	    set numTracks [llength $line]

	    set j 0
	    foreach I $line {
		if {[lsearch -exact $kCMMovie_Types $I] == -1} {
		    error "Bad track type: $I in line: $line"
		}
		set tracks($j) $I
		# puts "tracks($j) = $tracks($j)"
		incr j
	    }

	    foreach line $script {
		switch -- [lindex $line 0] {
		    clip {
			set trk [lindex $line 1]
			if {![CMApp_CkNum $trk] || $trk < 0 || $trk >= $numTracks} {
			    error "bad track in line: $line"
			}
			append tracks($trk) "\n	"
			lappend tracks($trk) [CMApp_ParseClipLine [lrange $line 2 end]]
		    }
		    default {
			error "Not a clip line: $line"
		    }
		}
	    }

	    lappend out $tags
	    for {set j 0} {$j < $numTracks} {incr j} {
		append out "\n"
		lappend out $tracks($j)
	    }

	    return $out
	}
    
	proc Hypertool_NewScript {script {action {}}} {
	    global currentScript Tags lts;

	    ResetPlayer;
	    Extern_SetActionCmd $action

	    set Tags(title) {}
	    set currentScript $script;

	    catch {grab .status}
	    setCursor watch

	    if {[catch {CMApp_ResolveMovie [Hypertool_ReadScript $script]} script] == 0 } {
		SetupScript $script;
		SetupForPlay;
		grab release .status
		setCursor {}
	    } else {
		global errorInfo
		puts "Error: $script\n$errorInfo"
		tk_dialog .error Error "Error:\n$script" "" 0 "Reset Cmplayer"
		ResetPlayer
		grab release .status
		setCursor {}
	    }
	}
    }

    # First, we define content-type handlers for continuous media document
    # types.  We also include a content-type handler for CMT scripts.
    # Then, various procedures are defined for communicating with the
    # hypertool (the protocol).

    # Register content-type handlers.  These handle files that have been
    # copied to the local cache (similar to images) by passing them to
    # the hypertool.  Unfortunately we have the latency of copying the file 
    # first :-(
    # Future: 1) handle data as it arrives from the network
    #	  2) Display inline video (shouldn't be too hard)

    # Of course, what you really want to do is use CMT scripts.

    foreach {type suffixes} {
	video/mpeg .mpg
	video/mjpg .mjpg
	audio/basic .au
	audio/audiofile .acf
	application/x-cmt-script .script
    } {catch {PRregister_type $type $suffixes CM_process {} CM_close 80}}

    # Register cmtp: network protocol handler.  Of course, we'll just pass
    # the URL to cmplayer.  cmtp: URLs don't behave quite like normal URLs;
    # they are, in a way, documents themselves.

    PRregister_type movie/cmt .cmt PRpassNpost CM_movie_handler CM_movie_close 80

    PRregister_scheme cmtp CM_protocol PRparser_collapse 80
}

}

proc CM_file {state filename} {
    upvar #0 $state var

    set var(eof) 1
    return $filename
}

proc CM_protocol {state method data put_data} {
    upvar #0 $state var
    upvar 2 $data data_return

    # Just send the URL straight to cmplayer
    Hypertool_send CM "Hypertool_NewFile [list $var(url)]\; play"
    set var(eof) 1
    return {}
}

proc CM_movie_handler {dataState renderState data} {
    upvar #0 $dataState var
    append var(CMmovie) $data
}

proc CM_movie_close {dataState renderState} {
    upvar #0 $dataState var

    Hypertool_send CM "Hypertool_NewScript [list $var(CMmovie)]\; play"
    unset var(CMmovie)
}

# If this is a CMT script then store the text
# so that CM_close can send it to cmplayer
# Otherwise ignore it since we'll use the cached file

proc CM_process {dataState dataType data} {
    upvar #0 $dataState var

    switch $dataType {
	PRfile {
	    if {[info exists var(CMprocessed)]} {
		return {}
	    } else {set var(CMprocessed) 1}

	    Hypertool_send CM "Hypertool_NewFile [list $var(file)]\; play"
	    # We're done so override normal protocol handling
	    set var(eof) 1
	    catch {close $var(fd); unset var(fd)}
	}
	PRfd {}
	default {
	    append var(CMscript) $data
	}
    }
    return {}
}

proc CM_close {dataState renderState} {
    upvar #0 $dataState var

    if {[info exists var(CMclosed)]} {return}  else {set var(CMclosed) 1}

    switch $var(dataType) {
	PRfile {# Was played in CM_process}
	PRfd {
	    # Play it
	    Hypertool_send CM "Hypertool_NewFile [list $var(cachefile)]\; play"
	}
	default {
	    Hypertool_send CM "Hypertool_NewScript [list $var(CMscript)]\; play"
	    unset var(CMscript)
	}
    }
}

###
###  USENET News hypertool - nn-tk
###

proc NEWS_init {} {
    global surfit

    if {[info exists surfit(CFGnews_handler)] && [string match *nn-tk* $surfit(CFGnews_handler)]} {
	Hypertool_register USENET {} nn nn-tk {}

	PRregister_scheme nntp nn-tk_handler PRparser_generic 80
	PRregister_scheme news nn-tk_handler PRparser_generic 80
    }
}

proc nn-tk_handler {state method data put_data} {
    upvar #0 $state var
    upvar 2 $data data_return

    # method "put" might indicate posting an article - not yet supported

    # Normal URL parsing doesn't apply
    regexp {(news|nntp):(.*)} $var(url) all protocol newsgroup

    Hypertool_send USENET "eval .groups.list tag remove \[.groups.list tag ranges sely\]
	set gm_type 'g' 
	set ev_param $newsgroup 
	ev_type_menu 
	rec_c \$EV_FUNCT \$token(K_SEL_GROUP)"

    return {}
}


} \
mailto.tcl {# $Id: mailto.tcl,v 1.6 1996/08/06 11:54:29 steve Exp $
#
#	Copyright (c) 1996,1997 Australian National University
#
#--------
#
#	PASTIME Project
#	Cooperative Research Centre for Advanced Computational Systems
#	COPYRIGHT NOTICE AND DISCLAIMER.
#
# Copyright (c) 1995,1996 ANU and CSIRO on behalf of the participants in 
# the  CRC for Advanced Computational Systems ('ACSys').
# 
# ACSys makes this software and all associated data and documentation
# ('Software') available free of charge for non-commercial purposes only. You
# may make copies of the Software but you must include all of this notice on
# any copy.
# 
# The Software was developed for research purposes and ACSys does not warrant
# that it is error free or fit for any purpose.  ACSys disclaims any
# liability for all claims, expenses, losses, damages and costs any user may
# incur as a result of using, copying or modifying the Software.
#
###
#
# Handlers for mailto: URLs.  Included are a basic handler which uses
# SMTP directly, and a hypertool handler which uses exmh.

package provide mailto [lindex {$Revision: 1.6 $} 1]

proc mailto_init {} {
    global surfit

    # Define a hypertool handler using exmh, if configured

    if {[info exists surfit(CFGmailto_handler)] && [string match *exmh* $surfit(CFGmailto_handler)]} {
	PRregister_scheme mailto Exmh_mailto PRparser_generic 80
	Hypertool_register SMTP {} exmh exmh {}
    } else {
	# Generic-brand version
	catch {
	    PRregister_scheme mailto SMTP_protocol PRparser_generic 20
	}
    }
}

proc Exmh_mailto {state method data put_data} {
    upvar #0 $state var

    Hypertool_send SMTP "Sedit_Mailto $var(url)"
    set var(eof) 1
    return {}
}


##########################################################################
## Authors: Jonathan Perret and Gerald Lester				##
## Purpose: This sends SMTP mail	
##    Date: December 27, 1995						##
##########################################################################

## Send SMTP to mail host.
proc SMTP_sendmail {smtphost toList from subject body {trace 0}} {
	if $trace then {
		puts stdout "Connecting to $smtphost:25"
	}
	set sockid [socket $smtphost 25]
	puts $sockid "HELO $smtphost\nMAIL From:<$from>"
	flush $sockid
	set result [gets $sockid]
	if $trace then {
		puts stdout "MAIL From:<$from>\n\t$result"
	}
	foreach to $toList {
	    puts $sockid "RCPT To:<$to>"
	    flush $sockid
	}
	set result [gets $sockid]
	if $trace then {
		puts stdout "RCPT To:<$to>\n\t$result"
	}
	puts  $sockid "DATA "
	flush $sockid
	set result [gets  $sockid]
	if $trace then {
		puts stdout "DATA \n\t$result"
	}
	puts  $sockid "From: <$from>"
	puts  $sockid "To: <$to>"
	puts  $sockid "Subject: $subject"
	puts  $sockid "\n"
	foreach line [split $body  "\n"] {
		puts  $sockid "[join $line]"
	}
	puts  $sockid "."
	puts  $sockid "QUIT"
	flush $sockid
	set result [gets  $sockid]
	if $trace then {
		puts stdout "QUIT\n\t$result"
	}
	close $sockid 
	return;
}

###
### Protocol handler for mailto: URLs
### Author: Steve Ball
###

# Hopefully this protocol handler will be overridden by a 
# mail user agent acting as a hypertool.

# The mailto: URL causes a mail message to be composed and sent
# to the given address.  This protocol never returns data to be
# rendered

catch {
    PRregister_scheme mailto SMTP_protocol PRparser_generic 20
}

proc SMTP_protocol {state method data put_data} {
    upvar #0 $state var
    upvar 2 $data data_return

    # Normal URL parsing does not apply to mail addresses
    regexp {mailto:([^@]+)@(.*)} $var(url) all user host

    # Construct the composition window

    set p [toplevel .mailto$state]
    wm title $p "Send Mail to User $user at $host"

    label $p.subjLabel -text "Subject: "
    entry $p.subject -width 40
    grid $p.subjLabel -row 0 -column 0
    grid $p.subject -sticky ew -row 0 -column 1

    label $p.toLabel -text "To: "
    entry $p.to -width 40
    grid $p.toLabel -row 1 -column 0
    grid $p.to -sticky ew -row 1 -column 1
    $p.to insert 0 "$user@$host"

    label $p.ccLabel -text "Cc: "
    entry $p.cc -width 40
    grid $p.ccLabel -row 2 -column 0
    grid $p.cc -sticky ew -row 2 -column 1

    label $p.bodyLabel -text "Message to send:"
    text $p.body
    grid $p.bodyLabel -row 3 -column 0 -columnspan 2 -sticky ew
    grid $p.body -row 4 -column 0 -columnspan 2 -sticky news

    set q [frame $p.ctl]
    grid $q -row 5 -column 0 -columnspan 2 -sticky w
    button $q.send -text "Send Message" -command "SMTP_send $p $user $host"
    button $q.cancel -text "Cancel Message" -command "SMTP_cancel $p $user $host"
    grid $q.send $q.cancel -sticky w

    return {}
}

proc SMTP_send {win user host} {
    global surfit

    if {[info exists surfit(CFGsmtphost)]} {set host $surfit(CFGsmtphost)}

    SMTP_sendmail $host  [$win.to get] $surfit(CFGusername)  [$win.subject get]  [$win.body get 1.0 end]

    destroy $win
}

proc SMTP_cancel {win user host} {
    if {![tk_dialog $win.confirm {Cancel Message?}  "Are you sure that you want to cancel sending this message  to $user at $host?" {} 1 Yes No]} {
	destroy $win
    }
}


} \
internal.tcl {# $Id: internal.tcl,v 1.9 1996/08/06 11:54:28 steve Exp $
#
#	Copyright (c) 1996,1997 Australian National University
#
#--------
#
#	PASTIME Project
#	Cooperative Research Centre for Advanced Computational Systems
#	COPYRIGHT NOTICE AND DISCLAIMER.
#
# Copyright (c) 1995,1996 ANU and CSIRO on behalf of the participants in 
# the  CRC for Advanced Computational Systems ('ACSys').
# 
# ACSys makes this software and all associated data and documentation
# ('Software') available free of charge for non-commercial purposes only. You
# may make copies of the Software but you must include all of this notice on
# any copy.
# 
# The Software was developed for research purposes and ACSys does not warrant
# that it is error free or fit for any purpose.  ACSys disclaims any
# liability for all claims, expenses, losses, damages and costs any user may
# incur as a result of using, copying or modifying the Software.
#
###
# Internal procedure declarations for surfit.  It is safe to re-source this
# file while SurfIt! is running.

# Documented procedures are prefixed with "SurfIt"
# Internal procedures and variables in this module are prefixed with "surfit"

package provide internals [lindex {$Revision: 1.9 $} 1]

proc surfit_init {} {
    ###
    ### Override routines for other modules
    ###

    if {[info commands surfit_setup_anchor] == {}} {
	rename HMlink_setup surfit_setup_anchor
    }

    proc HMlink_setup {state win href} {
	surfit_setup_anchor $state $win $href

	regsub -all {%} $href {%%} href2	;# make bind happy
	set top [SurfIt_hyperpage $win]
	$win tag bind L:$href <Enter> "+set surfit_highlightURL($top) \{$href2\}"
	$win tag bind L:$href <Leave> "+set surfit_highlightURL($top) {}"
    }

    proc HMlink_window_setup {but link} {
	set top [SurfIt_hyperpage $but]
	regsub -all {%} $link {%%} link2	;# make bind happy
	bind $but <Enter> "+set surfit_highlightURL($top) \{$link2\}"
	bind $but <Leave> "+set surfit_highlightURL($top) {}"
    }

    # Override the library link-callback routine for the sample app.

    proc HMlink_callback {state win href} {
	global surfit_url_history_idx surfit_currPage
	upvar #0 $state var

	# Check if only a destination anchor was specified
	if {[regexp -nocase {^#(.+)} $href all fragment]} {
	    HMgoto $state $fragment
	    return
	}

	set hyperpage [SurfIt_hyperpage $win]
	incr surfit_url_history_idx($hyperpage)
	# Set insert position for applets
	upvar #0 PS$hyperpage newState
	set newState(index) [$win index insert]	;# Assumes Button-1 invokes link
	set newState(referer) $surfit_currPage($hyperpage)
	set newState(win) $hyperpage
	SurfIt_loadURL [URI_merge $var(base) $href PRpath_handler] $hyperpage
    }

    # Supply an image callback function
    # NB. This function bypasses HMgot_image

    proc HMset_image {state win handle src} {
	global surfit_currPage
	set hyperpage [SurfIt_hyperpage $handle]
	upvar #0 $state var
	upvar #0 PS$hyperpage altvar

	set altvar(referer) $surfit_currPage($hyperpage)
	#   loadDocument [URI_merge $var(base) $src PRpath_handler] $handle {} fileevent surfit_image_start
	if {[catch {PRloadDocument [URI_merge $var(base) $src PRpath_handler] $handle {} surfit_image_start} result]} {
	    puts "unable to load inline due to \"$result\""
	    # Do something about not being able to load the image
	}
    }

    # Callback for loading applets.

    proc HMload_applet {state win scriptURL level {formID {}}} {
	upvar #0 $state var

	set uniq [HMgetUniqueNum $state]
	upvar #0 $state$uniq script
	set script(AppWin) $win
	set script(index) [$win index $var(S_insert)]
	if {$level == "form"} {
	    set script(form) $formID
	    set script(renderState) $state
	}
	PRloadDocument [URI_merge $var(base) $scriptURL PRpath_handler] $win $state$uniq surfit_image_start
    }

    proc HMload_icon {state win icon} {
	upvar #0 $state var

	if {![catch {label $win.[HMgetUniqueNum $state] -image ICON$icon} item]} {
	    $win window create $var(S_insert) -window $item -align center
	} else {
	    $win configure -image {}
	}
    }

    proc HMset_icon {state win icon} {
	if {[catch {$win configure -image ICON$icon}]} {
	    $win configure -image {}
	}
    }

    # Override the default HMsubmit_form proc to handle additional
    # features.  If there is an applet attached to the form then
    # call into it and override query if necessary

    proc HMsubmit_form {state win form_id param query} {
	upvar #0 $state var
	upvar #0 AppFormsMap$win formmap

	set method get
	HMextract_param $param method
	set method [string tolower $method]

	# Call into applet if attached
	if {![catch {set formmap($form_id)} slave]} {
	    if {[$slave eval info commands HMsubmit_form] != {}} {
		if {[catch {$slave eval HMsubmit_form $method \{$query\}} query]} {
		    # Cancel the form submission
		    puts "HMsubmit_form: submission cancelled due to \"$query\""
		    return
		}
	    }
	}

	switch $method {
	    get {
		set result ""
		set sep ""
		foreach i $query {
		    append result $sep [HMmap_reply $i]
		    if {$sep != "="} {set sep =} {set sep &}
		}
		if {[HMextract_param $param action]} {
		    SurfIt_loadURL [URI_merge $var(base) $action?$result PRpath_handler] [SurfIt_hyperpage $win]
		}
	    }
	    put {
		puts "PUT method not yet supported"
	    }
	    default {error "unknown method \"$method\""}
	}
    }

    # Let the attached applet know that the form has been reset
    # This gives it a chance to reset its internal state

    proc HMreset_form {win form_id} {
	upvar #0 HM[SurfIt_hyperpage $win]
	upvar #0 AppFormsMap$win formmap

	# Call into applet if attached
	if {![catch {set formmap($form_id)} slave]} {
	    catch {$slave eval HMreset_form}
	}
    }

    proc PRhandlerFinished {state size} {
	global surfit_pageStatus
	upvar #0 $state var
	upvar #0 PS$var(cookie) intState

	if {[info exists intState(win)]} {
	    set win $intState(win)
	} elseif {[interp exists $var(cookie)]} {
	    upvar #0 AppData$var(cookie) app
	    set win $app(win)
	} else {set win $var(cookie)}
	set hyperpage [SurfIt_hyperpage $win]
	catch {incr surfit_pageStatus($hyperpage,loaded) $size}

	# Make sure we don't call this twice
	if {$var(eof) && ![info exists var(surfit_done)]} {
	    incr surfit_pageStatus($hyperpage,numdocs) -1
	    set var(surfit_done) 1
	}
	SurfIt_setFeedback $hyperpage
    }

    proc PRhandlerError {state result} {
	global surfit_pageStatus errorInfo
	upvar #0 $state var
	upvar #0 PS$var(cookie) intState

	debug_puts [list PRhandlerError $state $result]

	set cause $errorInfo	;# Save the text
	if {[info exists intState(win)]} {
	    set win $intState(win)
	} elseif {[interp exists $var(cookie)]} {
	    upvar #0 AppData$var(cookie) app
	    set win $app(win)
	} else {set win $var(cookie)}
	set hyperpage [SurfIt_hyperpage $win]

	if {$win == $hyperpage} {
	    # Has a document load been scheduled?
	    #if {[info exists surfit_pageStatus($win,pending)]} {
		#set url [lindex $surfit_pageStatus($win,pending) 0]
		#set type [lindex $surfit_pageStatus($win,pending) 1]
		#unset surfit_pageStatus($win,pending)
		#SurfIt_loadURL $url $win $type
	    #}
	    return 0
	} else {

	    if {[info exists var(surfit_started)]} {
		incr surfit_pageStatus($hyperpage,numdocs) -1
		if {$var(HDRcontent-length) != {}} {incr surfit_pageStatus($hyperpage,total) -$var(HDRcontent-length)}
		# Should also subtract the # bytes loaded so far for this document
	    }
	    SurfIt_setFeedback $hyperpage

	    # Terminate loading
	    surfit_terminate_doc $state
	    return 1
	}
    }

    # From html_library/sample.tcl:
    # downloading fonts can take a long time.  We'll override the default
    # font-setting routine to permit better user feedback on fonts.  We'll
    # keep our own list of installed fonts on the side, to guess when delays
    # are likely

    proc HMset_font {win tag font} {
	global surfit_feedback surfit_Fonts
	set hyperpage [SurfIt_hyperpage $win]
	set FBwin [winfo toplevel $win].bottom-panel.feedback
	if {![info exists surfit_Fonts($font)]} {
	    set surfit_Fonts($font) 1
	    $FBwin configure -fg blue
	    set surfit_feedback($hyperpage) "downloading font $font"
	    update idletasks
	}
	catch {$win tag configure $tag -font $font} fonterr1
	catch {$win tag configure $tag -kanjifont [HMfont2kfont $font]} fonterr2
	# For the majority of wishes -kanjifont will
	# produce an error, so don't check for that error
	# wishes that do have Kanji support will have to supply an alternative
	# for this procedure
	if {[string compare {} $fonterr1]} {
	    $FBwin configure -fg red
	    set surfit_feedback($hyperpage) $fonterr1
	} else {
	    $win tag lower $tag mark
	}
    }

    proc HMget_banner {win} {
	set banner [winfo toplevel $win].banner
	if {[grid info $banner] == {}} {
	    grid $banner -row 3 -sticky ew
	}
	return $banner
    }

    # Override the interp_delete_slave procedure to detect when applets are destroyed

    if {[info commands surfit_interp_delete_slave] == {}} {
	rename interp_delete_slave surfit_interp_delete_slave
    }
    proc interp_delete_slave {slave} {
	global surfit_Applets

	surfit_interp_delete_slave $slave

	if {[set object $surfit_Applets($slave)] == "browser"} {
	    set menu .ctl-panel.applets.m
	} else {
	    if {[winfo exists $object]} {
		set menu [winfo toplevel $object].top-ctl-panel.applets.m
	    } else return;
	}
	$menu delete "Applet $slave"
    }

    # Augment unknown procedure to catch malformed comments
    # This really belongs in the htmllib module

    if {[info commands surfit_unknown] == {}} {
	rename unknown surfit_unknown
    }
    proc unknown {cmdName args} {
	if {[regexp {^HMtag_!--(.+)} $cmdName all extra]} {
	    lassign $args state win param text
	    uplevel HMtag_!-- $state $win [list $extra$param] $text
	    return
	}
	uplevel 1 surfit_unknown $cmdName $args
    }
}

###
### Register the HTML parser as a content-type handler for text/html
###

catch {
    PRregister_type text/html {.html .htm} SurfIt_parse_html SurfIt_startrender SurfIt_close 99 text.document
}

set surfit(S_update) 10	;# num of tags to process at a time

proc SurfIt_parse_html {state dataType data} {
    global surfit
    upvar #0 $state var

    debug_puts [list SurfIt_parse_html $state ...]

    # The PRfd or PRfile read handler will pass data to this proc

    switch $dataType {
	PRfd -
	PRfile -
	PRdata {
	    # Find start tag
	    if {[info exists var(start)]} {
		set start $var(start)
	    } else {
		set start hmstart
		set var(start) hmcont
	    }
	    # Find cache array for post processed data
	    if {![info exists var(postArray)]} {
		set var(postArray) SurfIt_$state
		Cache_add_post $state $var(postArray) text/html
	    }
	    set var(nextElement)  [HMparse_html $data $var(postArray) $var(nextElement) $surfit(S_update) leftovers HMrender $start]

	    set var(putback_data) $leftovers
	    return $var(postArray)
	}
	PRpost {
	    set var(postArray) $data
	}
	default {
	    error "unknown data type \"$dataType\""
	}
    }
    return {}
}

proc SurfIt_startrender {dataState renderState data} {
    upvar #0 $dataState var

    debug_puts [list SurfIt_startrender $dataState $renderState ...]

    set var(postArray) $data
    if {![info exists var(renderID)]} {
	set var(renderID) {}
	SurfIt_render $dataState
    }
    SurfIt_setFeedback $var(cookie)
}

# Make this as lean and mean as possible
# renderElement is the index of the next (group of) element(s)
# to render

# Workaround: this used to be an idle handler (after idle ...)
# and called update after setting the next handler.  However, on
# large docs & Mac it crashed.  'after 1' waits almost no time
# between rendering blocks and allows interactivity.

proc SurfIt_render {state} {
    upvar #0 $state var
    upvar #0 $var(postArray) post
    if {![catch {set post([incr var(renderElement)])} html]} {
	regsub -all {\\win\\} $html $var(cookie) html
	catch {eval $html}
	set var(renderID) [after 1 SurfIt_render $state]
    } else {
	incr var(renderElement) -1	;# reset to original value
	catch {unset var(renderID)}
	if {$var(eof)} {
	    global surfit_pageStatus
	    HMrender $var(cookie) {hmstart} {/} {} {}
	    set surfit_pageStatus($var(cookie),busy) 0
	    SurfIt_setFeedback $var(cookie)
	}
    }
}

proc surfit_image_start {dataState renderState} {
    global surfit_pageStatus
    upvar #0 $dataState var
    upvar #0 PS$var(cookie) intState

    if {[info exists intState(win)]} {set win $intState(win)}  else {set win $var(cookie)}
    set hyperpage [SurfIt_hyperpage $win]
    incr surfit_pageStatus($hyperpage,numdocs)
    lappend surfit_pageStatus($hyperpage,docList) $dataState
    # Sometimes the content-length is not set even at this stage.
    # If content-length is subsequently set the feedback will be wrong
    catch {incr surfit_pageStatus($hyperpage,total) $var(HDRcontent-length)}
    set var(surfit_started) 1
    SurfIt_setFeedback $hyperpage
}

### Display HTML source.  Must have already been parsed by HTML parser.

catch {PRregister_type surfit/htmlsource {} PRnoop SurfIt_source_start SurfIt_close 99}

proc SurfIt_source_start {dataState renderState data} {
    upvar #0 $dataState var
    set var(postArray) $data
    # Prepare Text widget
    $var(cookie) configure -wrap none
    $var(cookie) tag configure bold -font [HMx_font times 14 bold r]
    $var(cookie) tag configure italic -font [HMx_font times 14 medium i]
    wm title [winfo toplevel $var(cookie)] "Source of $var(url)"
    if {![info exists var(renderID)]} {
	set var(renderID) {}
	SurfIt_sourceRender $dataState
    }
    SurfIt_setFeedback $var(cookie)
}

proc SurfIt_sourceRender {state} {
    upvar #0 $state var
    upvar #0 $var(postArray) post
    set elem [incr var(renderElement)]
    if {![catch {set post($elem)} html]} {
	regsub -all {\\win\\} $html $var(cookie) html
	regsub -all HMrender $html surfit_source_render html
	catch {eval $html}
	set var(renderID) [after 1 SurfIt_sourceRender $state]
    } else {
	incr var(renderElement) -1	;# reset to original value
	catch {unset var(renderID)}
	global surfit_pageStatus
	set surfit_pageStatus($var(cookie),busy) 0
	SurfIt_setFeedback $var(cookie)
    }
}

proc surfit_source_render {cookie tag not param text} {
    switch $tag {
	hmstart -
	hmcont {# Ignore - used by htmllib}
	default {
	    $cookie insert end "<$not$tag" bold
	    if {$param != {}} {$cookie insert end " $param" italic}
	    $cookie insert end ">" bold
	    #if {$not != ""} {$cookie insert end "\n"}
	    $cookie insert end $text
	}
    }
}

###
### Load a URL into a hyperpage.
###
### url is the URL of the document to load
### win is the toplevel of the hyperpage to load the document into
###
### NB. That the given URL may not cause anything to be loaded
### (or at least cause a page reset), so don't jump the gun and
### clear the page at this point.
###

proc SurfIt_loadURL {url win {type {}}} {
    global gk_library
    if {[info exists gk_library]} {
	gk_toAll surfit_realLoadURL $url $win $type
    } else {
	surfit_realLoadURL $url $win $type
    }
}

proc surfit_realLoadURL {url win {type {}}} {
    global surfit_feedback surfit_pageStatus
    set win [SurfIt_hyperpage $win]
    upvar #0 HM$win parser

    # Tell the applets about it
    Applet_anchor_activation $win $url

    # Check if a local destination anchor was specified
    if {[regexp -nocase {([^#]*)#(.+)} $url all base name]} {
	if {$base == {} || $base == $parser(base)} {
	    HMgoto $win $name
	    return
	}
    }

    # Invoke pre-load scripts...

    set surfit_feedback($win) "Loading $url"
    SurfIt_start_download $win

    if {![PRloadDocument $url $win HM$win surfit_new_page $type]} {
	set surfit_feedback($win) {}
    }
}

# For compatibility with Brent Welch's WebEdit (WebTk)

proc Url_DisplayNew {uri} {
    debug_puts [list Url_DisplayNew $uri]
    global main
    if {[info exists main] && [winfo exists $main]} {
	debug_puts [list displaying on main $main]
	SurfIt_loadURL $uri $main
    } else {
	set win {}
	foreach c [winfo children .] {
	    if {[regexp {^\.hyperdoc} $c]} {
		set win $c
		break
	    }
	}
	if {$win != {}} {
	    SurfIt_loadURL $uri $win
	} else {
	    surfit_create_window $uri
	}
    }
}

# Clear the hyperpage in readiness for the new document

proc surfit_new_page {dataState renderState} {
    global surfit_pageStatus surfit_feedback
    upvar #0 $dataState var

    # Clear the page for inlined documents (quality == 99).
    # Exceptions are Tcl applets, which must request for 
    # the page to be cleared explicitly.
    set win $var(cookie)
    if {$var(type) != "application/x-tcl" && $var(typeHandlerQuality) == 99} {
	if {$surfit_pageStatus($win,busy)} {
	    set surfit_pageStatus($win,pending) $dataState
	    surfit_terminate_loading $win
	}

	# Initialise the window
	Applet_destroy_page $win
	if {$surfit_pageStatus($win,scrollMode) != "line"} {
	    surfit_line_scrolling $win
	}
	array set surfit_pageStatus [list $win,busy 0 $win,numdocs 0 $win,total 0 $win,loaded 0 $win,dataState $dataState $win,docList {}]
	array set var {renderElement 0 nextElement 0}
	# This routine takes care of the rest
	Applet_new_page $var(url) $win
    }

    lappend surfit_pageStatus($win,docList) $dataState

    if {!$surfit_pageStatus($win,busy)} {
	set surfit_pageStatus($win,busy) $var(id)
    }

    # Initialise hyperpage load stats
    incr surfit_pageStatus($win,numdocs)
    set var(surfit_started) 1
    if {[info exists var(HDRcontent-length)] &&  $var(HDRcontent-length) != {} && $var(HDRcontent-length) != 0} {
	incr surfit_pageStatus($win,total) $var(HDRcontent-length)
	set surfit_feedback($win) "Displaying $var(url), $var(HDRcontent-length) bytes"
    }
}

# Special version of surfit_new_page for Applets

proc Applet_new_page {url win} {
    global surfit_url_history surfit_url_history_idx surfit_highlightURL surfit_pageStatus
    upvar #0 HM$win var

    HMreset_win $win
    set banner [winfo toplevel $win].banner
    catch {HMreset_win $banner}
    catch {$banner mark set $var(S_insert) 1.0
    $banner mark gravity $var(S_insert) left
    $banner insert end "\n " last
    $banner mark gravity $var(S_insert) right}
    catch {grid forget $banner}

    URI_parse $url scheme path query fragment
    set var(base) [URI_make $scheme $path {}]
    surfit_setOpenPage $win $url
    set surfit_highlightURL($win) {}
    update idletasks

    if {$fragment != {}} {
	HMgoto $win $fragment
    }

    set t [winfo toplevel $win]
    set p $t.top-ctl-panel
    set s $t.second-ctl-panel
    set hyperpage [SurfIt_hyperpage $win]

    # Adjust navigation history
    if {$url == [lindex $surfit_url_history($hyperpage) $surfit_url_history_idx($hyperpage)]} {
	# Do nothing - visiting previous page
    } elseif {$surfit_url_history_idx($hyperpage) == [llength $surfit_url_history($hyperpage)]} {
	lappend surfit_url_history($hyperpage) $url
    } else {
	set surfit_url_history($hyperpage)  [lrange $surfit_url_history($hyperpage) 0 [expr $surfit_url_history_idx($hyperpage) - 1]]
	lappend surfit_url_history($hyperpage) $url
    }

    # Adjust button states
    if {$surfit_url_history_idx($hyperpage) == 0} {
	$s.back configure -state disabled
    } elseif {[llength $surfit_url_history($hyperpage)] > 1} {
	$s.back configure -state normal
    }
    if {$surfit_url_history_idx($hyperpage) < [expr [llength $surfit_url_history($hyperpage)] - 1]} {
	$s.forward configure -state normal
    } else {
	$s.forward configure -state disabled
    }
    # These really only needs to happen once...
    if {$surfit_url_history_idx($hyperpage) >= 0} {
	$s.start configure -state normal
    }
    $p.doc.m entryconfigure Reload -state normal
    $p.doc.m entryconfigure Source -state normal
    $p.file.m entryconfigure "Save As..." -state normal
}

# Set background to a more subdued red
proc SurfIt_start_download {win} {
    global surfit_pageStatus
    [winfo toplevel $win].second-ctl-panel.stop configure -state normal -background #df0000
}

# Reset stop button, set background to default value
proc SurfIt_stopped_download {win} {
    global surfit_pageStatus
    [winfo toplevel $win].second-ctl-panel.stop configure -state disabled  -background $surfit_pageStatus($win,stopbg)
    set surfit_pageStatus($win,busy) 0
}

# Protocol handler callbacks.  These let us know when loads complete
# or fail

proc SurfIt_close {dataState renderState} {
    global surfit_pageStatus
    upvar #0 $dataState var

    SurfIt_setFeedback $var(cookie)
}

proc SurfIt_setFeedback {win} {
    global surfit_pageStatus surfit_feedback surfit_currPage

    set hyperpage [SurfIt_hyperpage $win]
    upvar #0 $surfit_pageStatus($hyperpage,dataState) var
    set FBwin [winfo toplevel $win].bottom-panel.feedback
    $FBwin configure -fg black
    # numdocs < 0 is an internal error...
    if {$surfit_pageStatus($hyperpage,numdocs) <= 0} {
	set surfit_feedback($hyperpage) "Document done"
	if {![info exists var(renderID)] && $var(renderElement) >= $var(nextElement)} {
	    SurfIt_stopped_download $win
	    # Let the applets know
	    Applet_pageloaded $hyperpage $surfit_currPage($hyperpage)
	}
    } elseif {$surfit_pageStatus($hyperpage,total) > 0} {
	set surfit_feedback($hyperpage) "Loading $surfit_pageStatus($hyperpage,numdocs) document[expr $surfit_pageStatus($hyperpage,numdocs) != 1 ? \"s\" : \"\"],  [expr $surfit_pageStatus($hyperpage,loaded) * 100 / $surfit_pageStatus($hyperpage,total)]% completed"
    } else {
	set surfit_feedback($hyperpage)  "Loading $surfit_pageStatus($hyperpage,numdocs) document[expr $surfit_pageStatus($hyperpage,numdocs) != 1 ? \"s\" : \"\"]"
    }

    if {[info exists var(renderID)]} {
	append surfit_feedback($hyperpage) ", currently rendering text"
    }
    update idletasks
}

# Terminate all data transfer associated with a hyperwindow(s)
# If no window given, terminate all loading

proc surfit_terminate_loading {{hyperwinList {}}} {
    global surfit_pageStatus
    debug_puts [list surfit_terminate_loading]
    if {$hyperwinList == {}} {
	# Find all hyperwindows
	foreach w [winfo children .] {
	    if {[regexp {.(hyperdoc[0-9]+)} $w all win]}  {lappend hyperwinList $win}
	}
    }
    foreach hyperwin $hyperwinList {
	if {[info exists surfit_pageStatus($hyperwin,docList)]} {
	    foreach d $surfit_pageStatus($hyperwin,docList) {
		surfit_terminate_doc $d
	    }
	}
	SurfIt_stopped_download $hyperwin
    }
}

proc surfit_terminate_doc {state} {
    upvar #0 $state var
    debug_puts [list surfit_terminate_doc $state called from [info level -1]]
    if {![catch {after cancel $var(renderID)}]} {
	unset var(renderID)
    }
    set stopped 0
    if {![catch {fileevent $var(fd) readable {}}]} {
	set stopped 1
    }
    if {![catch {fileevent $var(fd) writable {}}]} {
	set stopped 1
    }
    if {$stopped} {
	debug_puts [list surfit_terminate_doc: unsetting var(fd)]
	unset var(fd)
    }
}

# Applet Management: We need to keep track of which applets are attached
# at what level and what they are attached to.

# Applet_created callback procedure to detect when applets are created
# NB. This gets called before the applet's code is evaluated, hence the
# user may at least have some control over the applet!
# Applets are created at the hyperpage level by default.
# NB. #2: Numerals denote menu indexes, so had to add "Applet" to menu item labels

proc Applet_created {dataState renderState slave} {
    global surfit_Applets
    upvar #0 $dataState var
    upvar #0 $renderState ren

    if {[info exists ren(AppWin)]} {
	set surfit_Applets($slave) [set win $ren(AppWin)]	;# What the slave is attached to
    } else {
	set surfit_Applets($slave) [set win $var(cookie)]
    }
    set surfit_Applets($slave,url) $var(url)

    set menu [winfo toplevel $win].top-ctl-panel.applets.m

    # Hyperpage-level applets are added above the hyperwindow separator
    $menu insert "Window" cascade -label "Applet $slave" -menu [surfit_create_applet_menu $menu $slave $win $var(url)]
}

proc surfit_create_applet_menu {menu slave win url} {
    menu $menu.$slave -tearoff no
    $menu.$slave add command -label "Reload" -state disabled -command "
	Applet_destroy $slave
	Cache_expire_document $url
	SurfIt_loadURL $url $win
    "
    $menu.$slave add command -label "Suspend" -state disabled
    $menu.$slave add command -label "Delete" -command "Applet_destroy $slave"

    return $menu.$slave
}

# Define the Applet level change callback procedure

proc Applet_levelChange {slave old new} {
    global surfit_Applets

    set object $surfit_Applets($slave)
    set menu [winfo toplevel $object].top-ctl-panel.applets.m
    set menuitem [$menu entrycget "Applet $slave" -menu]

    if {$old == "hyperpage" && $new == "hyperwindow"} {
	# Simply move the menu
	$menu delete "Applet $slave"
	$menu add cascade -label "Applet $slave" -menu $menuitem
    } else {
	# Either hyperpage->browser or hyperwindow->browser

	# Remove the applet from the old object
	$menu delete "Applet $slave"
	destroy $menuitem

	# Add the applet to the browser object
	set menu .ctl-panel.applets.m
	$menu add cascade -label "Applet $slave" -menu [surfit_create_applet_menu $menu $slave browser $surfit_Applets($slave,url)]
	set surfit_Applets($slave) browser
    }
}

proc Applet_browserversion {} {
    global surfit
    return $surfit(version)
}

# Tclets may not load file: URLs

proc Applet_loadURI {url win {type {}}} {
    URI_parse $url scheme path query fragment
    if {$scheme == "file"} {
	error "cannot load file: scheme URL"
    }
    SurfIt_loadURL $url $win $type
}

###
### Window management routines
###

proc surfit_create_window {{open_url {}}} {
    global surfit_win_num surfit_url_history surfit_url_history_idx surfit_open_vars
    global surfit_currPage surfit_openPage surfit_feedback surfit_pageStatus
    global gk_library

    # Allocate unique name for this hyperdoc window
    set t .hyperdoc[incr surfit_win_num]

    toplevel $t
    wm title $t "Hyperdocument $surfit_win_num"
    wm iconname $t "Hyperdoc $surfit_win_num"
    set hyperpage [SurfIt_hyperpage $t]
    set surfit_url_history($hyperpage) {}
    set surfit_url_history_idx($hyperpage) -1
    set surfit_open_vars($hyperpage) {}
    set surfit_currPage($hyperpage) {}
    set surfit_openPage($hyperpage) {}
    array set surfit_pageStatus "$hyperpage,busy 0 $hyperpage,scrollMode line"

    set p [frame $t.top-ctl-panel -borderwidth 2 -relief raised]

    menubutton $p.file -text File -underline 0 -menu $p.file.m
    pack $p.file -side left
    menu $p.file.m
    $p.file.m add command -label "Save As..." -underline 0 -state disabled  -command "surfit_save_doc $t \$surfit_currPage($hyperpage)"
    if {[info exists gk_library]} {
	$p.file.m add command -label Quit -underline 0 -command _gk_properQuit
    } else {
	$p.file.m add command -label Dismiss -command "surfit_destroy_window $t" -underline 0
    }

    menubutton $p.doc -text Document -underline 0 -menu $p.doc.m
    pack $p.doc -side left
    menu $p.doc.m
    $p.doc.m add command -label "Reload" -command "
	Cache_expire_document \$surfit_currPage($hyperpage)
	SurfIt_loadURL \$surfit_currPage($hyperpage) $hyperpage
    " -state disabled -underline 0
    $p.doc.m add command -label "Source" -command "surfit_load_source $t" -state disabled -underline 0
    $p.doc.m add command -label "Presentation" -command "surfit_disp_mode presentation $t" -underline 0

    menubutton $p.applets -text "Applets" -menu $p.applets.m
    pack $p.applets -side left -anchor w
    menu $p.applets.m
    $p.applets.m add command -label "Forms" -state disabled
    $p.applets.m add command -label "Page" -state disabled
    $p.applets.m add command -label "Window" -state disabled

    grid $p -row 0 -sticky new

    set p [frame $t.second-ctl-panel]

    button $p.start -text "Start" -command "
	set surfit_url_history_idx($hyperpage) 0
	SurfIt_loadURL \[lindex \$surfit_url_history($hyperpage) 0\] $hyperpage
    " -state disabled
    pack $p.start -side left
    button $p.back -text "Back" -command "
	incr surfit_url_history_idx($hyperpage) -1
	SurfIt_loadURL \[lindex \$surfit_url_history($hyperpage) \$surfit_url_history_idx($hyperpage)\] $hyperpage
    " -state disabled
    pack $p.back -side left
    button $p.forward -text "Forward" -command "
	incr surfit_url_history_idx($hyperpage)
	SurfIt_loadURL \[lindex \$surfit_url_history($hyperpage) \$surfit_url_history_idx($hyperpage)\] $hyperpage
    " -state disabled
    pack $p.forward -side left
    button $p.stop -text "Stop" -command "
	surfit_terminate_loading $hyperpage
	set surfit_pageStatus($hyperpage,busy) 0
	catch {unset surfit_pageStatus($hyperpage,pending)}
	set surfit_feedback($hyperpage) {}
    " -state disabled
    pack $p.stop -side left
    set surfit_pageStatus($hyperpage,stopbg)  [[set stop $p.stop] cget -background]

    grid $p -row 1 -sticky new

    # Create the information panel
 
    set p [frame $t.info-panel]

    button $p.currPageLabel -text "Current Page:"  -state disabled -relief flat -borderwidth 0  -disabledforeground black
    entry $p.currPage -textvariable surfit_openPage($hyperpage) -relief sunken -border 2
    bind $p.currPage <Key> "surfit_enable_openlab $p"
    bind $p.currPage <Control-Key-u> "set surfit_openPage($hyperpage) {}"
    grid $p.currPageLabel -row 0 -column 0 -sticky e -pady 4 -padx 2
    grid $p.currPage -row 0 -column 1 -sticky news -pady 4

    label $p.highlightLabel -text "URL: "
    label $p.highlightURL -textvariable surfit_highlightURL($hyperpage) -relief ridge -border 2 -anchor w
    grid $p.highlightLabel -row 1 -column 0 -sticky e -ipadx 6
    grid $p.highlightURL -row 1 -column 1 -padx 2 -sticky ew
    grid columnconfigure $p 1 -weight 1

    grid $p -row 2 -sticky new

    # Create the banner panel

    text $t.banner -height 3 -relief raised
    # grid $t.banner -row 3 -sticky ew
    HMinit_state $t.banner

    # Create the bottom feedback panel

    set p [frame $t.bottom-panel]

    label $p.feedback -textvariable surfit_feedback($hyperpage)  -relief raised -border 2 -anchor nw
    pack $p.feedback -side left -anchor nw -fill both -expand yes

    grid $p -row 5 -sticky sew

    # Create the main text widget for the hyperdocument
    # If the widget path changes, be sure to update the SurfIt_hyperpage procedure
 
    # We use a canvas to implement pixel-level vertical scrolling
    # horizontal scrolling is done normally

    set p [frame $t.main]
    frame $p.b -borderwidth 2 -relief raised
    set can [canvas $p.b.c -borderwidth 0]
    if {$hyperpage != [text $p.b.text -borderwidth 0  -xscrollcommand "$p.xscroll set"]} {
	# Sanity check
	error "internal error: hyperpage pathname incorrect"
    }
    $can configure -width [winfo reqwidth $hyperpage] -height [winfo reqheight $hyperpage]
    set surfit_pageStatus($hyperpage,height) [$hyperpage cget -height]	;# For resizing properly

    # Initialise the text widgets for HTML parsing & rendering
    HMinit_hyperwin $hyperpage
    HMset_state $hyperpage -symbols \154\155\147\151\152\153\121\122\123\124\125\126\127\130\131\132\133\134\135\136

    # Initialise transparent GIF handling
    global TRANSPARENT_GIF_COLOR; set TRANSPARENT_GIF_COLOR [$hyperpage cget -background]

    if {[info exists gk_library]} {
	gk_scrollbar $p.xscroll -command "$hyperpage xview" -orient horizontal
    } else {
	scrollbar $p.xscroll -command "$hyperpage xview" -orient horizontal
    }
    button $p.mode
    scrollbar $p.yscroll

    grid $p.b -row 0 -column 0 -sticky news
    grid $p.yscroll -row 0 -column 1 -sticky ns
    grid $p.xscroll -row 1 -column 0 -sticky ew
    grid $p.mode -row 1 -column 1 -sticky news
    grid columnconfigure $p 0 -weight 1
    grid columnconfigure $p 1 -weight 0
    grid rowconfigure $p 0 -weight 1
    grid rowconfigure $p 1 -weight 0

    grid $p -row 4 -sticky news
    grid rowconfigure $t 4 -weight 1
    grid columnconfigure $t 0 -weight 1

    # Pack widgets in scroll-by-line mode initially
    surfit_line_scrolling $p

    # Explicitly setting the size locks in the size of the info-panel widgets
    bind lockGeom <Visibility> [list surfit_lock_geometry $t]
    bindtags $t "lockGeom [bindtags $t]"

    # If an initial URL was specified, load it now
    if {$open_url != {}} {
	incr surfit_url_history_idx($hyperpage)
	SurfIt_loadURL $open_url $hyperpage
    }

    if {[info exists gk_library]} {
	gk_specializeWidgetTreeTelepointer $p.b.text
    }

    return $t
}

# Given a window, return the pathname for the text widget in the hyperwindow.
# This pathname is used in many places, so it is critical to get it right!

proc SurfIt_hyperpage {t} {
    return [winfo toplevel $t].main.b.text
}

# Adjust the geometry of the hyperwindow appropriately
# Also make sure that the window fits within the screen

proc surfit_lock_geometry {win} {
    set scrhgt [winfo screenheight $win]
    set scrwid [winfo screenwidth $win]
    debug_puts [list geometry of $win is [wm geometry $win]]
    regexp {([^x]+)x([^+-]+)(\+|-)(.[^+-]*)(\+|-)(.+)} [wm geometry $win] all width height dx x dy y
    debug_puts [list got geometry $width x $height $dx $x $dy $y]

    # normalize
    if {$dx == "-"} {
	set x [expr $scrwid - $width]
    }
    if {$dy == "-"} {
	set y [expr $scrhgt - $height]
    }

    # Make sure window is visible
    if {$y + $height > $scrhgt} {
	if {$height > $scrhgt} {
	    set height $scrhgt
	    set y 0
	} else {
	    set y [expr $scrhgt - $height]
	}
    }
    if {$x + $width > $scrwid} {
	if {$width > $scrwid} {
	    set width $scrwid
	    set x 0
	} else {
	    set x [expr $scrwid - $width]
	}
    }

    # Lock in new geometry
    debug_puts [list setting geometry ${width}x${height}+$x+$y]
    wm geometry $win ${width}x${height}+$x+$y
    bindtags $win [lrange [bindtags $win] 1 end]
}

# Manage vertical scrolling

# By default the Text widget manages scrolling normally.
# However, the user can enable pixel-level scrolling

proc surfit_line_scrolling {win} {
    global surfit_pageStatus

    set win [winfo toplevel $win].main
    $win.mode configure -text L -command [list surfit_pixel_scrolling $win]
    if {[catch {$win.mode configure -image ICONtext.document}]} {
	$win.mode configure -image {}
    }
    lassign [$win.yscroll get] first last
    $win.b.c configure -yscrollcommand {}
    $win.b.text configure -yscrollcommand [list $win.yscroll set]
    $win.yscroll configure -command [list $win.b.text yview]
    #$win.yscroll set $first $last

    grid forget $win.b.text $win.b.c	;# unpack everything

    grid $win.b.text -sticky news	;# skip the canvas
    grid rowconfigure $win.b 0 -weight 1
    grid columnconfigure $win.b 0 -weight 1

    set surfit_pageStatus($win.b.text,scrollMode) line
}

proc surfit_pixel_scrolling {win} {
    global surfit_pageStatus

    set win [winfo toplevel $win].main
    set surfit_pageStatus($win.b.text,scrollMode) pixel
    $win.mode configure -text P -command [list surfit_line_scrolling $win]
    if {[catch {$win.mode configure -image ICONcompressed.document}]} {
	$win.mode configure -image {}
    }
    lassign [$win.yscroll get] first last
    $win.b.text configure -yscrollcommand [list surfit_scroll $win.b.c]
    $win.b.text yview moveto 0
    $win.b.c configure -yscrollcommand [list $win.yscroll set]
    $win.yscroll configure -command [list $win.b.c yview]
    #$win.yscroll set $first $last
    catch {$win.b.c delete main}
    $win.b.c create window 0 0 -anchor nw -window $win.b.text -tags main
    surfit_scroll $win.b.c {} {}
    $win.b.c focus main
    bind $win.b.c <Configure> [list $win.b.c itemconfigure main -width %w]

    grid forget $win.b.text $win.b.c	;# unpack everything

    grid $win.b.c -sticky news
    grid rowconfigure $win.b 0 -weight 1
    grid columnconfigure $win.b 0 -weight 1

    update idletasks
    $win.b.c itemconfigure main -width [winfo width $win.b.c]
    UTIL_resetFitText $win.b.text
    UTIL_fitTextVertical $win.b.text $surfit_pageStatus($win.b.text,height) {} 1
}

# We need to set the scrollbar appropriately

proc surfit_scroll {w first last} {
    $w configure -scrollregion [$w bbox main]
}

# Destroy given window and other storage associated with it
proc surfit_destroy_window {t} {
    global surfit_url_history surfit_url_history_idx surfit_open_vars surfit_currPage

    set hyperpage [SurfIt_hyperpage $t]
    unset surfit_url_history($hyperpage)
    unset surfit_url_history_idx($hyperpage)
    unset surfit_open_vars($hyperpage)
    unset surfit_currPage($hyperpage)
    Applet_destroy_window $hyperpage
    # HMreset_state ...
    # unset state...
    destroy $t
}

proc surfit_load_source {win} {
    global surfit_url_history surfit_url_history_idx surfit_open_vars
    global surfit_currPage

    set hyperpage [SurfIt_hyperpage $win]
    SurfIt_loadURL $surfit_currPage($hyperpage) $hyperpage surfit/htmlsource
    $win.top-ctl-panel.doc.m entryconfigure Source -label Render -command "
	SurfIt_loadURL \$surfit_currPage($hyperpage) $hyperpage
	$win.top-ctl-panel.doc.m entryconfigure Render -label Source  -command \"surfit_load_source $win\"
    "
}

proc surfit_enable_openlab {p} {
    set hyperpage [SurfIt_hyperpage $p]

    bind $p.currPage <Key> {}
    bind $p.currPage <Key-Return> "
	upvar #0 PS$hyperpage altvar
	set altvar(win) $hyperpage
	incr surfit_url_history_idx($hyperpage)
	SurfIt_loadURL \$surfit_openPage($hyperpage) $hyperpage
    "

    $p.currPageLabel configure -state normal -relief raised -borderwidth 2  -text "Open:" -width [string length [$p.currPageLabel cget -text]] -command "
	    upvar #0 PS$hyperpage altvar
	    set altvar(win) $hyperpage
	    incr surfit_url_history_idx($hyperpage)
	    SurfIt_loadURL \$surfit_openPage($hyperpage) $hyperpage
	"
    grid configure $p.currPageLabel $p.currPage -pady 0 -padx 0
    grid propagate $p off
}

proc surfit_setOpenPage {win url} {
    global surfit_currPage surfit_openPage

    set hyperpage [SurfIt_hyperpage $win]
    set p [winfo toplevel $win].info-panel
    set surfit_currPage($hyperpage) [set surfit_openPage($hyperpage) $url]

    $p.currPageLabel configure -text "Current Page:" -state disabled  -relief flat -borderwidth 0
    bind $p.currPage <Key-Return> {}
    bind $p.currPage <Key> "surfit_enable_openlab $p"
}

proc surfit_save_doc {win url} {
    set state surfit_save$win
    upvar #0 $state var
    toplevel $win.saveas
    wm title $win.saveas "Save Document"
    URI_parse $url scheme path query fragment
    set initial [file tail $path]
    array set var [list win $win.saveas.fsbox top $win.saveas url $url ignoreMissing 1 exit surfit_saveas_exit open surfit_saveas_open initialSel $initial flist [list [list Everything *]]  ]
    Fsbox $state
    pack $var(win) -side top -fill both -expand yes
    grab $win.saveas
}

proc surfit_saveas_exit {state} {
    upvar #0 $state var
    grab release $var(top)
    destroy $var(top)
    unset var
}

proc surfit_saveas_open {state} {
    upvar #0 $state var
    if {[file exists $var(selFile)]} {
	if {[tk_dialog $var(top).overwrite "Overwrite $var(selFile)?"  "The file $var(selFile) already exists\nAre you sure you want to overwrite it?"  {} 1 "OK" "Cancel"]} {
	    return
	}
    }

    if {[catch {open $var(selFile) w} savefd]} {
	tk_dialog $var(top).cannotopen "File Open Error"  "Cannot open $var(selFile) for writing"  {} 0 "OK"
	return;
    } else {
	fconfigure $savefd -translation binary
	if {![catch {open [eval file join [Cache_get_file $var(url)]] r} cachefd]} {
	    fconfigure $cachefd -translation binary
	    fcopy $cachefd $savefd
	    close $cachefd
	} else {
	    close $savefd
	    error $cachefd
	}
	close $savefd
    }

    surfit_saveas_exit $state
}

### Other miscellaneous procedures

proc tkAboutDialog {} {
    toplevel .about
    wm title .about "About Plume"
    message .about.text -justify center -aspect 250  -text "The Plume World Wide Web browser

By Steve Ball,
Department of Computer Science,
Australian National University.

http://tcltk.anu.edu.au/

Copyright (c) 1996, 1997 Australian National University

Copyright (c) 1995,1996 ANU and CSIRO on behalf of the participants in 
the  CRC for Advanced Computational Systems ('ACSys').

ACSys makes this software and all associated data and documentation
('Software') available free of charge for non-commercial purposes only. You
may make copies of the Software but you must include all of this notice on
any copy.

The Software was developed for research purposes and ACSys does not warrant
that it is error free or fit for any purpose.  ACSys disclaims any
liability for all claims, expenses, losses, damages and costs any user may
incur as a result of using, copying or modifying the Software.
"
    button .about.ok -text OK -command {destroy .about}
    grid .about.ok -row 1
    grid .about.text -row 0 -sticky news
    grid rowconfigure .about 0 -weight 1
    grid columnconfigure .about 0 -weight 1
    grab .about.ok
}

proc tkOpenDocument args {
}


###
### Preference editing
###

proc SurfIt_Preferences {} {
    global surfit surfit_prefs

    if {[winfo exists .prefs]} return;	# already been here

    set p [toplevel .prefs]
    wm title .prefs "Plume Preferences"
    wm minsize .prefs 500 225	;# Should calc this, taking screen size into account

    button $p.ok -text OK -command "surfit_prefs_apply; surfit_prefs_save; destroy $p"
    button $p.apply -text Apply -command "surfit_prefs_apply; destroy $p"
    button $p.cancel -text Cancel -command "destroy $p"

    set c [canvas $p.c  -width [min 500 [winfo screenwidth .]] -height 40 -borderwidth 0 -highlightthickness 0]

    grid $p.c - - -row 0 -sticky news
    grid $p.ok -row 2 -column 0 -sticky w
    grid $p.apply -row 2 -column 1
    grid $p.cancel -row 2 -column 2 -sticky e
    grid rowconfigure $p 1 -weight 1
    grid columnconfigure $p 1 -weight 1

    ### Browser preferences
    set id [$c create text 25 27 -text "Browser" -anchor sw -tags {browser browser_text}]
    eval surfit_prefs_folderTop $c [$c bbox $id] browser
    $c bind browser <ButtonRelease-1> [list surfit_prefs_activate $p browser]
    # default home page, prefs script, cache directory
    set f [frame $p.browser]
    label $f.homeLabel -text "Default Home page:"
    entry $f.home -width 40
    catch {$f.home insert 0 $surfit(CFGwwwhome)}
    if {[info exists surfit(CFGcachedir)]} {
	set cachedir $surfit(CFGcachedir)
    } else {
	set cachedir "$surfit(userdir) Cache"
    }
    label $f.cachedirLabel -text "Cache directory:"
    entry $f.cachedir -width 40
    $f.cachedir insert 0 [eval file join $cachedir]
    label $f.usernameLabel -text "User name:"
    entry $f.username -width 40
    $f.username insert 0 $surfit(CFGusername)
    label $f.addnLabel -text "Additional Preferences (Tcl commands):"
    scrollbar $f.addnCmdsX -orient horizontal -command "$f.addnCmds xview"
    scrollbar $f.addnCmdsY -orient vertical -command "$f.addnCmds yview"
    text $f.addnCmds -width 40 -height 10 -wrap none  -xscrollcommand "$f.addnCmdsX set" -yscrollcommand "$f.addnCmdsY set"
    catch {$f.addnCmds insert 1.0 $surfit(CFGaddnCmds)}
    grid $f.homeLabel -row 0 -column 0 -sticky e
    grid $f.home -row 0 -column 1 -sticky ew
    grid $f.cachedirLabel -row 1 -column 0 -sticky e
    grid $f.cachedir -row 1 -column 1 -sticky ew
    grid $f.usernameLabel -row 2 -column 0 -sticky e
    grid $f.username -row 2 -column 1 -sticky ew
    grid $f.addnLabel - -row 3 -column 0 -sticky w
    grid $f.addnCmds - -row 4 -column 0 -sticky news
    grid $f.addnCmdsY -row 4 -column 2 -sticky ns
    grid $f.addnCmdsX - -row 5 -column 0 -sticky ew
    grid rowconfigure $f 4 -weight 1
    grid columnconfigure $f 1 -weight 1

    ### Network preferences
    set id [$c create text 100 27 -text "Network" -anchor sw -tags {network network_text}]
    eval surfit_prefs_folderTop $c [$c bbox $id] network
    $c bind network <ButtonRelease-1> [list surfit_prefs_activate $p network]
    # HTTP proxy server, FTP proxy server
    set f [frame $p.network]

    label $f.httpProxyLabel -text "HTTP proxy server:"
    entry $f.httpProxy -width 30
    catch {$f.httpProxy insert 0 $surfit(CFGproxy)}
    label $f.httpProxyPortLabel -text "port:"
    entry $f.httpProxyPort -width 10
    catch {$f.httpProxyPort insert 0 $surfit(CFGproxy-port)}
    grid $f.httpProxyLabel -sticky ne -row 0 -column 0
    grid $f.httpProxy -sticky new -row 0 -column 1
    grid $f.httpProxyPortLabel -sticky ne -row 0 -column 2
    grid $f.httpProxyPort -sticky new -row 0 -column 3

    label $f.ftpProxyLabel -text "FTP proxy server:"
    entry $f.ftpProxy -width 30
    catch {$f.ftpProxy insert 0 $surfit(CFGftpproxy)}
    label $f.ftpProxyPortLabel -text "port:"
    entry $f.ftpProxyPort -width 10
    catch {$f.ftpProxyPort insert 0 $surfit(CFGftpproxy-port)}
    grid $f.ftpProxyLabel -sticky ne -row 1 -column 0
    grid $f.ftpProxy -sticky new -row 1 -column 1
    grid $f.ftpProxyPortLabel -sticky ne -row 1 -column 2
    grid $f.ftpProxyPort -sticky new -row 1 -column 3

    set g [frame $f.noProxy]
    message $g.label -text "Don't use proxies for these hosts\n(* matches any name)" -aspect 125
    text $g.noProxy -width 40 -height 5 -wrap none -xscrollcommand "$g.xscroll set" -yscrollcommand "$g.yscroll set"
    if {[info exists surfit(CFGnoproxy)]} {
	regsub " |	|\t" $surfit(CFGnoproxy) "\n" noProxies
	$g.noProxy insert end $noProxies
    }
    scrollbar $g.xscroll -command "$g.noProxy xview" -orient horizontal
    scrollbar $g.yscroll -command "$g.noProxy yview" -orient vertical
    grid $g.label -sticky ne -row 0 -column 0 -rowspan 2
    grid $g.noProxy -sticky news -row 0 -column 1
    grid $g.xscroll -sticky ew -row 1 -column 1
    grid $g.yscroll -sticky ns -row 0 -column 2
    grid rowconfigure $g 0 -weight 1
    grid columnconfigure $g 1 -weight 1
    grid $g - - - -row 2 -column 0 -sticky news

    grid rowconfigure $f 2 -weight 1
    grid columnconfigure $f 1 -weight 1
    grid columnconfigure $f 3 -weight 1

    ### Hypertool preferences
    set id [$c create text 175 27 -text "Hypertools" -anchor sw -tags {hypertools hypertools_text}]
    eval surfit_prefs_folderTop $c [$c bbox $id] hypertools
    $c bind hypertools <ButtonRelease-1> [list surfit_prefs_activate $p hypertools]
    # MAILTO, NNTPHANDLER, CMHANDLER hypertools
    set f [frame $p.hypertools]
    label $f.mailtoLabel -text "mailto: hypertool program:"
    entry $f.mailto -width 20
    catch {$f.mailto insert 0 $surfit(CFGmailto_handler)}
    label $f.newsLabel -text "news: or nntp: hypertool program:"
    entry $f.news -width 20
    catch {$f.news insert 0 $surfit(CFGnews_handler)}
    label $f.cmLabel -text "continuous media hypertool program:"
    entry $f.cm -width 20
    catch {$f.cm insert 0 $surfit(CFGcm_handler)}
    grid $f.mailtoLabel -row 0 -column 0 -sticky ne
    grid $f.mailto -row 0 -column 1 -sticky new
    grid $f.newsLabel -row 1 -column 0 -sticky ne
    grid $f.news -row 1 -column 1 -sticky new
    grid $f.cmLabel -row 2 -column 0 -sticky ne
    grid $f.cm -row 2 -column 1 -sticky new
    grid rowconfigure $f 2 -weight 1
    grid columnconfigure $f 1 -weight 1

    # Default initial page
    set surfit_prefs(prev_active) browser
    surfit_prefs_activate $p browser
}

proc surfit_prefs_apply {} {
    global surfit wwwhome
    array set tmp [array get surfit CFG*]
    array set surfit [list  CFGwwwhome [.prefs.browser.home get] CFGcachedir [file split [.prefs.browser.cachedir get]] CFGusername [.prefs.browser.username get] CFGaddnCmds [string trim [.prefs.browser.addnCmds get 1.0 end]]  CFGproxy [.prefs.network.httpProxy get] CFGproxy-port [.prefs.network.httpProxyPort get] CFGftpproxy [.prefs.network.ftpProxy get] CFGftpproxy-port [.prefs.network.ftpProxyPort get]  CFGmailto_handler [.prefs.hypertools.mailto get] CFGnews_handler [.prefs.hypertools.news get] CFGcm_handler [.prefs.hypertools.cm get]  ]
    set noproxies [.prefs.network.noProxy.noProxy get 1.0 end]
    regsub -all " |	|\n" $noproxies { } surfit(CFGnoproxy)
    # Re-initialise as appropriate
    if {![info exists tmp(CFGcachedir)] && $surfit(CFGcachedir) != {}} {
	Cache_init
    } elseif {$tmp(CFGcachedir) != $surfit(CFGcachedir)} {
	Cache_init
    }
    if {[info exists tmp(CFGwwwhome)] && $surfit(CFGwwwhome) == {}} {
	set wwwhome {}
	.open-panel.home configure -state disabled
    } else {
	set wwwhome $surfit(CFGwwwhome)
	.open-panel.home configure -state normal
    }

    set cmds_changed 0
    if {![info exists tmp(CFGaddnCmds)] && $surfit(CFGaddnCmds) != {}} {
	set cmds_changed 1
    } elseif {[info exists tmp(CFGaddnCmds)] && $tmp(CFGaddnCmds) != $surfit(CFGaddnCmds)} {
	set cmds_changed 1
    }
    if {$cmds_changed} {
	tk_dialog .msg {Additional Commands} {Additional Tcl commands will not take effect until you restart Plume}  {} 0 OK
    }

    if {![info exists tmp(CFGmailto_handler)] && $surfit(CFGmailto_handler) != {}} {
	mailto_init
    } elseif {[info exists tmp(CFGmailto_handler)] &&  $tmp(CFGmailto_handler) != $surfit(CFGmailto_handler)} {
	mailto_init
	if {[string match *exmh* $tmp(CFGmailto_handler)]} {
	    PRremove_scheme mailto Exmh_mailto
	    Hypertool_remove SMTP
	}
    }
    if {![info exists tmp(CFGcm_handler)] && $surfit(CFGcm_handler) != {}} CM_init  elseif {[info exists tmp(CFGcm_handler)] && $tmp(CFGcm_handler) != $surfit(CFGcm_handler)} CM_init;
    if {![info exists tmp(CFGnews_handler)] && $surfit(CFGnews_handler) != {}} NEWS_init  elseif {[info exists tmp(CFGnews_handler)] && $tmp(CFGnews_handler) != $surfit(CFGnews_handler)} NEWS_init;
}

# Write the current preferences out to the preferences script

proc surfit_prefs_save {} {
    global surfit
    if {[catch {open [eval file join $surfit(userdir) Prefs.tcl] w} pfd]} {
    }
    puts $pfd {#
# Plume preferences script
#
# This script is written automatically - DO NOT EDIT MANUALLY
# Use the Plume Preferences dialog instead
#

}
    puts -nonewline $pfd {array set surfit_preferences }
    puts $pfd [list [array get surfit CFG*]]
    puts $pfd {
	foreach {pref value} [array get surfit_preferences] {
	    if {$value != {}} {set surfit($pref) $value}
	}
    }
    puts $pfd {
	if {$surfit(CFGaddnCmds) != {} && [catch {eval $surfit(CFGaddnCmds)} err]} {
	    puts stderr "An error occurred while processing additional user commands:\n$err"
	}
    }
    close $pfd
}

proc surfit_prefs_activate {win tag} {
    global surfit_prefs
    grid forget $win.$surfit_prefs(prev_active)
    $win.c raise $tag
    grid $win.$tag -row 1 -column 0 -columnspan 3 -sticky news
    set surfit_prefs(prev_active) $tag
}

proc surfit_prefs_folderTop {win l t r b tags} {
    set t [expr $t + 3]
    set b [expr $b + 3]
    $win create polygon  -10 $b -10 $b [expr $l - 10] $b [expr $l - 10] $b [expr $l - 6.5] $b [expr $l - 5] [expr $b - 5]  [expr $l - 5] [expr $b - 5] [expr $l - 5] $t  [expr $l - 5] $t [expr $l - 3.5] [expr $t - 3.5] $l [expr $t - 5]  $l [expr $t - 5] $r [expr $t - 5]  $r [expr $t - 5] [expr $r + 3.5] [expr $t - 3.5] [expr $r + 5] $t  [expr $r + 5] $t [expr $r + 5] [expr $b - 5]  [expr $r + 5] [expr $b - 5] [expr $r + 6.5] [expr $b - 1] [expr $r + 10] $b  [expr $r + 10] $b 2000 $b  2000 $b 2000 100 2000 100 -10 100 -10 100 -10 $b  -smooth true -outline black -fill [$win cget -background]  -tags "$tags ${tags}_folderTop"
    $win lower ${tags}_folderTop ${tags}_text
}

#
# Presentation mode
#
# For presentations, the control panels are removed, the hyperwindow
# maximised to cover the display and new bindings are set.
# Bindings:
#	Control-Return, Control-Enter restores browsing mode
#	<1>, Right-Arrow goes to next page (<LINK>)
#	Left-Arrow goes to previous page (<LINK>)
#	Up-Arrow goes back
#	Down-Arrow goes forward
#	Home goes to Home page (<LINK>)
#

bind presentation <Control-Return> {surfit_disp_mode browse %W}
bind presentation <Control-KP_Enter> {surfit_disp_mode browse %W}
bind presentation <1> {surfit_go next %W}
bind presentation <Left> {surfit_go prev %W}
bind presentation <Right> {surfit_go next %W}
bind presentation <Up> {surfit_go back %W}
bind presentation <Down> {surfit_go forward %W}
bind presentation <Home> {surfit_go home %W}

proc surfit_disp_mode {mode t} {
    global surfit_pageStatus

    set t [winfo toplevel $t]
    switch $mode {
	presentation {
	    grid forget $t.top-ctl-panel $t.second-ctl-panel $t.info-panel $t.bottom-panel
	    set surfit_pageStatus([SurfIt_hyperpage $t],geom) [wm geometry $t]
	    # Maximise window
	    bindtags $t "presentation [bindtags $t]"
	}
	browse {
	    # Re-grid control panels
	    grid $t.top-ctl-panel -row 0 -sticky new
	    grid $t.second-ctl-panel -row 1 -sticky new
	    grid $t.info-panel -row 2 -sticky new
	    grid $t.bottom-panel -row 5 -sticky sew

	    # Restore window geometry
	    wm geometry $t $surfit_pageStatus([SurfIt_hyperpage $t],geom)
	    # Remove binding
	    regsub presentation [bindtags $t] {} tags
	    bindtags $t $tags
	}
    }
}
} \
]
array set bitmaps [list \
archive {#define archive_width 20
#define archive_height 23
static char archive_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0xe0, 0x00, 0x30, 0x3c, 0x00,
   0xe0, 0x06, 0x00, 0x80, 0x07, 0x00, 0xc0, 0x0c, 0x00, 0xfc, 0xff, 0x0f,
   0x42, 0x00, 0x08, 0xfe, 0xff, 0x07, 0x40, 0x00, 0x00, 0xfe, 0xff, 0x07,
   0x41, 0x00, 0x04, 0xff, 0xff, 0x03, 0xc0, 0x00, 0x00, 0xf8, 0xff, 0x07,
   0x84, 0x01, 0x04, 0xfc, 0xff, 0x03, 0x00, 0x07, 0x00};
} \
audio {#define audio_width 20
#define audio_height 23
static char audio_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x02, 0x01,
   0x00, 0x23, 0x02, 0x80, 0x42, 0x02, 0x40, 0x92, 0x04, 0x3c, 0xa2, 0x04,
   0x04, 0xaa, 0x04, 0x04, 0xaa, 0x04, 0x04, 0xaa, 0x04, 0x04, 0xaa, 0x04,
   0x3c, 0xaa, 0x04, 0x7c, 0xa2, 0x04, 0xc0, 0x92, 0x04, 0x80, 0x43, 0x02,
   0x00, 0x23, 0x02, 0x00, 0x02, 0x01, 0x00, 0x80, 0x00};
} \
binary_document {#define binary_document_width 20
#define binary_document_height 23
static char binary_document_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0xfc, 0x1f, 0x00, 0x04, 0x30, 0x00, 0x04, 0x50, 0x00, 0x04, 0x90, 0x00,
   0x04, 0x10, 0x01, 0x04, 0xf0, 0x03, 0x74, 0x0e, 0x02, 0x64, 0x1b, 0x02,
   0x64, 0x1b, 0x02, 0x64, 0x1b, 0x02, 0x64, 0x0e, 0x02, 0x04, 0x00, 0x02,
   0xe4, 0x1c, 0x02, 0xb4, 0x19, 0x02, 0xb4, 0x19, 0x02, 0xb4, 0x19, 0x02,
   0xe4, 0x18, 0x02, 0x04, 0x00, 0x02, 0xfc, 0xff, 0x03};
} \
binhex_document {#define binhex_document_width 20
#define binhex_document_height 23
static char binhex_document_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0xfc, 0x1f, 0x00, 0x04, 0x30, 0x00, 0x04, 0x50, 0x00, 0x04, 0x90, 0x00,
   0x04, 0x10, 0x01, 0x04, 0xf0, 0x03, 0x04, 0x00, 0x02, 0x04, 0x00, 0x02,
   0x00, 0x00, 0x00, 0x6c, 0x6f, 0x03, 0x6c, 0x63, 0x03, 0x6c, 0xc3, 0x01,
   0x7c, 0x87, 0x00, 0x6c, 0xc3, 0x01, 0x6c, 0x63, 0x03, 0x6c, 0x6f, 0x03,
   0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0xfc, 0xff, 0x03};
} \
calculator {#define calculator_width 20
#define calculator_height 23
static char calculator_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfc, 0xff, 0x01, 0x04, 0x00, 0x01,
   0xf4, 0x7f, 0x01, 0x14, 0x40, 0x01, 0x14, 0x40, 0x01, 0xf4, 0x7f, 0x01,
   0x04, 0x00, 0x01, 0x04, 0x00, 0x01, 0x74, 0x77, 0x01, 0x54, 0x55, 0x01,
   0x74, 0x77, 0x01, 0x04, 0x00, 0x01, 0x74, 0x77, 0x01, 0x54, 0x55, 0x01,
   0x74, 0x77, 0x01, 0x04, 0x00, 0x01, 0x74, 0x77, 0x01, 0x54, 0x55, 0x01,
   0x74, 0x77, 0x01, 0x04, 0x00, 0x01, 0xfc, 0xff, 0x01};
} \
caution {#define caution_width 20
#define caution_height 23
static char caution_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x00,
   0x00, 0x0f, 0x00, 0x00, 0x0f, 0x00, 0x80, 0x19, 0x00, 0x80, 0x19, 0x00,
   0xc0, 0x30, 0x00, 0xc0, 0x30, 0x00, 0x60, 0x66, 0x00, 0x60, 0x66, 0x00,
   0x30, 0xc6, 0x00, 0x30, 0xc6, 0x00, 0x18, 0x86, 0x01, 0x18, 0x86, 0x01,
   0x0c, 0x06, 0x03, 0x0c, 0x00, 0x03, 0x06, 0x06, 0x06, 0x06, 0x06, 0x06,
   0x03, 0x00, 0x0c, 0xff, 0xff, 0x0f, 0xfe, 0xff, 0x07};
} \
cd_i {#define cd-i_width 20
#define cd-i_height 23
static char cd-i_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x0f, 0x00, 0x60, 0x30, 0x00,
   0x18, 0xc0, 0x00, 0x04, 0x00, 0x01, 0x04, 0xfc, 0x0f, 0xb2, 0xfd, 0x0f,
   0x02, 0x04, 0x08, 0x01, 0x07, 0x08, 0x81, 0xfc, 0x0f, 0x95, 0x24, 0x09,
   0x81, 0x24, 0x09, 0x01, 0xff, 0x0f, 0x02, 0x24, 0x09, 0x02, 0x24, 0x09,
   0xb4, 0xfc, 0x0f, 0x04, 0x24, 0x09, 0x18, 0x24, 0x09, 0x60, 0xfc, 0x0f,
   0x80, 0x27, 0x09, 0x00, 0x24, 0x09, 0x00, 0xfc, 0x0f};
} \
cd_rom {#define cd-rom_width 20
#define cd-rom_height 23
static char cd-rom_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x80, 0x0f, 0x00, 0x60, 0x30, 0x00, 0x18, 0xc0, 0x00, 0xc4, 0x07, 0x01,
   0x04, 0x00, 0x01, 0xe2, 0x1f, 0x02, 0x02, 0x00, 0x02, 0x01, 0x07, 0x04,
   0x81, 0x0f, 0x04, 0x81, 0x4f, 0x05, 0x81, 0x0f, 0x04, 0x01, 0x07, 0x04,
   0x02, 0x00, 0x02, 0xe2, 0x1f, 0x02, 0x04, 0x00, 0x01, 0xc4, 0x07, 0x01,
   0x18, 0xc0, 0x00, 0x60, 0x30, 0x00, 0x80, 0x0f, 0x00};
} \
clock {#define clock_width 20
#define clock_height 23
static char clock_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0xc0, 0x1f, 0x00, 0x30, 0x60, 0x00, 0x08, 0x82, 0x00, 0x04, 0x02, 0x01,
   0x02, 0x02, 0x02, 0x02, 0x02, 0x02, 0x01, 0x02, 0x04, 0x01, 0x02, 0x04,
   0x01, 0x02, 0x04, 0xe1, 0x07, 0x04, 0x01, 0x02, 0x04, 0x01, 0x00, 0x04,
   0x01, 0x00, 0x04, 0x02, 0x00, 0x02, 0x02, 0x00, 0x02, 0x04, 0x00, 0x01,
   0x08, 0x80, 0x00, 0x30, 0x60, 0x00, 0xc0, 0x1f, 0x00};
} \
compressed_document {#define compressed_document_width 20
#define compressed_document_height 23
static char compressed_document_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0xe0, 0x03, 0x00, 0x20, 0x06, 0x00, 0x20, 0x0a, 0x00, 0x20, 0x12, 0x00,
   0x20, 0x22, 0x00, 0x20, 0x7e, 0x00, 0x21, 0x40, 0x08, 0x23, 0x40, 0x0c,
   0x27, 0x40, 0x0e, 0x2f, 0x40, 0x0f, 0x27, 0x40, 0x0e, 0x23, 0x40, 0x0c,
   0x21, 0x40, 0x08, 0x20, 0x40, 0x00, 0x20, 0x40, 0x00, 0x20, 0x40, 0x00,
   0x20, 0x40, 0x00, 0x20, 0x40, 0x00, 0xe0, 0x7f, 0x00};
} \
disk_drive {#define disk_drive_width 20
#define disk_drive_height 23
static char disk_drive_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0xff, 0xff, 0x0f, 0x01, 0x00, 0x08, 0x0d, 0x00, 0x08, 0x0d, 0x00, 0x08,
   0x81, 0x1f, 0x08, 0xfd, 0xff, 0x0b, 0x81, 0x1f, 0x08, 0x01, 0x00, 0x08,
   0x01, 0x00, 0x08, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
diskette {#define diskette_width 20
#define diskette_height 23
static char diskette_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0xfe, 0xff, 0x01, 0x22, 0x40, 0x02, 0x22, 0x40, 0x04,
   0x22, 0x40, 0x04, 0x22, 0x40, 0x04, 0xe2, 0x7f, 0x04, 0x02, 0x00, 0x04,
   0x02, 0x06, 0x04, 0x02, 0x09, 0x04, 0x02, 0x09, 0x04, 0x02, 0x06, 0x04,
   0x02, 0x00, 0x04, 0x02, 0x00, 0x04, 0x02, 0x00, 0x04, 0x02, 0x00, 0x04,
   0x02, 0x00, 0x04, 0x02, 0x00, 0x04, 0xfe, 0xff, 0x07};
} \
display {#define display_width 20
#define display_height 23
static char display_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0xfe, 0xff, 0x07, 0xfe, 0xff, 0x07, 0x1e, 0x80, 0x07, 0x0e, 0x00, 0x07,
   0x06, 0x00, 0x06, 0x06, 0x00, 0x06, 0x06, 0x00, 0x06, 0x06, 0x00, 0x06,
   0x06, 0x00, 0x06, 0x06, 0x00, 0x06, 0x06, 0x00, 0x06, 0x06, 0x00, 0x06,
   0x0e, 0x00, 0x07, 0x1e, 0x80, 0x07, 0xfe, 0xff, 0x07, 0xfc, 0xff, 0x03,
   0x02, 0x00, 0x04, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
document {#define document_width 20
#define document_height 23
static char document_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0xfc, 0x1f, 0x00, 0x04, 0x30, 0x00, 0x04, 0x50, 0x00, 0x04, 0x90, 0x00,
   0x04, 0x10, 0x01, 0x04, 0xf0, 0x03, 0x04, 0x00, 0x02, 0x04, 0x00, 0x02,
   0x04, 0x00, 0x02, 0x04, 0x00, 0x02, 0x04, 0x00, 0x02, 0x04, 0x00, 0x02,
   0x04, 0x00, 0x02, 0x04, 0x00, 0x02, 0x04, 0x00, 0x02, 0x04, 0x00, 0x02,
   0x04, 0x00, 0x02, 0x04, 0x00, 0x02, 0xfc, 0xff, 0x03};
} \
fax {#define fax_width 20
#define fax_height 23
static char fax_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0xfc, 0x1f, 0x00, 0x04, 0x30, 0x00, 0x04, 0x50, 0x00, 0x04, 0x90, 0x00,
   0x04, 0x10, 0x01, 0x04, 0xf0, 0x03, 0x04, 0x00, 0x02, 0x04, 0x00, 0x02,
   0xff, 0xff, 0x0f, 0x01, 0x00, 0x08, 0x01, 0x00, 0x08, 0x01, 0x00, 0x08,
   0xe1, 0x7f, 0x08, 0xf9, 0xff, 0x09, 0x7d, 0xe0, 0x0b, 0x3d, 0xc0, 0x0b,
   0x39, 0xc0, 0x09, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
filing_cabinet {#define filing_cabinet_width 20
#define filing_cabinet_height 23
static char filing_cabinet_bits[] = {
   0x00, 0x00, 0x00, 0xf8, 0xff, 0x03, 0xfe, 0xff, 0x03, 0x02, 0x00, 0x03,
   0xfa, 0x7f, 0x03, 0x0a, 0x40, 0x03, 0x0a, 0x40, 0x03, 0x8a, 0x44, 0x03,
   0x8a, 0x47, 0x03, 0x0a, 0x40, 0x03, 0x0a, 0x40, 0x03, 0xfa, 0x7f, 0x03,
   0x02, 0x00, 0x03, 0xfa, 0x7f, 0x03, 0x0a, 0x40, 0x03, 0x0a, 0x40, 0x03,
   0x8a, 0x44, 0x03, 0x8a, 0x47, 0x03, 0x0a, 0x40, 0x03, 0x0a, 0x40, 0x03,
   0xfa, 0x7f, 0x03, 0x02, 0x00, 0x03, 0xfe, 0xff, 0x01};
} \
film {#define film_width 20
#define film_height 23
static char film_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0xff, 0xff, 0x07, 0xf9, 0xff, 0x04, 0x19, 0xc0, 0x04, 0x1f, 0xc0, 0x07,
   0x19, 0xc0, 0x04, 0x19, 0xc0, 0x04, 0x1f, 0xc0, 0x07, 0x19, 0xc0, 0x04,
   0xf9, 0xff, 0x04, 0xff, 0xff, 0x07, 0xf9, 0xff, 0x04, 0x19, 0xc0, 0x04,
   0x1f, 0xc0, 0x07, 0x19, 0xc0, 0x04, 0x19, 0xc0, 0x04, 0x1f, 0xc0, 0x07,
   0x19, 0xc0, 0x04, 0xf9, 0xff, 0x04, 0xff, 0xff, 0x07};
} \
fixed_disk {#define fixed_disk_width 20
#define fixed_disk_height 23
static char fixed_disk_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x0f, 0x01, 0x00, 0x08,
   0x0d, 0x00, 0x08, 0xcd, 0xff, 0x0b, 0x01, 0x00, 0x08, 0x01, 0x00, 0x08,
   0xc1, 0xff, 0x0b, 0x01, 0x00, 0x08, 0x01, 0x00, 0x08, 0xc1, 0xff, 0x0b,
   0x01, 0x00, 0x08, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
folder {#define folder_width 20
#define folder_height 23
static char folder_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfc, 0x03, 0x00,
   0x02, 0x04, 0x00, 0x01, 0xf8, 0x0f, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f,
   0x01, 0x00, 0x08, 0x01, 0x00, 0x08, 0x01, 0x00, 0x08, 0x01, 0x00, 0x08,
   0x01, 0x00, 0x08, 0x01, 0x00, 0x08, 0x01, 0x00, 0x08, 0x01, 0x00, 0x08,
   0x01, 0x00, 0x08, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
form {#define form_width 20
#define form_height 23
static char form_bits[] = {
   0xff, 0xff, 0x00, 0x09, 0x81, 0x00, 0x09, 0x81, 0x00, 0x09, 0x81, 0x00,
   0xf9, 0x81, 0x00, 0x01, 0x88, 0x08, 0x01, 0x8c, 0x04, 0xfd, 0x87, 0x0a,
   0x1d, 0x83, 0x05, 0xb9, 0x81, 0x0a, 0xe9, 0x41, 0x0d, 0x69, 0xa1, 0x0e,
   0xf9, 0x51, 0x07, 0x01, 0x90, 0x03, 0x01, 0x08, 0x01, 0xf9, 0xc9, 0x00,
   0x09, 0xbd, 0x00, 0x09, 0x8d, 0x00, 0x09, 0x81, 0x00, 0x09, 0x81, 0x00,
   0xf9, 0x81, 0x00, 0x01, 0x80, 0x00, 0xff, 0xff, 0x00};
} \
ftp {#define ftp_width 20
#define ftp_height 23
static char ftp_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x0f,
   0x01, 0x00, 0x08, 0xfd, 0xef, 0x09, 0x0d, 0x63, 0x0b, 0x0d, 0x63, 0x0b,
   0x3d, 0xe3, 0x09, 0x0d, 0x63, 0x08, 0x0d, 0x63, 0x08, 0x0d, 0x63, 0x08,
   0x01, 0x00, 0x08, 0xff, 0xff, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x10, 0x80, 0x00, 0x18, 0x80, 0x01, 0xfc, 0xff, 0x03, 0xfe, 0xff, 0x07,
   0xfc, 0xff, 0x03, 0x18, 0x80, 0x01, 0x10, 0x80, 0x00};
} \
glossary {#define glossary_width 20
#define glossary_height 23
static char glossary_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x9c, 0xff, 0x01, 0x00, 0x00, 0x00,
   0x80, 0x7f, 0x00, 0x00, 0x00, 0x00, 0x80, 0xff, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x8c, 0xff, 0x00, 0x00, 0x00, 0x00, 0x80, 0x1f, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x9c, 0xff, 0x01, 0x00, 0x00, 0x00,
   0x80, 0xff, 0x00, 0x00, 0x00, 0x00, 0x80, 0x3f, 0x00};
} \
gopher {#define gopher_width 20
#define gopher_height 23
static char gopher_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x0f, 0x00, 0x80, 0x14, 0x00,
   0x40, 0x21, 0x00, 0xce, 0x3f, 0x07, 0x69, 0x4f, 0x09, 0x19, 0x80, 0x08,
   0xd1, 0xb2, 0x08, 0x12, 0x86, 0x04, 0x24, 0xc0, 0x03, 0x38, 0x25, 0x00,
   0x40, 0x55, 0x00, 0x90, 0x88, 0x00, 0x0c, 0x17, 0x03, 0x00, 0x19, 0x00,
   0x80, 0x20, 0x00, 0x40, 0x38, 0x00, 0x40, 0x60, 0x00};
} \
home {#define home_width 20
#define home_height 23
static char home_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x0f,
   0x01, 0x00, 0x08, 0xfd, 0x07, 0x08, 0xfd, 0x03, 0x08, 0xfd, 0x01, 0x08,
   0xfd, 0x01, 0x08, 0xfd, 0x03, 0x08, 0xfd, 0x07, 0x08, 0xfd, 0x0f, 0x08,
   0xcd, 0x0f, 0x08, 0x85, 0x07, 0x08, 0x01, 0x03, 0x08, 0x01, 0x00, 0x08,
   0x01, 0x00, 0x08, 0x01, 0x00, 0x08, 0x01, 0x00, 0x08, 0x01, 0x00, 0x08,
   0x01, 0x00, 0x08, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
image {#define image_width 20
#define image_height 23
static char image_bits[] = {
   0xe0, 0xff, 0xff, 0x20, 0xe0, 0xff, 0xe0, 0xff, 0xff, 0x30, 0xff, 0xff,
   0xe8, 0xf8, 0xff, 0xdf, 0xf7, 0xff, 0xbb, 0xff, 0xff, 0x7b, 0xff, 0xff,
   0xfb, 0xfe, 0xff, 0xfb, 0xfd, 0xff, 0xff, 0xff, 0xff, 0xff, 0x7f, 0xff,
   0xcf, 0x1f, 0xfc, 0x03, 0x0e, 0xf8, 0x20, 0x70, 0xf0, 0x18, 0x80, 0xf1,
   0x07, 0x00, 0xf0, 0x00, 0x1e, 0xf0, 0xf8, 0x01, 0xf0, 0x00, 0x00, 0xf0,
   0xc0, 0x7f, 0xf3, 0x00, 0x80, 0xf0, 0x40, 0x00, 0xf0};
} \
index {#define index_width 20
#define index_height 23
static char index_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x20, 0x00, 0x00, 0x70, 0x00, 0x00, 0x50, 0x00, 0x00, 0xd8, 0x00, 0x00,
   0xf8, 0x00, 0x00, 0x88, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x00, 0x60, 0x00, 0x00, 0x30, 0x00,
   0x00, 0x18, 0x00, 0x00, 0x0c, 0x00, 0x00, 0xfc, 0x00};
} \
keyboard {#define keyboard_width 20
#define keyboard_height 23
static char keyboard_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x0f, 0x01, 0x00, 0x08,
   0x01, 0x00, 0x08, 0x6d, 0xdb, 0x0b, 0x6d, 0xdb, 0x0b, 0x01, 0x00, 0x08,
   0xdd, 0xb6, 0x0b, 0xdd, 0xb6, 0x0b, 0x01, 0x00, 0x08, 0xe1, 0x7f, 0x08,
   0xe1, 0x7f, 0x08, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
mail {#define mail_width 20
#define mail_height 23
static char mail_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfe, 0xff, 0x07, 0x06, 0x00, 0x06,
   0x0a, 0x00, 0x05, 0x12, 0x80, 0x04, 0x62, 0x60, 0x04, 0x82, 0x10, 0x04,
   0x02, 0x09, 0x04, 0x02, 0x06, 0x04, 0x02, 0x00, 0x04, 0x02, 0x00, 0x04,
   0x02, 0x00, 0x04, 0x02, 0x00, 0x04, 0xfe, 0xff, 0x07};
} \
mail_in {#define mail-in_width 20
#define mail-in_height 23
static char mail-in_bits[] = {
   0xfe, 0xff, 0x07, 0x06, 0x00, 0x06, 0x0a, 0x00, 0x05, 0x12, 0x80, 0x04,
   0x62, 0x60, 0x04, 0x82, 0x10, 0x04, 0x02, 0x09, 0x04, 0x02, 0x06, 0x04,
   0x02, 0x00, 0x04, 0x02, 0x00, 0x04, 0x02, 0x00, 0x04, 0x02, 0x1c, 0x04,
   0xfe, 0xff, 0x07, 0x00, 0x1c, 0x00, 0x00, 0x1c, 0x00, 0x00, 0x7f, 0x00,
   0x00, 0x3e, 0x00, 0x00, 0x1c, 0x00, 0x00, 0x08, 0x00, 0x0f, 0x00, 0x0f,
   0xf9, 0xff, 0x09, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
mail_out {#define mail-out_width 20
#define mail-out_height 23
static char mail-out_bits[] = {
   0x00, 0x01, 0x00, 0x80, 0x03, 0x00, 0xc0, 0x07, 0x00, 0xe0, 0x0f, 0x00,
   0x80, 0x03, 0x00, 0x80, 0x03, 0x00, 0xfe, 0xff, 0x07, 0x86, 0x03, 0x06,
   0x0a, 0x00, 0x05, 0x12, 0x80, 0x04, 0x62, 0x60, 0x04, 0x82, 0x10, 0x04,
   0x02, 0x09, 0x04, 0x02, 0x06, 0x04, 0x02, 0x00, 0x04, 0x02, 0x00, 0x04,
   0x02, 0x00, 0x04, 0xfe, 0xff, 0x07, 0x00, 0x00, 0x00, 0x0f, 0x00, 0x0f,
   0xf9, 0xff, 0x09, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
map {#define map_width 20
#define map_height 23
static char map_bits[] = {
   0x24, 0x40, 0x00, 0x24, 0x40, 0x00, 0x44, 0x40, 0x00, 0xbf, 0xff, 0x0f,
   0x44, 0x40, 0x00, 0x44, 0x40, 0x00, 0x84, 0x40, 0x00, 0x84, 0x40, 0x00,
   0x84, 0xb9, 0x0f, 0x84, 0xbf, 0x0f, 0xc4, 0x47, 0x00, 0xc4, 0x40, 0x00,
   0x64, 0x40, 0x00, 0x64, 0x40, 0x00, 0x64, 0x40, 0x00, 0x9f, 0xff, 0x0f,
   0x64, 0x40, 0x00, 0x64, 0x40, 0x00, 0x64, 0x40, 0x0e, 0x64, 0xb0, 0x01,
   0xc4, 0x4f, 0x00, 0xc4, 0x40, 0x00, 0xc4, 0x40, 0x00};
} \
mouse {#define mouse_width 20
#define mouse_height 23
static char mouse_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x01, 0x00, 0x00, 0x06, 0x00,
   0x00, 0x18, 0x00, 0x00, 0x20, 0x00, 0x00, 0x20, 0x00, 0xf0, 0xff, 0x00,
   0x08, 0x04, 0x01, 0x08, 0x04, 0x01, 0x08, 0x04, 0x01, 0x08, 0x04, 0x01,
   0x08, 0x04, 0x01, 0xf8, 0xff, 0x01, 0x08, 0x00, 0x01, 0x08, 0x00, 0x01,
   0x08, 0x80, 0x01, 0x08, 0x00, 0x01, 0x08, 0x80, 0x01, 0x08, 0x40, 0x01,
   0x10, 0xd0, 0x00, 0xe0, 0x6a, 0x00, 0x80, 0x1f, 0x00};
} \
network {#define network_width 20
#define network_height 23
static char network_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xf8, 0x03, 0x00, 0x08, 0x02, 0x00,
   0x08, 0x02, 0x00, 0x08, 0x02, 0x00, 0xf8, 0x03, 0x00, 0xfc, 0x07, 0x00,
   0x04, 0x04, 0x00, 0xfc, 0x07, 0x00, 0x40, 0x00, 0x00, 0x40, 0x00, 0x00,
   0xff, 0xff, 0x0f, 0x00, 0x20, 0x00, 0x00, 0x20, 0x00, 0x00, 0xfc, 0x01,
   0x00, 0x04, 0x01, 0x00, 0x04, 0x01, 0x00, 0x04, 0x01, 0x00, 0xfc, 0x01,
   0x00, 0xfe, 0x03, 0x00, 0x02, 0x02, 0x00, 0xfe, 0x03};
} \
next {#define next_width 20
#define next_height 23
static char next_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x0f,
   0x01, 0x00, 0x08, 0x81, 0x00, 0x08, 0x81, 0x01, 0x08, 0x81, 0x03, 0x08,
   0x81, 0x07, 0x08, 0x81, 0x0f, 0x08, 0x81, 0x1f, 0x08, 0x81, 0x3f, 0x08,
   0x81, 0x7f, 0x08, 0x81, 0x7f, 0x08, 0x81, 0x3f, 0x08, 0x81, 0x1f, 0x08,
   0x81, 0x0f, 0x08, 0x81, 0x07, 0x08, 0x81, 0x03, 0x08, 0x81, 0x01, 0x08,
   0x81, 0x00, 0x08, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
notebook {#define notebook_width 20
#define notebook_height 23
static char notebook_bits[] = {
   0x00, 0x00, 0x00, 0x90, 0x24, 0x01, 0xfc, 0xff, 0x0f, 0x94, 0x24, 0x09,
   0x94, 0x24, 0x09, 0x04, 0x00, 0x08, 0x04, 0x00, 0x08, 0x04, 0x00, 0x08,
   0x04, 0x0e, 0x08, 0x84, 0x09, 0x08, 0x44, 0x09, 0x08, 0x24, 0x06, 0x08,
   0x14, 0x06, 0x08, 0x0c, 0x03, 0x08, 0x84, 0x01, 0x08, 0xc2, 0xe0, 0x0f,
   0x61, 0x20, 0x0c, 0x31, 0x20, 0x0a, 0x1d, 0x20, 0x09, 0x0e, 0xa0, 0x08,
   0x04, 0x60, 0x08, 0xfc, 0x3f, 0x08, 0xf8, 0xff, 0x0f};
} \
parent {#define parent_width 20
#define parent_height 23
static char parent_bits[] = {
   0x00, 0xf8, 0x01, 0x00, 0x08, 0x02, 0x00, 0x08, 0x04, 0x80, 0x08, 0x08,
   0x80, 0x09, 0x08, 0xc0, 0x0b, 0x08, 0xe0, 0x0f, 0x08, 0xf0, 0x0b, 0x08,
   0xb8, 0x09, 0x08, 0xb8, 0x08, 0x08, 0x38, 0x08, 0x08, 0x38, 0x08, 0x08,
   0x38, 0x08, 0x08, 0x38, 0xf8, 0x0f, 0x38, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x1c, 0x00, 0x00, 0x24, 0x00, 0x00, 0x44, 0x00, 0x00, 0x44, 0x00, 0x00,
   0x44, 0x00, 0x00, 0x44, 0x00, 0x00, 0x7c, 0x00, 0x00};
} \
postscript {#define postscript_width 20
#define postscript_height 23
static char postscript_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x08, 0x00,
   0x00, 0x0d, 0x00, 0x80, 0x06, 0x00, 0x7c, 0xe3, 0x03, 0xbc, 0xe1, 0x03,
   0xde, 0xf0, 0x07, 0x6e, 0xf0, 0x07, 0x36, 0xf8, 0x07, 0xe2, 0xf8, 0x07,
   0x9e, 0xfc, 0x07, 0xfe, 0xfe, 0x07, 0x7e, 0xfe, 0x07, 0x7e, 0xff, 0x07,
   0x7c, 0xff, 0x03, 0xbc, 0xff, 0x07, 0x48, 0x55, 0x01, 0x02, 0x00, 0x04,
   0xa4, 0xa9, 0x02, 0x8a, 0xa4, 0x02, 0x28, 0xaa, 0x00};
} \
previous {#define previous_width 20
#define previous_height 23
static char previous_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x0f,
   0x01, 0x00, 0x08, 0x01, 0x10, 0x08, 0x01, 0x18, 0x08, 0x01, 0x1c, 0x08,
   0x01, 0x1e, 0x08, 0x01, 0x1f, 0x08, 0x81, 0x1f, 0x08, 0xc1, 0x1f, 0x08,
   0xe1, 0x1f, 0x08, 0xe1, 0x1f, 0x08, 0xc1, 0x1f, 0x08, 0x81, 0x1f, 0x08,
   0x01, 0x1f, 0x08, 0x01, 0x1e, 0x08, 0x01, 0x1c, 0x08, 0x01, 0x18, 0x08,
   0x01, 0x10, 0x08, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
printer {#define printer_width 20
#define printer_height 23
static char printer_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0xfc, 0x1f, 0x00, 0x04, 0x30, 0x00, 0x04, 0x50, 0x00, 0x04, 0x90, 0x00,
   0x04, 0x10, 0x01, 0x04, 0xf0, 0x03, 0x04, 0x00, 0x02, 0x04, 0x00, 0x02,
   0x04, 0x00, 0x02, 0x04, 0x00, 0x02, 0x07, 0x00, 0x0e, 0x05, 0x00, 0x0a,
   0xfd, 0xff, 0x0b, 0x01, 0x00, 0x08, 0x01, 0x00, 0x08, 0x01, 0x00, 0x08,
   0x01, 0x00, 0x08, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
sadsmiley {#define sadsmiley_width 20
#define sadsmiley_height 23
#define sadsmiley_x_hot 9
#define sadsmiley_y_hot 15
static char sadsmiley_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x1f, 0x00, 0x60, 0x60, 0x00,
   0x10, 0x80, 0x00, 0x08, 0x00, 0x01, 0x04, 0x00, 0x02, 0x62, 0x60, 0x04,
   0xf2, 0xf0, 0x04, 0x61, 0x60, 0x08, 0x01, 0x06, 0x08, 0x01, 0x06, 0x08,
   0x01, 0x06, 0x08, 0x01, 0x06, 0x08, 0x01, 0x00, 0x08, 0x02, 0x0f, 0x04,
   0xc2, 0x30, 0x04, 0x24, 0x40, 0x02, 0x08, 0x00, 0x01, 0x10, 0x80, 0x00,
   0x60, 0x60, 0x00, 0x80, 0x1f, 0x00, 0x00, 0x00, 0x00};
} \
smiley {#define smiley_width 20
#define smiley_height 23
#define smiley_x_hot 9
#define smiley_y_hot 15
static char smiley_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x1f, 0x00, 0x60, 0x60, 0x00,
   0x10, 0x80, 0x00, 0x08, 0x00, 0x01, 0x04, 0x00, 0x02, 0x62, 0x60, 0x04,
   0xf2, 0xf0, 0x04, 0x61, 0x60, 0x08, 0x01, 0x06, 0x08, 0x01, 0x06, 0x08,
   0x01, 0x06, 0x08, 0x01, 0x06, 0x08, 0x01, 0x00, 0x08, 0x12, 0x80, 0x04,
   0x22, 0x40, 0x04, 0xc4, 0x30, 0x02, 0x08, 0x0f, 0x01, 0x10, 0x80, 0x00,
   0x60, 0x60, 0x00, 0x80, 0x1f, 0x00, 0x00, 0x00, 0x00};
} \
stop {#define stop_width 20
#define stop_height 23
static char stop_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0xe0, 0x7f, 0x00, 0x10, 0x80, 0x00, 0xc8, 0x3f, 0x01, 0xe4, 0x7f, 0x02,
   0xf2, 0xff, 0x04, 0xf9, 0xff, 0x09, 0x4d, 0x24, 0x0a, 0xed, 0xaa, 0x0a,
   0xed, 0xaa, 0x0a, 0x8d, 0x2a, 0x0a, 0xbd, 0xaa, 0x0b, 0xbd, 0xaa, 0x0b,
   0x8d, 0xa2, 0x0b, 0xf9, 0xff, 0x09, 0xf2, 0xff, 0x04, 0xe4, 0x7f, 0x02,
   0xc8, 0x3f, 0x01, 0x10, 0x80, 0x00, 0xe0, 0x7f, 0x00};
} \
summary {#define summary_width 20
#define summary_height 23
static char summary_bits[] = {
   0x00, 0x00, 0xf0, 0x00, 0x00, 0xf0, 0x00, 0x00, 0xf0, 0x00, 0x00, 0xf0,
   0xc0, 0x1f, 0xf0, 0xf0, 0x7f, 0xf0, 0xf8, 0xff, 0xf0, 0xfc, 0xff, 0xf1,
   0x1e, 0xc0, 0xf3, 0x3e, 0x9f, 0xf3, 0x7f, 0xfe, 0xf7, 0xff, 0xfc, 0xf7,
   0xff, 0xf9, 0xf7, 0xff, 0xf9, 0xf7, 0xff, 0xfe, 0xf7, 0x7f, 0xff, 0xf7,
   0xbf, 0xbf, 0xf7, 0x1e, 0x80, 0xf3, 0x1e, 0xc0, 0xf3, 0xfc, 0xff, 0xf1,
   0xf8, 0xff, 0xf0, 0xf0, 0x7f, 0xf0, 0xc0, 0x1f, 0xf0};
} \
telephone {#define telephone_width 20
#define telephone_height 23
static char telephone_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x38, 0x00, 0x00, 0x44, 0x00, 0x00, 0x44, 0x00, 0x00, 0x82, 0x00, 0x00,
   0x82, 0x00, 0x00, 0x82, 0x00, 0x00, 0x44, 0x00, 0x00, 0x44, 0x00, 0x00,
   0x88, 0x00, 0x00, 0x08, 0x01, 0x00, 0x10, 0x02, 0x00, 0x10, 0x44, 0x00,
   0x20, 0xa8, 0x01, 0x40, 0x10, 0x02, 0x80, 0x00, 0x04, 0x00, 0x01, 0x04,
   0x00, 0x06, 0x04, 0x00, 0x18, 0x03, 0x00, 0xe0, 0x00};
} \
telnet {#define telnet_width 20
#define telnet_height 23
static char telnet_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x0f,
   0x01, 0x00, 0x08, 0xfd, 0xdf, 0x08, 0x31, 0xc3, 0x08, 0x31, 0xc3, 0x08,
   0x31, 0xcf, 0x08, 0x31, 0xc3, 0x08, 0x31, 0xc3, 0x08, 0x31, 0xdf, 0x0b,
   0x01, 0x00, 0x08, 0xff, 0xff, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x10, 0x80, 0x00, 0x18, 0x80, 0x01, 0xfc, 0xff, 0x03, 0xfe, 0xff, 0x07,
   0xfc, 0xff, 0x03, 0x18, 0x80, 0x01, 0x10, 0x80, 0x00};
} \
text_document {#define text_document_width 20
#define text_document_height 23
static char text_document_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0xfc, 0x1f, 0x00, 0x04, 0x30, 0x00, 0x04, 0x50, 0x00, 0x04, 0x90, 0x00,
   0xf4, 0x17, 0x01, 0x04, 0xf0, 0x03, 0xf4, 0x01, 0x02, 0x04, 0x00, 0x02,
   0x04, 0x00, 0x02, 0xf4, 0x3f, 0x02, 0x04, 0x00, 0x02, 0x04, 0x00, 0x02,
   0xf4, 0x0f, 0x02, 0x04, 0x00, 0x02, 0xf4, 0x3f, 0x02, 0x04, 0x00, 0x02,
   0x04, 0x00, 0x02, 0x04, 0x00, 0x02, 0xfc, 0xff, 0x03};
} \
tn3270 {#define tn3270_width 20
#define tn3270_height 23
static char tn3270_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x0f,
   0x01, 0x00, 0x08, 0xfd, 0x46, 0x08, 0x31, 0x4e, 0x08, 0x31, 0x4e, 0x08,
   0x31, 0x5a, 0x08, 0x31, 0x72, 0x08, 0x31, 0x72, 0x08, 0x31, 0x62, 0x08,
   0x01, 0x00, 0x08, 0xff, 0xff, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x10, 0x80, 0x00, 0x18, 0x80, 0x01, 0xfc, 0xff, 0x03, 0xfe, 0xff, 0x07,
   0xfc, 0xff, 0x03, 0x18, 0x80, 0x01, 0x10, 0x80, 0x00};
} \
toc {#define toc_width 20
#define toc_height 23
static char toc_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x0f, 0x00,
   0x80, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfc, 0x1f, 0x03,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x03, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0xfc, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x7c, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x03,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfc, 0x07, 0x03};
} \
trash {#define trash_width 20
#define trash_height 23
static char trash_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfc, 0xff, 0x03,
   0x02, 0x00, 0x04, 0xfe, 0xff, 0x07, 0x06, 0x00, 0x06, 0x44, 0x44, 0x02,
   0x24, 0x22, 0x02, 0x24, 0x22, 0x02, 0x24, 0x22, 0x02, 0x24, 0x22, 0x02,
   0x24, 0x22, 0x02, 0x24, 0x22, 0x02, 0x24, 0x22, 0x02, 0x24, 0x22, 0x02,
   0x24, 0x22, 0x02, 0x24, 0x22, 0x02, 0x24, 0x22, 0x02, 0x24, 0x22, 0x02,
   0x24, 0x22, 0x02, 0x04, 0x00, 0x02, 0xf8, 0xff, 0x01};
} \
unknown_document {#define unknown_document_width 20
#define unknown_document_height 23
static char unknown_document_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0xfc, 0x1f, 0x00, 0x04, 0x30, 0x00, 0x04, 0x50, 0x00, 0x04, 0x90, 0x00,
   0x04, 0x10, 0x01, 0x04, 0xf0, 0x03, 0x04, 0x00, 0x02, 0x04, 0x0f, 0x02,
   0x84, 0x19, 0x02, 0x84, 0x19, 0x02, 0x04, 0x18, 0x02, 0x04, 0x0c, 0x02,
   0x04, 0x06, 0x02, 0x04, 0x06, 0x02, 0x04, 0x00, 0x02, 0x04, 0x06, 0x02,
   0x04, 0x06, 0x02, 0x04, 0x00, 0x02, 0xfc, 0xff, 0x03};
} \
uuencoded_document {#define uuencoded_document_width 20
#define uuencoded_document_height 23
static char uuencoded_document_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0xfc, 0x1f, 0x00, 0x04, 0x30, 0x00, 0x04, 0x50, 0x00, 0x04, 0x90, 0x00,
   0x04, 0x10, 0x01, 0x04, 0xf0, 0x03, 0x04, 0x00, 0x02, 0x04, 0x00, 0x02,
   0x04, 0x00, 0x02, 0xb4, 0xd9, 0x02, 0xb4, 0xd9, 0x02, 0xb4, 0xd9, 0x02,
   0xb4, 0xd9, 0x02, 0xb4, 0xd9, 0x02, 0xb4, 0xd9, 0x02, 0xe4, 0x70, 0x02,
   0x04, 0x00, 0x02, 0x04, 0x00, 0x02, 0xfc, 0xff, 0x03};
} \
trademark {#define copyright_width 16
#define copyright_height 16
static char copyright_bits[] = {
   0xf0, 0x07, 0x08, 0x08, 0x04, 0x10, 0xe2, 0x21, 0x21, 0x42, 0x21, 0x44,
   0x21, 0x44, 0xe1, 0x43, 0x21, 0x41, 0x21, 0x41, 0x21, 0x42, 0x22, 0x22,
   0x04, 0x10, 0x08, 0x08, 0xf0, 0x07, 0x00, 0x00};
} \
copyright {#define copyright_width 15
#define copyright_height 15
static unsigned char copyright_bits[] = {
 0xc0,0x01,0x30,0x06,0x0c,0x18,0xe4,0x11,0x32,0x26,0x1a,0x2e,0x19,0x40,0x19,
 0x40,0x19,0x40,0x1a,0x2e,0x32,0x26,0xe4,0x11,0x0c,0x18,0x30,0x06,0xc0,0x01};
} \
author {#define author_width 16
#define author_height 16
static char author_bits[] = {
   0x00, 0x00, 0x00, 0x10, 0x00, 0x49, 0x00, 0x2a, 0x60, 0x7a, 0x80, 0x9e,
   0x98, 0x27, 0xe0, 0x49, 0x68, 0x92, 0xb2, 0xa4, 0x3c, 0x29, 0x58, 0x29,
   0x4c, 0x0a, 0x96, 0x0a, 0xa6, 0x02, 0x01, 0x00};
} \
top {#define top_width 16
#define top_height 16
static char top_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfc, 0x3f, 0xfc, 0x3f, 0xfc, 0x3f,
   0x00, 0x00, 0x80, 0x01, 0xc0, 0x03, 0xe0, 0x07, 0xf0, 0x0f, 0xf8, 0x1f,
   0xfc, 0x3f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
} \
made {#define author_width 16
#define author_height 16
static char author_bits[] = {
   0x00, 0x00, 0x00, 0x10, 0x00, 0x49, 0x00, 0x2a, 0x60, 0x7a, 0x80, 0x9e,
   0x98, 0x27, 0xe0, 0x49, 0x68, 0x92, 0xb2, 0xa4, 0x3c, 0x29, 0x58, 0x29,
   0x4c, 0x0a, 0x96, 0x0a, 0xa6, 0x02, 0x01, 0x00};
} \
contents {#define toc_width 20
#define toc_height 23
static char toc_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x0f, 0x00,
   0x80, 0x0f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfc, 0x1f, 0x03,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x03, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0xfc, 0x03, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x7c, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x03,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfc, 0x07, 0x03};
} \
prev {#define previous_width 20
#define previous_height 23
static char previous_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x0f,
   0x01, 0x00, 0x08, 0x01, 0x10, 0x08, 0x01, 0x18, 0x08, 0x01, 0x1c, 0x08,
   0x01, 0x1e, 0x08, 0x01, 0x1f, 0x08, 0x81, 0x1f, 0x08, 0xc1, 0x1f, 0x08,
   0xe1, 0x1f, 0x08, 0xe1, 0x1f, 0x08, 0xc1, 0x1f, 0x08, 0x81, 0x1f, 0x08,
   0x01, 0x1f, 0x08, 0x01, 0x1e, 0x08, 0x01, 0x1c, 0x08, 0x01, 0x18, 0x08,
   0x01, 0x10, 0x08, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
begin {#define begin_width 20
#define begin_height 23
static char begin_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x0f,
   0x01, 0x00, 0x08, 0x61, 0x20, 0x08, 0x61, 0x30, 0x08, 0x61, 0x38, 0x08,
   0x61, 0x3c, 0x08, 0x61, 0x3e, 0x08, 0x61, 0x3f, 0x08, 0xe1, 0x3f, 0x08,
   0xe1, 0x3f, 0x08, 0xe1, 0x3f, 0x08, 0xe1, 0x3f, 0x08, 0x61, 0x3f, 0x08,
   0x61, 0x3e, 0x08, 0x61, 0x3c, 0x08, 0x61, 0x38, 0x08, 0x61, 0x30, 0x08,
   0x61, 0x20, 0x08, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
first {#define begin_width 20
#define begin_height 23
static char begin_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x0f,
   0x01, 0x00, 0x08, 0x61, 0x20, 0x08, 0x61, 0x30, 0x08, 0x61, 0x38, 0x08,
   0x61, 0x3c, 0x08, 0x61, 0x3e, 0x08, 0x61, 0x3f, 0x08, 0xe1, 0x3f, 0x08,
   0xe1, 0x3f, 0x08, 0xe1, 0x3f, 0x08, 0xe1, 0x3f, 0x08, 0x61, 0x3f, 0x08,
   0x61, 0x3e, 0x08, 0x61, 0x3c, 0x08, 0x61, 0x38, 0x08, 0x61, 0x30, 0x08,
   0x61, 0x20, 0x08, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
last {#define last_width 20
#define last_height 23
static char last_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x0f,
   0x01, 0x00, 0x08, 0x21, 0x60, 0x08, 0x61, 0x60, 0x08, 0xe1, 0x60, 0x08,
   0xe1, 0x61, 0x08, 0xe1, 0x63, 0x08, 0xe1, 0x67, 0x08, 0xe1, 0x6f, 0x08,
   0xe1, 0x7f, 0x08, 0xe1, 0x7f, 0x08, 0xe1, 0x6f, 0x08, 0xe1, 0x67, 0x08,
   0xe1, 0x63, 0x08, 0xe1, 0x61, 0x08, 0xe1, 0x60, 0x08, 0x61, 0x60, 0x08,
   0x21, 0x60, 0x08, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
end {#define last_width 20
#define last_height 23
static char last_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0x0f,
   0x01, 0x00, 0x08, 0x21, 0x60, 0x08, 0x61, 0x60, 0x08, 0xe1, 0x60, 0x08,
   0xe1, 0x61, 0x08, 0xe1, 0x63, 0x08, 0xe1, 0x67, 0x08, 0xe1, 0x6f, 0x08,
   0xe1, 0x7f, 0x08, 0xe1, 0x7f, 0x08, 0xe1, 0x6f, 0x08, 0xe1, 0x67, 0x08,
   0xe1, 0x63, 0x08, 0xe1, 0x61, 0x08, 0xe1, 0x60, 0x08, 0x61, 0x60, 0x08,
   0x21, 0x60, 0x08, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
origin {#define top_width 16
#define top_height 16
static char top_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xfc, 0x3f, 0xfc, 0x3f, 0xfc, 0x3f,
   0x00, 0x00, 0x80, 0x01, 0xc0, 0x03, 0xe0, 0x07, 0xf0, 0x0f, 0xf8, 0x1f,
   0xfc, 0x3f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
} \
child {#define child_width 20
#define child_height 23
static char child_bits[] = {
   0x00, 0xf8, 0x01, 0x00, 0x08, 0x02, 0x00, 0x08, 0x04, 0x00, 0x08, 0x08,
   0xe0, 0x0f, 0x08, 0xf0, 0x0f, 0x08, 0xf0, 0x0f, 0x08, 0x78, 0x08, 0x08,
   0x38, 0x08, 0x08, 0x38, 0x08, 0x08, 0x38, 0x08, 0x08, 0x38, 0x08, 0x08,
   0xfe, 0x08, 0x08, 0x7c, 0xf8, 0x0f, 0x38, 0x00, 0x00, 0x10, 0x00, 0x00,
   0x1c, 0x00, 0x00, 0x24, 0x00, 0x00, 0x44, 0x00, 0x00, 0x44, 0x00, 0x00,
   0x44, 0x00, 0x00, 0x44, 0x00, 0x00, 0x7c, 0x00, 0x00};
} \
sibling {#define sibling_width 20
#define sibling_height 23
static char sibling_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x3f, 0xf8, 0x01, 0x41, 0x08, 0x02, 0x81, 0x08, 0x04,
   0x01, 0x09, 0x08, 0x01, 0x09, 0x08, 0x41, 0x29, 0x08, 0x21, 0x49, 0x08,
   0xf1, 0xff, 0x08, 0x21, 0x49, 0x08, 0x41, 0x29, 0x08, 0x01, 0x09, 0x08,
   0x01, 0x09, 0x08, 0x01, 0x09, 0x08, 0xff, 0xf9, 0x0f, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
} \
navigator {#define navigator_width 20
#define navigator_height 23
static char navigator_bits[] = {
   0x00, 0x00, 0x00, 0xff, 0xf8, 0x07, 0xab, 0x58, 0x05, 0xd5, 0xa8, 0x06,
   0xab, 0x58, 0x05, 0xd5, 0xa8, 0x06, 0xab, 0x58, 0x05, 0xd5, 0xa8, 0x06,
   0xab, 0x58, 0x05, 0xd5, 0xaf, 0x06, 0xab, 0x5f, 0x05, 0xd5, 0xa8, 0x06,
   0xfe, 0xff, 0x03, 0xfe, 0xff, 0x03, 0x7e, 0xf0, 0x03, 0x7e, 0xf0, 0x03,
   0x7e, 0xf0, 0x03, 0x7e, 0xf0, 0x03, 0x38, 0xe0, 0x00, 0x38, 0xe0, 0x00,
   0x38, 0xe0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
} \
disclaimer {#define disclaimer_width 20
#define disclaimer_height 23
static char disclaimer_bits[] = {
   0x00, 0x0e, 0x00, 0x00, 0x0e, 0x00, 0x00, 0x1b, 0x00, 0x00, 0x1b, 0x00,
   0x80, 0x31, 0x00, 0x80, 0x31, 0x00, 0xc0, 0x30, 0x00, 0xc0, 0x64, 0x00,
   0x60, 0x6e, 0x00, 0x60, 0xce, 0x00, 0x20, 0xce, 0x00, 0x30, 0xce, 0x00,
   0x30, 0x8e, 0x01, 0x18, 0x8e, 0x01, 0x18, 0x0e, 0x03, 0x0c, 0x0e, 0x03,
   0x0c, 0x04, 0x02, 0x06, 0x04, 0x06, 0x06, 0x00, 0x06, 0x03, 0x04, 0x0c,
   0x03, 0x04, 0x0c, 0x01, 0x00, 0x08, 0xff, 0xff, 0x0f};
} \
publisher {#define publisher_width 20
#define publisher_height 23
static char publisher_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x60, 0x60, 0x00, 0xe0, 0x7f, 0x00, 0xe0, 0x7f, 0x00,
   0x60, 0x66, 0x00, 0x00, 0x06, 0x00, 0x00, 0x06, 0x00, 0x00, 0x06, 0x00,
   0x00, 0x06, 0x00, 0xe0, 0x7f, 0x00, 0x20, 0x46, 0x00, 0x20, 0x46, 0x00,
   0x20, 0x46, 0x00, 0x20, 0x46, 0x00, 0xfc, 0xff, 0x03, 0xfc, 0xff, 0x03,
   0x04, 0x00, 0x02, 0xfc, 0xff, 0x03, 0xfc, 0xff, 0x03};
} \
meta {#define meta_width 20
#define meta_height 23
static char meta_bits[] = {
   0x00, 0x00, 0x00, 0xe0, 0x3f, 0x00, 0x10, 0x40, 0x00, 0x08, 0x80, 0x00,
   0x04, 0x07, 0x01, 0x02, 0x07, 0x02, 0x01, 0x00, 0x04, 0x81, 0x07, 0x04,
   0x01, 0x07, 0x04, 0x01, 0x07, 0x04, 0x01, 0x07, 0x04, 0x01, 0x07, 0x04,
   0x01, 0x07, 0x04, 0x01, 0x07, 0x04, 0x01, 0x07, 0x04, 0x02, 0x07, 0x02,
   0xe4, 0x3f, 0x01, 0x08, 0x80, 0x00, 0x10, 0x40, 0x00, 0xe0, 0x3f, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
} \
]
proc surfit_install {} {
    global tcl_platform
    toplevel .install
    wm title .install {Install Plume}
    label .install.lib -text {Install library files in directory:}
    entry .install.libdir -width 40
    label .install.bin -text {Install executable program as:}
    entry .install.binpath -width 40
    label .install.user -text {Default Plume user directory:}
    entry .install.userdir -width 40
    switch $tcl_platform(platform) {
	unix {
	    .install.libdir insert 0 /usr/local/share/lib/plume
	    .install.binpath insert 0 /usr/local/bin/plume
	    .install.userdir insert 0 \$env(HOME)/.plume
	}
	windows {
	    .install.libdir insert 0 c:/plume/lib
	    .install.binpath insert 0 c:/plume/surfit.tcl
	    .install.userdir insert 0 c:/plume
	}
    }
    label .install.feedback
    button .install.ok -text Install -command surfit_do_install
    button .install.cancel -text Cancel -command {destroy .install}
    grid .install.lib .install.libdir -row 0 -sticky news
    grid .install.bin .install.binpath -row 1 -sticky news
    grid .install.user .install.userdir -row 2 -sticky news
    grid .install.feedback - -row 3 -sticky news
    grid .install.ok -row 4 -column 0 -sticky w
    grid .install.cancel -row 4 -column 1 -sticky e
    grid rowconfigure .install 2 -weight 1
    grid columnconfigure .install 1 -weight 1
}
proc surfit_cancel_install {} {
    global surfit_install_cancel
    .install.feedback configure -text {installation cancelled}
    .install.cancel configure -command {destroy .install}
    catch {unset surfit_install_cancel}
}
proc surfit_do_install {} {
    global scripts bitmaps PRAbout tcl_platform surfit_install_cancel

    set libdir [file split [File_normalizeFile [.install.libdir get]]]
    set binpath [file split [File_normalizeFile [.install.binpath get]]]
    set userdir [file split [File_normalizeFile [.install.userdir get]]]
    # Preparation of directories, etc.
    if {[file exists [eval file join $libdir]]} {
	if {![file isdirectory [eval file join $libdir]]} {
	    .install.feedback configure -text "[eval file join $libdir] is not a directory"
	    bell
	    return
	} elseif {![file writable [eval file join $libdir]]} {
	    .install.feedback configure -text "cannot write to directory [eval file join $libdir]"
	    bell
	    return
	}
    } else {
	if {[catch {File_mkdir $libdir}]} {
	    .install.feedback configure -text "error creating directory [eval file join $libdir]"
	    bell
	    return
	}
    }
    if {[file exists [eval file join $binpath]]} {
	if {[file isdirectory [eval file join $binpath]]} {
	    .install.feedback configure -text "[eval file join $binpath] is a directory"
	    bell
	    return
	}
    }
    if {![file writable [file dirname [eval file join $binpath]]]} {
	.install.feedback configure -text "cannot write to directory [file dirname [eval file join $binpath]]"
	bell
	return
    }

    set bmpdir "$libdir icons"
    if {[file exists [eval file join $bmpdir]]} {
	if {![file isdirectory [eval file join $bmpdir]]} {
	    .install.feedback configure -text "[eval file join $bmpdir] is not a directory"
	    bell
	    return
	} elseif {![file writable [eval file join $bmpdir]]} {
	    .install.feedback configure -text "cannot write to directory [eval file join $bmpdir]"
	    bell
	    return
	}
    } else {
	if {[catch {File_mkdir $bmpdir}]} {
	    .install.feedback configure -text "error creating directory [eval file join $bmpdir]"
	    bell
	    return
	}
    }

    .install.cancel configure -command {surfit_cancel_install}

    .install.feedback configure -text {installing library scripts}
    update
    if {[info exists surfit_install_cancel]} return

    foreach script [array names scripts] {
    	set filename "$libdir $script"
	set fd [open [eval file join $filename] w]
	puts $fd $scripts($script)
	close $fd
    }
    pkg_mkIndex [eval file join $libdir] fsbox.tcl uri.tcl prot.tcl cache.tcl htmllib.tcl safe-tk.tcl applets.tcl handlers.tcl hypetool.tcl mailto.tcl internal.tcl  utils.tcl
    .install.feedback configure -text {installing bitmaps}
    update
    if {[info exists surfit_install_cancel]} return
    foreach bitmap [array names bitmaps] {
	set filename "$bmpdir $bitmap.xbm"
	set fd [open [eval file join $filename] w]
	puts $fd $bitmaps($bitmap)
	close $fd
    }

    ### Write out surfit program
    .install.feedback configure -text {installing plume program}
    update
    if {[info exists surfit_install_cancel]} return

    set fd [open [eval file join $binpath] w]
    puts $fd #!/bin/sh\n#\ -*-\ tcl\ -*-\ \\\nexec\ wish8.0\ \"\$0\"\ \"\$@\"\n\n#\tCopyright\ (c)\ 1996,1997\ Australian\ National\ University\n#\n#--------\n#\n#\ Copyright\ (c)\ 1995,1996\ ANU\ and\ CSIRO\ on\ behalf\ of\ the\ participants\ in\ the\ \n#\ CRC\ for\ Advanced\ Computational\ Systems\ ('ACSys').\n#\n#\ ACSys\ makes\ this\ software\ and\ all\ associated\ data\ and\ documentation\n#\ ('Software')\ available\ free\ of\ charge\ for\ non-commercial\ purposes\ only.\ You\n#\ may\ make\ copies\ of\ the\ Software\ but\ you\ must\ include\ all\ of\ this\ notice\ on\n#\ any\ copy.\n#\n#\ The\ Software\ was\ developed\ for\ research\ purposes\ and\ ACSys\ does\ not\ warrant\n#\ that\ it\ is\ error\ free\ or\ fit\ for\ any\ purpose.\ \ ACSys\ disclaims\ any\n#\ liability\ for\ all\ claims,\ expenses,\ losses,\ damages\ and\ costs\ any\ user\ may\n#\ incur\ as\ a\ result\ of\ using,\ copying\ or\ modifying\ the\ Software.\n
    puts $fd "lappend auto_path [eval file join $libdir]"
    puts $fd $scripts(plume-1.tcl)
    puts $fd "set datadir [list $libdir]"
    puts $fd "set userdir [list $userdir]"
    puts $fd "array set PRAbout [list [array get PRAbout]]"
    puts $fd $scripts(plume-2.tcl)
    puts $fd $scripts(plume-3.tcl)
    close $fd

    if {$tcl_platform(platform) == {unix}} {
	exec chmod a+x [eval file join $binpath]
    }

    .install.feedback configure -text {installation complete}
    .install.cancel configure -state disabled
    .install.ok configure -text OK -command {destroy .install}
}
foreach f {utils.tcl} {eval $scripts($f)}
eval $scripts(plume-1.tcl)
catch {package require Img}
foreach f {fsbox.tcl uri.tcl prot.tcl cache.tcl htmllib.tcl safe-tk.tcl} {eval $scripts($f)}
PR_init; HMlibrary_init; interp_init
foreach {f i} {applets.tcl Applet_init handlers.tcl handlers_init hypetool.tcl HT_init mailto.tcl mailto_init internal.tcl surfit_init } {eval $scripts($f); $i}
foreach bitmap [array names bitmaps] {
    regsub -all _ $bitmap . img
    image create bitmap ICON$img -data $bitmaps($bitmap)
}
array set PRAbout {plume {
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2//EN">

<html>
<head>
<title>Plume</title>
<link rev="made" href="mailto:Steve.Ball@tcltk.anu.edu.au">
</head>

<body>
<a rel=embed href="http://tcltk.anu.edu.au/tcl/welcome.tcl">
<h1 align=center>Plume - The All Tcl WWW Browser</h1>
</a>

<hr>

Copyright &copy; 1996, 1997 Australian National University.
<p>

Copyright &copy; 1995,1996 ANU and CSIRO on behalf of the participants in the 
CRC for Advanced Computational Systems ('ACSys').

<p>
ACSys makes this software and all associated data and documentation
('Software') available free of charge for non-commercial purposes only. You
may make copies of the Software but you must include all of this notice on
any copy.

<p>
The Software was developed for research purposes and ACSys does not warrant
that it is error free or fit for any purpose.  ACSys disclaims any
liability for all claims, expenses, losses, damages and costs any user may
incur as a result of using, copying or modifying the Software.

<hr>

Welcome to the Plume World Wide Web browser.  You can find out more
about Plume on the <a href="http://tcltk.anu.edu.au/">Plume
Home Page</a>.  You can also find:

<ul>
<li dingbat="mouse"><a href="http://tcltk.anu.edu.au/0.6.1alpha/install.html">Installation instructions and help</a>
<li dingbat="mouse"><a href="http://tcltk.anu.edu.au/0.6.1alpha/release.notes.html">Release Notes</a>
<li dingbat="mouse"><a href="http://tcltk.anu.edu.au/demos/demos.html">Demos and Examples</a>
</ul>

</body>
</html>
}
}
if {$surfit(CFGwwwhome) == {}} {set surfit(CFGwwwhome) about:plume}
eval $scripts(plume-3.tcl)
.ctl-panel.file.m insert Quit command -label "Install" -command surfit_install
